/*  Copyright 1993,2003 Itty Bitty Computers.  All rights reserved. */
/*  IOglue, TagLib, TagIO, TagStack, and UILib modules, rewritten in C */

/* class TagLibs -- 2006 January 14 */

/* Table of Contents..                   // State Machine Info */
/*  General Constants //                 // String & Symbol Table Data */
/*  Tree Node Types //                   // State Machine Procs */
/*  Tree Procedures //                   // String Table Procedures */
/*  Global I/O Vars //                   // Symbol Table Procedures */
/*  Utility Functions //                 // Parser Stack Info */
/*  Low-Level I/O Functions //           // TAG File I/O Utilities */
/*  Sparse Set Types & Data //           // Debugging Utility Procedures */
/*  Sparse Set Procedures //             // User Trace Utilities */
/*  FSM-Building Utilities //            // Compiled TAG Procedures */
/*  Memory Image Functions */            /******** Prototypes ********/

/**************************** General Constants ****************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <setjmp.h>

#define RunawayTrace  1000000
 /* C doesn't kill program on file errs */

#define CHAR unsigned char
#define INTEGER int
#define BOOLEAN int
#define boolean int
#define true  1
#define false 0
#define TRUE  1
#define FALSE 0
#define null  0
#define NIL   0

#define MaxImportTrees 400000
#define EchoInputBuffer 0 /* 256 */

#define NoisyTrace  0
 /* initial value for Debugging, now automated */
int Debugging = 0; /* =0 none, =1 a few, >5 for lots; +8 traces stack */
int PreMortemLevel = 2;
int KillingIt = 0;
#define Trace_Toggle  ((char) 4)
 /* if != Null_Char, Trace_Toggle (^D) in file toggles trace Debugging=2 */

#define  InputFileName "TAGin.txt"
#define OutputFileName "TAGout.txt"
#define  DebugFileName "TAGdbg.txt"
#define  LibryFileName "Libry.txt"
#define FailMsFileName "TAGexit.txt"
#define OutputFileType "wb"
#define UseFileInput  true
#define UseFileDebug  true
#define TopSetNo  262144
 /* 256K, about 2x what we use */
#define CharMax  32700
 /* available size of string table */
#define StringMax  (4096*3)
 /* last avail string index in string table */
#define FirstString  (10*3)
 /* leave low numbers for other uses */
#define HashTop  1024
 /* string hashtable */
#define StackMax 65500
#define TransMax  2000
#define DetrnMax 20000

#define Null_Char ((CHAR) 0)
#define EOF_Char ((CHAR) 255)
#define EOL_Char '\n'
#define CRLF ((CHAR) 13)
 /* '\n' // C does skanky things behind our backs */
#define LFCR EOF_Char
 /* ((CHAR) 10) // = EOF_Char if Mac or eunuchs */

char* NodeNameS = "no\0ty\0gl\0fn\0tk\0nt\0st\0pl\0qu\0tr\0ac\0at\0mt\0ot\0cn\0rt\0xf\0bt\0er\0al\0ca\0qm\0dl\0tf\0dr\0vr\0as\0eq\0ls\0gr\0ne\0an\0or\0ad\0su\0mp\0dv\0md\0ng\0tv\0rn\0nn\0nc\0co\0sa\0pk\0ar\0ax\0bg\0ch\0cl\0cm\0cs\0cv\0df\0do\0en\0ex\0fc\0fl\0fr\0ge\0hp\0if\0in\0io\0jm\0ki\0lc\0le\0lp\0lx\0nl\0nx\0of\0pm\0pt\0rc\0rg\0rp\0rx\0sh\0sl\0sr\0sw\0sz\0un\0up\0vt\0wh\0xo\0yk\0zz\0ab\0ag\0bf\0bs\0by\0cc\0cp\0ct\0dc\0dn\0dy\0el\0ez\0fd\0fk\0fp\0ft\0go\0hi\0it\0jp\0jz\0ld\0lg\0lo\0nm\0np\0nu\0op\0ov\0pf\0pr\0rf\0sg\0sm\0sy\0tc\0th\0tp\0va\0vd\0wd\0xc\0ys\0ze\0\0";

int RunPhase = 0;
jmp_buf CatchThrow;
char Int2StrTemp[16];
char StringTemp[256];
#define ErrLnCaptLen 128
char ErrLnCapture[ErrLnCaptLen+2];
int ErrLnCapIx = 0;

/**************************** Tree Node Types ****************************/

/* final int nodeclass = ( */
#define nono  0
#define intt  1
#define bool  2
#define snam  3
#define seti  4
#define wood  5
#define symt  6
#define pkat  7
#define mark  8
 /* higher bits count & mark nodes for export or compare.. */
#define mrkn  4
#define mrkb  16
 /* one extra bit for second flag */
#define mrkt  65536
 /* recursion stopper bit in ShoTreDee */

struct s_node {
  int kind, intval, nodename;
  struct s_node * subtrees[9];};
typedef struct s_node node;
typedef node * tree;
#define nodeSize sizeof(node)
#define nodeIsize (sizeof(int)*2)
#define nodeTsize (nodeSize-7*sizeof(tree))
#define strn int
typedef tree settt;
typedef tree table;
typedef tree trees;
  /* struct s_settt extends tree // intval is UniqueSets index */

/* pointer classes used for ref parameters.. * /
  class VARb {int b = false; VARb() {} VARb(int x) {b=x;}}
  class VARi {int i = 0;         VARi() {} VARi(int x)     {i=x;}}
  class VARt {tree t = NIL;      VARt() {} VARt(tree x)    {t=x;}}
  class VARs {settt s = null;    VARs() {} VARs(settt x)   {s=x;}}
  class VARa {table a = null;    VARa() {} VARa(table x)   {a=x;}} */
typedef int * VARi;
typedef int * VARb;
typedef tree * VARt;
typedef settt * VARs;
typedef tree * VARa;

int Tree_Count = 0;
/* tree NIL = null; */

/**************************** Global I/O Vars ****************************/

typedef FILE* FilePtr;  /* InFile, CodeFile, DbgFile, LibeFile, FailFile; */
FilePtr My_Files[7] = {0,0,0,0,0,0,0};
long int FileOpLen[7] = {0,-1,-1,-1,-1,-1,-1};  /* 0=scr,1=in,2=out,3=diag... */
CHAR File_Peek[7] =
    {EOF_Char, EOF_Char, EOF_Char, EOF_Char, EOF_Char, EOF_Char, EOF_Char};

int Line_Len; /* to keep lines from getting too long */
int sourceFile = 0; /* normally =1 */
int destFile = 0;   /* normally =2 */

int CkSumReplay[64];  /* to turn up Debugging near the failure stop */
int CkSumHere = 0;    int PriorChksum = 0;    int CheckOff;
int theChecksum = 0;
char TokenSho[256];    /* prior token, for synchronous trace */
int curTokSho = 0;
char ReadReplay[1024]; /* recent input&output history, for debug */
int ReplayHere = 0;
CHAR SavedChar = ' ';
int DbugFiLen = 0;
int DbugFile = 0;   /* normally =3 */
char InputEchoBuffer[EchoInputBuffer+4];
int EchoInputSofar = 0;
int InputLinePosn = 0;

tree MagiTrees[16];   /* to tag frequently recurring trees in trace */
int SeenMagiTr[16];
int DbgOutLns = 0;    /* odometer to measure SeenMagiTr against */

int* MemImage = null;
int ImageSize = 0;

int dbgNextTime;  /* Time_Tix; tick count for next pacifier write */
int evtNextTime;  /* Time_Tix; when to ask for next event (unused) */

/**************************** Sparse Set Types & Data ****************************/

struct s_UniSetRec { /* a collection of unique sets, so equality tests work .. */
  /*  UniqueSets is a lattice of sparse set elements: */
  /*    Every UniqueSets[s] is the highest value element for one unique set */
  /*    s is its reference number */
  /*    UniqueSets[s].HedLink is the next set with the same highest element */
  /*    UniqueSets[n].EltHeds is the first such set for each n */
  /*    UniqueSets[e].ThisElt is the value of each set element e */
  /*    UniqueSets[e].NextElt is the next element in this set, or 0 if none */
  /*    UniqueSets[0] is the empty set */
  /*    if s:{i,j,k} (i>j>k) and t:{j,k}, then UniqueSets[s].NextElt=t */
    /*    and UniqueSets[s].HedLink=u where u=0 or u:{i,...} */
  int ThisElt; /* element value */
  int NextElt; /* link to next element in this set */
  int HedLink; /* link to next element in equal elt list */
  int EltHeds; /* head of each equal element list */
  settt SetPtr;
  /* UniqueSeTyp = RECORD AllSets: ARRAY [0..TopSetNo] OF UniSetRec; END; */

  /* UniSetRec() {
    SetPtr = null;
    ThisElt = 0;
    NextElt = 0;
    HedLink = 0;
    EltHeds = 0;} */ };

typedef struct s_UniSetRec UniSetRec;
UniSetRec UniqueSets[TopSetNo]; /* init`d in InitSets */
int HugeEltHash[256]; /* EltHeds of elts>TopSetNo */
int LastUniSet;
int SawSetSet; /* to minimize long trace files */

/**************************** FSM-Building Data ****************************/

struct s_transItem { /* transition tables.. */
  settt onset;
  int onchar, fromst, tost, actset;};
typedef struct s_transItem transItem;
/* newTransItem(settt ons, int onch, int froms, int tos, int acts) {
    onset = ons; / * ND Tran * /
    onchar = onch; / * Detrn * /
    fromst = froms;
    tost = tos;
    actset = acts;}} */
  /* TransListRec = RECORD tranData: ARRAY [0..TransMax] OF transItem END; */
  /* TransListPtr = POINTER TO TransListRec; */
  /* DeTrnListRec = RECORD dtrnData: ARRAY [0..DetrnMax] OF transItem END; */
  /* DeTrnListPtr = POINTER TO DeTrnListRec; */
/* transItem TransList[TransMax], DeTrnList[DetrnMax]; */
transItem* TransList; transItem* DeTrnList;
int transnumber;
int detransnmber;

/**************************** State Machine Info ****************************/

#define MuchTooBig 132000
int badChar;
int CurrentState;
CHAR ScanNextCh; /* char scannextch; */
int nCharSets, topState, SourceHere, SourceThar;
int charClass[256];

struct s_StateTrans {int action; int nextState;};
  /* StateTrans() {action = 0; nextState = 0;}} */
typedef struct s_StateTrans StateTrans;
StateTrans StateTable[MuchTooBig];

/**************************** String & Symbol Table Data ****************************/

/* final int CharRange = [0..CharMax]; */
/* final int CharTabType = ARRAY CharRange) {CHAR; */

int EndStrings, EndChars;  /* next available spot for a string/char */
CHAR LastCharacter;
int* ResWordPtr;
char* EquivClassPtr;
int NexToken, Taken, RecentStrx;
char CharTable[CharMax]; /* init`d in InitStrings */

/* HashRange = [0..HashTop]; */
/* StringRange = [0..StringMax]; */
/* StrRef = RECORD theStr: CharRange; sLink: StringRange; StrType: INTEGER END; */
    /* offsets in 3-word record.. */
#define StrOffs 0
#define sLink   1
#define StrType 2
#define StrRef  3
 /* item size */
/* AllStrs = RECORD StrAry: ARRAY StringRange OF StrRef END; */
int WorkHash; /* HashRange; // hash of current partially build string */
int StringTable[StringMax]; /* ARRAY StringRange OF StrRef // init`d in InitStrings */
int StrHashTbl[HashTop]; /* [HashRange] of StringRange */

int TabNo = 0;  /* node type for symbol table, = 1st node built */
table RecenTable;  /* for limiting debug trace size */
boolean ShoLexLvl = false;

int InLineNum = 0;

/**************************** Parser Stack Info ****************************/

/* stackIndexType = [0..StackMax]; */
/* stackType = RECORD theStk: ARRAY stackIndexType OF INTEGER END; */
int stackTop, HiStack; /* stackIndexType; points to next available slot */
int theStack[StackMax]; /* POINTER TO stackType; init`d in TagBegin */


/**************************** Prototypes ****************************/

CHAR CHR(int theCh);
int ORD(CHAR theCh);
char* IntToStr(int value);
char* CatStr(char* theStr, char* aStr);
int Tick_Count(void);
void Sys_Beep(int howLong);
void System_Task(void);
int User_Killed(void);
void CalledOften(void);
void Stop_Run(char* ErrMsg);

boolean fnmin(int a, int b, VARi c);
boolean fnmax(int a, int b, VARi c);
boolean fnbitand(int a, int b, VARi c);

tree inttree (int val);
tree strtree (int val);
tree booltree(int truly);
settt newsettt(void);
tree nutrees(int noden, tree decorn, tree s1, tree s2, tree s3, tree s4, 
      tree s5, tree s6, tree s7, tree s8, int numsubs);
tree newtable(int nam, int lex, table oldtabTb, tree theValue);

boolean fnbooltree(int inBool, VARt outtree);
trees build(int noden, tree decorn, tree s1, tree s2, tree s3, tree s4, 
    tree s5, tree s6, tree s7, tree s8, int numsubs);
boolean comparetree(tree thisn, tree thatn);
void replacetree(tree dest, tree orig);
int treeint(tree thetree);
int treestrn(tree thetree);
int treepart(tree treeptr, int opList, int compareto, VARt destTree);
tree treetree(tree thetree);
boolean fninttree(int theint, VARt outtree);
boolean fnstrtree(int theint, VARt outtree);
boolean fnintstr(int theint, VARi outint);
boolean fnstrint(int theint, VARi outint);
boolean fnpackattributes(table tab1, settt set2, settt set3, settt set4,
    tree tree5, int int6, VARa outDataTb);
boolean fnsettree(settt theSet, VARt rset);
boolean fntabtree(table theTable, VARt theTree);
boolean fntreebool(tree intree, VARb outbool);
boolean fntreeint(tree thetree, VARi outint);
boolean fntreestr(tree thetree, VARi outint);
boolean fntreeset(tree thetree, VARs rset);
boolean fntreetab(tree thetree, VARa theTable);
boolean fnunpackatts(table indataTb, VARa tab1, VARs set2, VARs set3,
    VARs set4, VARt tree5, VARi int6);
void IO_Startup(void);
long int ofsize(FILE* fp);
void Open_File(int theFile, char* prompt);
void Close_File(int theFile);
CHAR PeekInChar(int theFile);
void EchoThisInput(CHAR ch);
CHAR ReadInChar(int theFile);
void WriteOutChar(int theFile, CHAR theChar);
int EndFi(void);
int EndLn(int theFile);
int EndLine(void);
CHAR ReadCh(void);
int ReadInt(int theFile);
void ReadLin(void);
void FileClose(int fileNumber);
int FileOpen(int mode, char* prompt);
void fileWriteC(int fileRef, CHAR outChar);
void fileWriteS(int fileRef, char* outString);
void EchoCh(CHAR ch);
void ShoTokSwap(void);
boolean InitSets(void);
int myaddset(int elt, int theSet);
int myinset(int elt, settt theSet);
boolean fnaddset(int elt, settt theSet, VARs rset);
boolean fndifference(settt aset, settt bset, VARs difset);
boolean fninset(int elt, settt theSet);
boolean fnintersect(settt aset, settt bset, VARs rset);
boolean fnnewelt(settt theSet, VARi nuelt);
boolean fnnotinset(int elt, settt theSet);
boolean fnunion(settt aset, settt bset, VARs rset);
boolean fnfixadetran(int atthistran, int atfrosty, int attosty, 
    int charsClass, int atactor);
boolean fnfixatran(int atthistran, int atfrosty,
    int attosty, settt ataset, int atactor);
boolean fngetadetran(int inTran, VARi outTran, VARi atfrosty, 
    VARi attosty, VARi ataset, VARi atactor, VARi nextTran);
boolean fngetatran(int inTran, VARi outTran, VARi atfrosty, 
    VARi attosty, VARs ataset, VARi atactor, VARi atnexttr);
boolean fninitScannerStuff(void);
boolean fnnotherdetran(int atfrosty, int attosty, int ataset, int atactor);
boolean fnnothertran(int atfrosty, int attosty, settt ataset, int action);
void NewStateTable(int nitems);
void ScanStateItem(int st, int ch, int ac, int ns);
boolean MatchToken(int tkno);
int ReservedWord(int strix);
void AddResWord(int tokno, int wordty, char* theWord);
boolean fnStartScanner(boolean closeIt);
boolean fnsrctoklinch(VARi linch);
boolean InitStrings(void);
int myLength(int item);
void ShowStrings(void);
void SpellIt(int thename, int showy);
boolean fnaddtbl(int theChar);
boolean fncharfrom(int index, int item, VARi thechar);
boolean fncharval(VARi charNum);
boolean fninitbl(void);
boolean fnlength(int item, VARi outLen);
boolean fnstleng(int item, VARi outLen);
boolean fnstrindex(int equivclass, int stringType, VARi stringIndex);
boolean fnuplino(int plus, VARi newly);               /******** Prototypes ********/
boolean fnixfrom(int indx, table oldtabTb, VARt theValue, VARi lexn, VARi thename);
boolean fnfrmlex(int thename, table oldtabTb, VARt theValue, VARi lexn);
boolean fnfrom(int thename, table oldtabTb, VARt theValue);
boolean fnfroms(int thename, table oldtabTb, VARt theValue);
boolean fninto(int thename, table oldtabTb, tree theValue, VARa newtabtb);
boolean fnsinto(int thename, table oldtabTb, tree theValue, VARa newtabtb);
boolean fnlxinto(int thename, table oldtabTb, tree theValue, int lex, VARa newtabtb);
boolean fnslxinto(int thename, table oldtabTb, tree theValue, int lex, VARa newtabtb);
boolean fnnothere(int thename, table oldtabTb);
boolean fnopen(table oldframe, VARa newFrame);
int pop(void);
void push(int value);
boolean fnnewimage(int size);
boolean fnputat(int locn, int datum);
boolean fngetat(int locn, VARi datum);
boolean fnascii(int item);
void WriteB(int outBool);
void WriteC(CHAR outChar);
void WriteI(int value);
void WriteLn(void);
void WriteS(char* outString);
boolean fnnumber(int value);
void WriteHexDigit(boolean logfi, int theH);
boolean fnhexnum(int value, int ndigs);
boolean fnspam(tree thename);
boolean fnspell(int thename);
void WriteIs(int value);
boolean fnexportree(int thename, tree theTree);
void unmark(tree theTree);
int marker(tree theTree, int theint);
void ExportNode(tree theTree);
void ExpoSet(int theSet);
void ExpoSet(int theSet);
trees GetixTree(trees theTree, int inx);
int Skip2Int(int theFile, boolean NLok);
boolean fnimportree(int thename, VARt theTree);
int ImportString(int theFile, int kin);
void LogReplay (CHAR theCh);
void SaveChecksum(int failed);
void GetChecksum(void);
void ChekChecksum(void);
void Abortit(char* ErrMsg);
void AbortInt(char* ErrMsg, int lino, int aNum);
void MemBomb(void);
void ShowTransList(void);
void ShoTreDeep(tree theTree, int numLevels, int indent);
void ShowK(int value, int refer);
void StatsReport(void);
void validtree(tree thetree);
void ShoBool(int outBool);
void ShoInt(int outInteger);
void ShoStrn(int outInteger);
void ShoTable(table outTable);
void ShoTree(tree outTree);
void ShoSett(settt outTree);
void sWriteB(int outBool);
void sWriteC(CHAR outChar);
void sWriteI(int value);
void sWriteLn(void);
void sWriteS(char* outString);
boolean fntpremolevel(int theLevel);
boolean fntroff(void);
boolean fntron(int theLevel);
boolean fntrlevel(VARi theLevel);
boolean fntrasci(int theChar);
boolean fntrnmbr(int theNum);
boolean fntrhexn(int value, int ndigs);
boolean fntrspel(int theName);
boolean fnttreedmp(tree theTree);
boolean fntsetdmp(settt theTree);
boolean fnttabdmp(table theTree);
boolean fnttoptree(tree theTree, int theChar);
boolean fnttranslsts(tree theTree);
boolean fntcheckpoint(void);
boolean fntreeflag(tree theTree, int theIndx);
boolean fntrunflag(tree theTree, VARi theIndx);
void SpellNodeName(int theNode);
void NewNodeNs(void);
void ModuleInitialization(void);
int TagBegin(void);
void TagEnd(void);
void SyntaxError(void);

void InitializeScanner(); /* actual code for these is in main file.. */
void ScannerFill();
void Getoken();
void do_err();

/**************************** Utility Functions ****************************/

CHAR CHR(int theCh) {return (CHAR) (theCh&255);}
int ORD(CHAR theCh) {return ((int) theCh)&255;}

char* IntToStr(int value) { /* C-specific hack, returns a pointer into its buffer */
  int ix;
  int sgn = 0;
  int oops = 0;
  if (value<0) {sgn++; value = -value;}
  if (value<0) {oops = 1; value = value-1;} /* max negative has no positive */
  /* strcpy(Int2StrTemp, "              0"); */
  Int2StrTemp[15] = '\0';
  Int2StrTemp[0] = ' ';
  for (ix=14; ix>0; ix--) {
    Int2StrTemp[ix] = (char)(value%10+48+oops);
    value = value/10;
    oops = 0;
    if (value==0) break;}
  if (ix>0) if (sgn>0) Int2StrTemp[--ix] = '-';
  return &Int2StrTemp[ix];} /* ~IntToStr; */

char* CatStr(char* theStr, char* aStr) { /* returns a pointer to its buffer */
  strcpy(StringTemp,theStr);
  return strcat(StringTemp,aStr);} /* ~CatStr; */

int Tick_Count(void) { /*  returns ticks since whenever  / *Time_Tix* / */
  return (int) (clock()/(CLOCKS_PER_SEC/60));} /* ~Tick_Count; */

void Sys_Beep(int howLong) {howLong++;} /* ~Sys_Beep; (not implemented in Java) */
void System_Task(void) {} /* ~System_Task; */
int User_Killed(void) {return false;} /* ~User_Killed; */

void Stop_Run(char* ErrMsg)
  {printf("Error: %s \n", ErrMsg); longjmp(CatchThrow,-1);}

void CalledOften(void) { /* used only locally */
  if (User_Killed()) Abortit("User killed");
    else System_Task();} /* ~CalledOften; */

boolean fnmin(int a, int b, VARi c) {if (a>b) *c=b; else *c=a; return true;}
boolean fnmax(int a, int b, VARi c) {if (a<b) *c=b; else *c=a; return true;}
boolean fnbitand(int a, int b, VARi c) {*c=(a&b); return true;}

/**************************** Tree Node Functions ****************************/

tree inttree (int val) {
  tree itree = (tree) malloc(nodeIsize);
  itree->kind = intt;
  itree->intval = val;
  Tree_Count++;
  return itree;}

tree strtree (int val) {
  tree itree = inttree (val);
  itree->kind = snam;
  return itree;}

tree booltree(int truly) {
  tree itree = (tree) malloc(nodeIsize);
  itree->kind = bool;
  itree->intval = truly;
  Tree_Count++;
  return itree;}

settt newsettt(void) {
  settt stree = (settt) inttree(LastUniSet);
  stree->kind = seti;
  return stree;}

tree nutrees(int noden, tree decorn, tree s1, tree s2, tree s3, tree s4, 
      tree s5, tree s6, tree s7, tree s8, int numsubs) {
  tree stree = (tree) malloc(nodeSize);
  stree->kind = wood;
  stree->nodename = noden;
  stree->intval=numsubs;
  stree->subtrees[0] = decorn;
  /* if (numsubs>0) */ stree->subtrees[1] = s1;
  /* if (numsubs>1) */ stree->subtrees[2] = s2;
  /* if (numsubs>2) */ stree->subtrees[3] = s3;
  /* if (numsubs>3) */ stree->subtrees[4] = s4;
  /* if (numsubs>4) */ stree->subtrees[5] = s5;
  /* if (numsubs>5) */ stree->subtrees[6] = s6;
  /* if (numsubs>6) */ stree->subtrees[7] = s7;
  /* if (numsubs>7) */ stree->subtrees[8] = s8;
  Tree_Count++;
  return stree;}

tree newtable(int nam, int lex, table oldtabTb, tree theValue) {
  /* class table extends tree // lex#=intval, value=subtr[0], link=subtr[1] */
  tree atable = (tree) malloc(nodeTsize);
  atable->kind = symt;
  atable->nodename = nam;
  atable->intval = lex;
  atable->subtrees[0] = theValue;
  atable->subtrees[1] = oldtabTb;
  Tree_Count++;
  return atable;}

/**************************** Tree Procedures ****************************/

boolean fnbooltree(int inBool, VARt outtree) {
  *outtree = booltree(inBool);
  return true;} /* ~fnbooltree; */

trees build(int noden, tree decorn, tree s1, tree s2, tree s3, tree s4, 
    tree s5, tree s6, tree s7, tree s8, int numsubs) { /* tree node constructor */
  trees nutree;
  if (TabNo==0) TabNo = noden;
  CalledOften();
  nutree = nutrees(noden,decorn,s1,s2,s3,s4,s5,s6,s7,s8,numsubs);
  if (nutree==null) MemBomb();
  /* WITH nutree^ DO */
    /* nutree^.kind = wood; */
    /* nutree^.nodename = noden; */
    /* nutree^.intval / *nsbtrees* / = numsubs; */
    /* nutree^.decor = decorn; */
    /* nutree^.subtrees[1] = s1; */
    /* nutree^.subtrees[2] = s2; */
    /* nutree^.subtrees[3] = s3; */
    /* nutree^.subtrees[4] = s4; */
    /* nutree^.subtrees[5] = s5; */
    /* nutree^.subtrees[6] = s6; */
    /* nutree^.subtrees[7] = s7; */
    /* nutree^.subtrees[8] = s8; */
    /* END */
  if (Debugging>0) {
    sWriteS("*Build:  ");
    ShoTreDeep(nutree,3,99);
    sWriteLn();}
  return nutree;} /* ~build; */

boolean comparetree(tree thisn, tree thatn) { /* true if same or similar */
  int ix, nx;
  if (thisn == thatn) return true;
  if (thisn == NIL) return false;
  if (thatn == NIL) return false;
  if (thisn->kind != thatn->kind) return false;
  if (thisn->nodename != thatn->nodename) return false;
  if (thisn->intval != thatn->intval) return false;
  if (thisn->kind >= mrkb) return true;
  if (thisn->kind < wood) return true;
  if (thisn->kind == symt) nx = 1;
  else if (thisn->kind == pkat) nx = 5;
    else nx = thisn->intval;
  ix = thisn->kind + mrkb;
  thisn->kind = ix;
  thatn->kind = ix;
  ix = 0;
  while (ix<=nx) {
    if (!comparetree(thisn->subtrees[ix], thatn->subtrees[ix])) nx = -1;
    ix++;}
  ix = thisn->kind - mrkb;
  thisn->kind = ix;
  thatn->kind = ix;
  return (nx>=0);} /* ~comparetree; */

void replacetree(tree dest, tree orig) { /* can only replace wood with wood.. */
  int ix, newn;
  trees treeo, treed;
  if (dest == NIL) Abortit("(Can't replace null tree)");
  if (orig == NIL) Abortit("(Can't replace with null tree)");
  validtree(dest);
  validtree(orig);
  treed = (trees) dest;
  treeo = (trees) orig;
  treed->nodename = treeo->nodename;
  newn = treeo->intval;
  /* if (treed->intval<newn) treed->subtrees = new tree[newn+1]; */
  treed->intval = newn;
  for (ix=0; ix<=newn; ix++) treed->subtrees[ix] = treeo->subtrees[ix];
  } /* ~replacetree; */

int treeint(tree thetree) {
  trees atree;
  if (thetree == NIL) return 0;
  switch (thetree->kind) { /* WITH thetree^ DO */
    case nono:
    case pkat:
      return 0;
    case intt:
    case snam:
    case symt:
    case seti:
    case bool: /* *ORD* /(/ *boolval* /); */
      return thetree->intval;
    case wood:
      atree = (trees) thetree;
      return treeint(atree->subtrees[0]);} /* CASE */ /* END WITH */;
  return 0; /* can`t get here, but the compiler complains anyway */
  } /* ~treeint; */

int treestrn(tree thetree) {return treeint(thetree);}

int treepart(tree treeptr, int opList, int compareto, VARt destTree) {
  /* OpList is a string of one or more one-digit opcodes, read from least */
    /* significant to most significant by mod`ing it by 10.  Note that every */
    /* opList ends with the implied zero that precedes any integer. */
  /* Opcode 0:  set destTree to current tree; return true */
  /* Opcodes 1-7:  set current tree to be the subtree of the current tree */
    /* corresponding to the opcode; if this fails (e.g., current tree is int */
    /* or null), return false and destTree is undefined */
  /* Opcode 8:  set current tree to be the current tree`s decoration; if */
    /* this fails (e.g., the current tree is null), return false and destTree */
    /* is undefined */
  /* Opcode 9:  compare nodename of the current tree to compareTo parameter; */
    /* if not equal, return false with an undefined destTree */
  /* Note: a negative compareTo parameter "equals" the nodename of a null tree. */

  int opCode, howdeep;
  int result;
  trees stree;
  if (Debugging>2) {
    opCode = 10;
    howdeep = 2;
    while (opList>opCode) {
      opCode = opCode*10;
      howdeep = howdeep+1;}
    sWriteS("*treepart: ");
    sWriteI(opList);
    sWriteS(",");
    SpellNodeName(compareto);
    sWriteS(",");
    ShoTreDeep(treeptr,howdeep,0);
    /* sWriteLn; */
    *destTree = NIL;}
  result = true;  /* assume success */
  opCode = opList%10;
  stree = (trees) treeptr;
  while (result && (opCode != 0)) {
    /* validptr(treeptr); */
    switch (opCode) {
      case 8:
        if (stree == null) result = false;
        else if (stree->kind != wood) result = false;
          else stree = (trees) stree->subtrees[0];
        break;
      case 9:
        if (compareto < 0) result = stree == null;
        else if (stree == null) {
          *destTree = NIL;
          result = false;}
        else result = compareto == stree->nodename;
        break;
      default: /* 1..7 */
        if (stree == null) result = false;
        else if (stree->kind != wood) result = false;
          else stree = (trees) stree->subtrees[opCode];} /* end CASE */
    opList = opList/10;
    opCode = opList%10;} /* end WHILE */
  *destTree = stree;
  if (Debugging>2) {
    if (result) sWriteS("="); else sWriteS("#");
    ShoTreDeep(*destTree,2,99);
    sWriteLn();}
  return result;} /* ~treepart; */

tree treetree(tree thetree) {return thetree;} /* ~function treetree */

/* tree inttree(int theint) {return new tree(theint);} /* ~inttree; */

boolean fninttree(int theint, VARt outtree) {
  *outtree = inttree(theint);
  return true;} /* ~fninttree; */

boolean fnstrtree(int theint, VARt outtree) {
  *outtree = strtree(theint);
  return true;} /* ~fnstrtree; */

boolean fnpackattributes(table tab1, settt set2, settt set3, settt set4,
    tree tree5, int int6, VARa outDataTb) {
  table otb = (table) build(0,NIL,tab1,set2,set3,set4,tree5,NIL,NIL,NIL,5);
  otb->kind = pkat;
  otb->intval=int6;
  /* NEW(outData); */
  /* if (outData==null) {MemBomb();} */
  /* outData^.kind = wood; */
  /* outData^.nodename = 0; */
  /* outData^.intval / *nsbtrees* / = 6; */
  /* outData^.decor = null; */
  /* outData^.subtrees[1] = / * tree  * / (tab1); */
  /* outData^.subtrees[2] = / * tree  * / (set2); */
  /* outData^.subtrees[3] = / * tree  * / (set3); */
  /* outData^.subtrees[4] = / * tree  * / (set4); */
  /* outData^.subtrees[5] = tree5; */
  /* outData^.subtrees[6] = inttree(int6); */
  /* outData^.subtrees[7] = 0; */
  /* outData^.subtrees[8] = 0; */
  /* outData^.kind = pkat; */
  /* outDataTb = / * table  * / (outData); */
  if (Debugging>2) {
    sWriteS("*packa:");
    ShoTable(otb);
    sWriteLn();}
  *outDataTb = otb;
  return true;} /* ~fnpackattributes; */

boolean fnsettree(settt theSet, VARt rset) {
  *rset = theSet;
  if (theSet == null) return true;
    else return theSet->kind == seti;} /* ~fnsettree; */

boolean fntabtree(table theTable, VARt theTree) {
  /* This was written for the "old style" symbol table (env) that`s */
  /* really a linked list of trees. */
  *theTree = /* tree */ (theTable);
  return true;} /* ~fntabtree; */

boolean fntreebool(tree intree, VARb outbool) {
  *outbool = intree->intval>0 /* boolval */;
  return true;} /* ~fntreebool; */

boolean fntreeint(tree thetree, VARi outint) {
  *outint = treeint(thetree);
  return true;} /* ~fntreeint; */

boolean fntreestr(tree thetree, VARi outint) {
  *outint = treeint(thetree);
  return true;} /* ~fntreestr; */

boolean fnintstr(int theint, VARi outint) {
  *outint = theint;
  return true;} /* ~fnintstr; */
boolean fnstrint(int theint, VARi outint) {
  *outint = theint;
  return true;} /* ~fnintstr; */

boolean fntreeset(tree thetree, VARs rset) {
  *rset = null;
  if (thetree == NIL) return true;
  if (thetree->kind != seti) return false;
  *rset = (settt) thetree;
  return true;} /* ~fntreeset; */

boolean fntreetab(tree thetree, VARa theTable) {
  /* This was written for the "old style" symbol table (env) that`s */
  /* really a linked list of trees. */
  *theTable = null;
  if (thetree == NIL) return true;
  if (thetree->kind != symt) return false;
  *theTable = (table) thetree;
  return true;} /* ~fntreetab; */

boolean fnunpackatts(table indataTb, VARa tab1, VARs set2, VARs set3,
    VARs set4, VARt tree5, VARi int6) {
  int oldebug = Debugging;
  if (Debugging>2) {
    sWriteS("*unpka:");
    if (Debugging<6) Debugging = 2;
    ShoTable(indataTb);
    sWriteLn();
    Debugging = oldebug;}
  if (indataTb->kind != pkat) return false;
  *int6 = indataTb->intval;
  *tree5 = indataTb->subtrees[5];
  return fntreetab(indataTb->subtrees[1],tab1)
    && fntreeset(indataTb->subtrees[2],set2)
    && fntreeset(indataTb->subtrees[3],set3)
    && fntreeset(indataTb->subtrees[4],set4);} /* ~fnunpackatts; */


/**************************** Low-Level I/O Functions ****************************/

char aFileName[256] = "";

void IO_Startup(void) {} /* ~IO_Startup; */

long int ofsize(FILE* fp) { /* file must be open; returns -1 if err */
  long int len;
  long int here = ftell(fp);
  return 99999999; /* 100M file limit */
  if (here<0) return -1;
  if (fseek(fp,0,SEEK_END)) return -1;
  len = ftell(fp);
  if (fseek(fp,here,SEEK_SET)) return -1;
  return len;} /* ~ofsize */

void Open_File(int theFile, char* prompt) {
  char* finame = prompt; /* not used; this just stops the compiler warning */
  char* fimode = "rb";
  char errmsg[333] = "Can't open ";
  int io = 1; /* 1=in, 0=out */
  switch (theFile) { /* 0=scr,1=in,2=out,3=diag,4=lib */
  case 1: 
    finame = InputFileName;
    break;
  case 2: 
    finame = OutputFileName;
    io = 0;
    fimode = OutputFileType;
    break;
#ifdef LibryFileName
  case 4: 
    finame = LibryFileName;
    break;
#endif
#ifdef FailMsFileName
  case 3: 
    finame = DebugFileName;
    io = 0;
    fimode = OutputFileType;
    break;
  case 5: 
    finame = FailMsFileName;
    break;
  case 6: 
    finame = FailMsFileName;
    io = 0;
    fimode = OutputFileType;
    break;
#endif
  default: return;}
  My_Files[theFile] = fopen(finame,fimode);
  if (My_Files[theFile]==0) {if (theFile<4) Stop_Run(strcat(errmsg,finame));}
  else if (io) {
    File_Peek[theFile] = Null_Char;
    FileOpLen[theFile] = ofsize(My_Files[theFile]);}
  else FileOpLen[theFile] = 0;} /* ~Open_File; */

void Close_File(int theFile) {
  File_Peek[theFile] = EOF_Char;
  if (FileOpLen[theFile]<0) return;
  fclose(My_Files[theFile]);
  FileOpLen[theFile] = -1-FileOpLen[theFile];} /* ~Close_File; */

void WriteOutChar(int theFile, CHAR theChar) {
  if (FileOpLen[theFile]<0) return;
  FileOpLen[theFile] = FileOpLen[theFile]+1;
  if (theFile==0) {
    if (KillingIt==0) if (FileOpLen[0]>64) if (theChar<CHR(33))
      theChar = CRLF;
    if (theChar<CHR(32)) {
      putchar('\n');  /* CRLF */
      /* if (LFCR != EOF_Char) putchar(LFCR); */
      FileOpLen[0] = 0;}
    else putchar(theChar);}
  else if ((theFile&2)==2) {
    if (theChar != LFCR) putc((int) theChar, My_Files[theFile]);
    if (theChar==CRLF) if (LFCR != EOF_Char)
      putc((int) LFCR, My_Files[theFile]);}
  } /* ~WriteOutChar; */

CHAR PeekInChar(int theFile) {
  CHAR ch = File_Peek[theFile];
  while (ch==Null_Char) {
    if (FileOpLen[theFile]>0) {
      FileOpLen[theFile] = FileOpLen[theFile]-1;
      ch = (CHAR) getc(My_Files[theFile]);
      if (feof(My_Files[theFile])) {
        ch = EOF_Char;
        FileOpLen[theFile] = 0;}
      /* else if (ch>'~') ch = Null_Char; */
      else if (ch==EOF_Char || ch==LFCR) ch = Null_Char;}
    else ch = EOF_Char;
    if (Trace_Toggle>Null_Char) if (ch==Trace_Toggle) {
      if (Debugging==0) Debugging = 2;
      else if (Debugging==2) Debugging = 0;
      ch = Null_Char;}}
  File_Peek[theFile] = ch;
  CalledOften();
  return ch;} /* ~PeekInChar; */

void EchoThisInput(CHAR ch) {
  int dump = 0;
  if (ch >= ' ' || ch=='\t') {
    InputLinePosn++;
    InputEchoBuffer[EchoInputSofar++] = ch;
    if (EchoInputSofar>EchoInputBuffer) dump = 1;}
  else dump = 1;
  if (dump==0) return;
  InputLinePosn = 0;
  InputEchoBuffer[EchoInputSofar++] = '\0';
  WriteS("// ");
  WriteS(InputEchoBuffer);
  WriteLn();
  EchoInputSofar = 0;} /* ~EchoThisInput; */

CHAR ReadInChar(int theFile) {
  CHAR ch = PeekInChar(theFile);
  if (ch != EOF_Char) File_Peek[theFile] = Null_Char;
  if (theFile==1) if (ch != EOF_Char) LogReplay(ch);
  if (EchoInputBuffer !=0) if (theFile==sourceFile) EchoThisInput(ch);
  return ch;} /* ~ReadInChar; */

int EndFi(void) {return PeekInChar(sourceFile)==EOF_Char;} /* ~EndFi; */

int EndLn(int theFile) {
  CHAR ch = PeekInChar(theFile);
  return (ch==CRLF) || (ch==EOL_Char) || (ch==EOF_Char);} /* ~EndLn; */

int EndLine(void) {return EndLn(sourceFile);} /* ~EndLine; */

CHAR ReadCh(void) { /* read input.. */
  CHAR ch = ReadInChar(sourceFile);
  SourceHere = SourceHere+1;
  if (ch == '\t') ch = ' ';
  return ch;} /* ~ReadCh; */

int ReadInt(int theFile) {
  CHAR inChar;
  int theInt = 0;
  int negatory = 1;
  int peek = false;
  do {inChar = ReadInChar(theFile);} while (inChar <= ' ');
  if (inChar == '-') {
    inChar = ReadInChar(theFile);
    negatory = -1;}
  while (inChar>='0' && inChar<='9') {
    if (peek) inChar = ReadInChar(theFile);
    theInt = theInt*10+ORD(inChar)-48;
    inChar = PeekInChar(theFile);
    peek = true;}
  return theInt*negatory;} /* ~ReadInt; */

void ReadLin(void) { /* Read through next CR (ASCII 13) */
  CHAR ch;
  SourceHere = (SourceHere+1)|0xFFF;
  while (!EndLn(sourceFile)) ch = ReadInChar(sourceFile);
  ch = ReadInChar(sourceFile);} /* ~ReadLin(); */

void FileClose(int fileNumber) {Close_File(fileNumber);} /* ~FileClose; */

int FileOpen(int mode, char* prompt) { /* File_Ref */
  /* Opens an existing TEXT file or creates a new one and opens it. */
  /* Input mode: */
    /* 1 = open existing ("read") */
    /* 2 = create new & open ("write") */
    /* 3 = create new & open debug file */
    /* 4 = open existing library file */
  Open_File(mode, prompt);
  return mode;} /* ~FileOpen; */

void fileWriteC(int fileRef, CHAR outChar) {
  /* Writes a character to given file, or perhaps the screen */
  WriteOutChar(fileRef,outChar);
  if (fileRef==DbugFile) {
    Line_Len = Line_Len+1;
    if (Line_Len>64) if (outChar<CHR(33)) if (outChar != CRLF) sWriteLn();
    DbugFiLen = DbugFiLen+1;
    if (DbugFiLen>RunawayTrace) {
      DbugFiLen = 0; /* so we don't repeat a runaway */
      CkSumHere = 0;
      sWriteLn();
      Abortit("RunawayTrace");}}
  else if (fileRef==destFile) LogReplay(outChar);} /* ~fileWriteC; */

void fileWriteS(int fileRef, char* outString) {
  /* Writes a string to the given file, or perhaps the screen */
  int ix;
  if (fileRef==destFile) ChekChecksum();
  SavedChar = outString[0];
  for (ix=0; ix<strlen(outString); ix++) 
    fileWriteC(fileRef,outString[ix]);} /* ~fileWriteS; */

void EchoCh(CHAR ch) {
  SourceThar=SourceHere;
  TokenSho[curTokSho] = ch;
  if (curTokSho<127) curTokSho++;} /* ~EchoCh */

void ShoTokSwap(void) {
  TokenSho[curTokSho] = Null_Char;
  if (Debugging>0) {
    sWriteC('\"'); sWriteS(TokenSho); sWriteC('\"');}
  curTokSho = 0;} /* ~ShoTokSwap */

/**************************** Sparse Set Procedures ****************************/

boolean InitSets(void) { /* initialize UniqueSets.. */
  /* UniqueSets = new UniSetRec[TopSetNo]; */
  /* HugeEltHash = new int[256]; */
  /* if (UniqueSets==null) return false; */
  LastUniSet = 0;
  ReplayHere = 0;
  DbugFiLen = 0;
  KillingIt = 0;
  do {
    if (LastUniSet<256) HugeEltHash[LastUniSet] = 0;
    /* UniqueSets[LastUniSet] = new UniSetRec(); */
    UniqueSets[LastUniSet].SetPtr = null;
    UniqueSets[LastUniSet].EltHeds = 0;
    LastUniSet = LastUniSet+1;}
    while (LastUniSet<TopSetNo);
  LastUniSet = 0;
  return true;} /* ~InitSets; */

int myaddset(int elt, int theSet) { /* UniqueSets constructor */
  settt tSet;                                 
  int eSet, hasher;
  /* create or find the set x=theSet+{elt} .. */
  if (theSet==0 || UniqueSets[theSet].ThisElt<elt) {
    /* elt goes in front of theSet; maybe it exists already... */
    hasher = -1;
    if (elt>=TopSetNo) {
      hasher = (elt/256 + elt)&255; /* simple hash code */
      eSet = HugeEltHash[hasher];}
    else eSet = UniqueSets[elt].EltHeds;
    while (eSet != 0) {
      if (UniqueSets[eSet].NextElt==theSet) {
        if (hasher<0) return eSet;
        else if (UniqueSets[eSet].ThisElt==elt) return eSet;}
      eSet = UniqueSets[eSet].HedLink;}
    LastUniSet = LastUniSet+1;
    if (LastUniSet==TopSetNo) Abortit("***Set Oflo***");
    UniqueSets[LastUniSet].ThisElt = elt;
    UniqueSets[LastUniSet].NextElt = theSet;
    if (hasher<0) {
      UniqueSets[LastUniSet].HedLink = UniqueSets[elt].EltHeds;
      UniqueSets[elt].EltHeds = LastUniSet;}
    else {
      UniqueSets[LastUniSet].HedLink = HugeEltHash[hasher];
      HugeEltHash[hasher] = LastUniSet;}
    tSet = newsettt(); /* inttree(LastUniSet) */
    /* tSet->kind = seti; */
    UniqueSets[LastUniSet].SetPtr = tSet;
    return LastUniSet;}
  else if (UniqueSets[theSet].ThisElt==elt) return theSet;
  /* else if myinset(elt, UniqueSets[theSet].SetPtr) return theSet; */
  else { /* must insert it somewhere inside, and build a new list on top... */
    eSet = myaddset(elt,UniqueSets[theSet].NextElt);
    return myaddset(UniqueSets[theSet].ThisElt,eSet);}} /* ~myaddset; */

int myinset(int elt, settt theSet) { /* UniqueSets examiner */
  int tset;                                  
  if (theSet != null) if (theSet->kind==seti) {
    tset = theSet->intval;
    if (UniqueSets[tset].ThisElt==elt) return true;
    if (UniqueSets[tset].ThisElt>elt) 
      return myinset(elt,UniqueSets[UniqueSets[tset].NextElt].SetPtr);}
  return false;} /* ~myinset; */

boolean fnaddset(int elt, settt theSet, VARs rset) {
  if (myinset(elt, theSet)) *rset = theSet;
  else
    if (Debugging>2) {
      sWriteS("*addset:");
      ShoInt(elt);
      /* ShoSett(*rset); */
      if (Debugging>7) ShoTreDeep(theSet,9,99); else ShoTreDeep(theSet,2,99);}
    if (theSet==null) *rset = UniqueSets[myaddset(elt,0)].SetPtr;
    else if (theSet->kind==seti) 
      *rset = UniqueSets[myaddset(elt,theSet->intval)].SetPtr;
    else return false;
  return true;} /* ~fnaddset; */

boolean fndifference(settt aset, settt bset, VARs difset) {
  settt xset, yset;
  int elt, tset, lvls;
  int ok = true;
  *difset = null;
  if ((aset == bset) || (aset == null)) ;
  else if (bset == null) *difset = aset;
  else if (aset->kind != seti) return false;
  else if (bset->kind != seti) return false;
  else {
    elt = UniqueSets[aset->intval].ThisElt;
    if (elt>UniqueSets[bset->intval].ThisElt) {
      *difset = UniqueSets[UniqueSets[aset->intval].NextElt].SetPtr;
      ok = fndifference(*difset,bset,&xset);
      tset = 0;
      if (ok) if (xset != null) tset = xset->intval;
      *difset = UniqueSets[myaddset(elt,tset)].SetPtr;}
    else if (elt==UniqueSets[bset->intval].ThisElt) {
      xset = UniqueSets[UniqueSets[aset->intval].NextElt].SetPtr;
      yset = UniqueSets[UniqueSets[bset->intval].NextElt].SetPtr;
      ok = fndifference(xset,yset,difset);}
    else {
      xset = UniqueSets[UniqueSets[bset->intval].NextElt].SetPtr;
      ok = fndifference(aset,xset,difset);}}
  if (Debugging>8) {
    sWriteS("*diff: ");
    if (Debugging>9) lvls = 9; else lvls = 1;
    ShoTreDeep(aset,lvls,0);
    sWriteS(",");
    ShoTreDeep(bset,lvls,0);
    sWriteS("=");
    ShoTreDeep(*difset,lvls,0);
    sWriteLn();}
  return ok;} /* ~fndifference; */

boolean fninset(int elt, settt theSet) {
  int retVal = myinset(elt, theSet);
  if (Debugging>3) {
    sWriteS("inset:");
    ShoInt(elt);
    if (retVal) sWriteC('T');
      else sWriteC('F');}
  return retVal;} /* ~fninset; */

boolean fnintersect(settt aset, settt bset, VARs rset) {
  settt xset, yset;
  int elt, tset;
  int ok = true;
  *rset = null;
  if (bset == null || aset == null) *rset = null;
  else if (aset == bset) *rset = aset;
  else if (aset->kind != seti) return false;
  else if (bset->kind != seti) return false;
  else {
    elt = UniqueSets[aset->intval].ThisElt;
    if (elt<UniqueSets[bset->intval].ThisElt) {
      xset = UniqueSets[UniqueSets[bset->intval].NextElt].SetPtr;
      ok = fnintersect(aset,xset,rset);}
    else if (elt>UniqueSets[bset->intval].ThisElt) {
      xset = UniqueSets[UniqueSets[aset->intval].NextElt].SetPtr;
      ok = fnintersect(xset,bset,rset);}
    else {
      *rset = UniqueSets[UniqueSets[aset->intval].NextElt].SetPtr;
      yset = UniqueSets[UniqueSets[bset->intval].NextElt].SetPtr;
      ok = fnintersect(*rset,yset,&xset);
      tset = 0;
      if (ok) if (xset != null) tset = xset->intval;
      *rset = UniqueSets[myaddset(elt,tset)].SetPtr;}}
  return ok;} /* ~fnintersect; */

boolean fnnewelt(settt theSet, VARi nuelt) {
  int index = 1;
  int dbg = Debugging;
  Debugging = 0;
  while (myinset(index, theSet)) index = index + 1;
  Debugging = dbg;
  *nuelt = index;
  return true;} /* ~fnnewelt; */

boolean fnnotinset(int elt, settt theSet) {
  int retVal = ! myinset(elt, theSet);
  if (Debugging>0) {
    sWriteS("*notinset:");
    sWriteI(elt);
    if (retVal) sWriteC('T');
    else {
      sWriteC('F');
      if (Debugging>1) ShoTreDeep(theSet, 999 ,0);}}
  return retVal;} /* ~fnnotinset; */

boolean fnunion(settt aset, settt bset, VARs rset) {
  settt xset, yset;
  int elt, tset;
  int ok = true;
  *rset = null;
  if ((bset == null) || (aset == bset)) *rset = aset;
  else if (aset == null) *rset = bset;
  else if (aset->kind != seti) return false;
  else if (bset->kind != seti) return false;
  else {
    elt = UniqueSets[aset->intval].ThisElt;
    tset = UniqueSets[bset->intval].ThisElt;
    if (elt<tset) {
      elt = tset;
      yset = UniqueSets[UniqueSets[bset->intval].NextElt].SetPtr;
      ok = fnunion(aset,yset,&xset);}
    else if (elt>tset) {
      yset = UniqueSets[UniqueSets[aset->intval].NextElt].SetPtr;
      ok = fnunion(yset,bset,&xset);}
    else {
      *rset = UniqueSets[UniqueSets[aset->intval].NextElt].SetPtr;
      yset = UniqueSets[UniqueSets[bset->intval].NextElt].SetPtr;
      ok = fnunion(*rset,yset,&xset);}
    tset = 0;
    if (ok) if (xset != null) tset = xset->intval;
    *rset = UniqueSets[myaddset(elt,tset)].SetPtr;}
  return ok;} /* ~fnunion; */


/**************************** FSM-Building Utilities ****************************/

boolean fnfixadetran(int atthistran, int atfrosty, int attosty, 
    int charsClass, int atactor) {
  if (Debugging>2) {
    sWriteS("*fixadetran (");
    ShoInt(atthistran);
    sWriteS(")");
    ShoInt(atfrosty);
    ShoInt(attosty);
    ShoInt(charsClass);
    ShoInt(atactor);
    sWriteLn();}
  else sWriteS("x");
  if (atthistran < 0) 
    Abortit(CatStr(" fnfixadetran error--no index < 0 allowed...illegal value: ",
      IntToStr(atthistran)));
  /* WITH DetrnPtr^[atthistran] DO */
    DeTrnList[atthistran].fromst = atfrosty;
    DeTrnList[atthistran].tost = attosty;
    DeTrnList[atthistran].onchar = charsClass;
    DeTrnList[atthistran].actset = atactor;
    /* END; */
  if (atthistran == 1 && atfrosty + attosty < 0) 
    Abortit(" Error in fnfixadetran");
  /* showDetransList; */
  return true;} /* ~fnfixadetran; */

boolean fnfixatran(int atthistran, int atfrosty,
    int attosty, settt ataset, int atactor) {
  if (Debugging>2) {
    sWriteS("*fixatran");
    ShoInt(atthistran);
    ShoInt(atfrosty);
    ShoInt(attosty);
    ShoInt(atactor);
    ShoSett(ataset);}
  /* WITH TransPtr^[atthistran] DO */
    TransList[atthistran].fromst = atfrosty;
    TransList[atthistran].tost = attosty;
    TransList[atthistran].onset = ataset;
    TransList[atthistran].actset = atactor;
    /* END; */
  if (atthistran == 1 && atfrosty + attosty < 0)
     Abortit("Error in ntfixatran");
  /* ShowTransList; */
  return true;} /* ~fnfixatran; */

boolean fngetadetran(int inTran, VARi outTran, VARi atfrosty, 
    VARi attosty, VARi ataset, VARi atactor, VARi nextTran) {
  /*  Return - If nontrivial transition is found, returns true and all */
  /*    outputs are defined.  If nontrivial transition is not found, */
  /*    returns false and all outputs are undefined. */
  /*  inTran - "first guess" at the transition to use.  If -1, then start */
  /*    at the end of the list. */
  /*  outTran - actual transition used.  This will be the first nontrivial */
  /*    transition found, with the hunt beginning with inTran. */
  int oTran = inTran;
  int result = false;
  CalledOften();
  if (inTran < 0) oTran = detransnmber;
  while (oTran>0 && DeTrnList[oTran].fromst<0) oTran = oTran - 1;
  if (oTran != 0) {
    /* WITH DetrnPtr^[oTran] DO */
      *atfrosty = DeTrnList[oTran].fromst;
      *attosty = DeTrnList[oTran].tost;
      *atactor = DeTrnList[oTran].actset;
      *ataset = DeTrnList[oTran].onchar;
      /* END; */
    *nextTran = oTran - 1;
    result = true;}
  if (Debugging>1) {
    sWriteS("*getadetran(");
    sWriteI(oTran);
    sWriteS(")");
    ShoInt(*atfrosty);
    ShoInt(*attosty);
    ShoInt(*ataset);
    ShoInt(*atactor);
    ShoInt(*nextTran);
    if (result) sWriteS(" T");
      else sWriteS(" F");
    sWriteLn();}
  *outTran = oTran;
  return result;} /* ~fngetadetran; */

boolean fngetatran(int inTran, VARi outTran, VARi atfrosty, 
    VARi attosty, VARs ataset, VARi atactor, VARi atnexttr) {
  int oTran = inTran;
  int retVal = false;
  CalledOften();
  if (inTran < 0) oTran = transnumber;
  while (oTran>0 && TransList[oTran].fromst<0) oTran = oTran - 1;
  if (oTran != 0) {
    /* WITH TransPtr[oTran] DO */
      *atfrosty = TransList[oTran].fromst;
      *attosty = TransList[oTran].tost;
      *atactor = TransList[oTran].actset;
      *ataset = TransList[oTran].onset;
      retVal = true;
      /* END; */
    *atnexttr = oTran - 1;}
  if (Debugging>1) {
    sWriteS("*getatran(");
    sWriteI(oTran);
    sWriteS(")");
    ShoInt(*atfrosty);
    ShoInt(*attosty);
    ShoSett(*ataset);
    ShoInt(*atactor);
    ShoInt(*atnexttr);
    if (retVal) sWriteS(" T");
      else sWriteS(" F");
    sWriteLn();}
  *outTran = oTran;
  return retVal;} /* ~fngetatran; */

boolean fninitScannerStuff(void) {
  /* TransPtr = new transItem[TransMax]; // create empty transition list */
  TransList = (transItem*) malloc(sizeof(transItem)*TransMax);
  transnumber = 0;
  /* DetrnPtr = new transItem[DetrnMax]; // create empty deterministic trans list */
  DeTrnList = (transItem*) malloc(sizeof(transItem)*DetrnMax);
  detransnmber = 0;
  return true;} /* ~fninitScannerStuff; */

boolean fnnotherdetran(int atfrosty, int attosty, int ataset, int atactor) {
  detransnmber = detransnmber + 1;
  if (Debugging>2) {
    sWriteS("*notherdetran(");
    sWriteI(detransnmber);
    sWriteC(')');
    ShoInt(atfrosty);
    ShoInt(attosty);
    ShoInt(ataset);
    ShoInt(atactor);}
  else sWriteS("m ");
  if (detransnmber >= DetrnMax) 
    Abortit(" Deterministic transition table is too small");
  /* WITH DetrnPtr^[detransnmber] DO */
     DeTrnList[detransnmber].onset = null;
     DeTrnList[detransnmber].fromst = atfrosty;
     DeTrnList[detransnmber].tost = attosty;
     DeTrnList[detransnmber].actset = atactor;
     DeTrnList[detransnmber].onchar = ataset;
    /* END WITH */;
  /* DetrnPtr[detransnmber] = new transItem(null,ataset,atfrosty,attosty,atactor); */
  /* showDetransList; */
  return true;} /* ~fnnotherdetran; */

boolean fnnothertran(int atfrosty, int attosty, settt ataset, int action) {
  transnumber = transnumber + 1;
  if (Debugging>2) {
    sWriteS("*nothertran (");
    ShoInt(transnumber);
    sWriteC(')');
    ShoInt(atfrosty);
    ShoInt(attosty);
    ShoInt(action);
    ShoSett(ataset);}
  else sWriteS("n ");
  if (transnumber >= TransMax) Abortit(" Transition table is too small");
  /* WITH TransPtr^[transnumber] DO */
     TransList[transnumber].fromst = atfrosty;
     TransList[transnumber].tost = attosty;
     TransList[transnumber].actset = action;  /* not really a set at all! */
     TransList[transnumber].onset = ataset;
     TransList[transnumber].onchar = 0;
    /* END WITH */;
  /* TransPtr[transnumber] = new transItem(ataset,0,atfrosty,attosty,action); */
  /* ShowTransList; */
  return true;} /* ~fnnothertran; */


/**************************** State Machine Procs ****************************/

void NewStateTable(int nitems) {
  /* int ix; */
  CHAR ch;
  topState = nitems;
  /* StateTable = new StateTrans[nitems]; // [MuchTooBig]; */
  /* for(ix=0; ix<nitems; ++ix) StateTable[ix] = new StateTrans(); */
  for(ch=CHR(0); ch<CHR(255); ++ch) charClass[ch] = 0;
  for(ch=CHR(8); ch<=CHR(12); ++ch) charClass[ch] = 2;
  charClass[CHR(13)] = 1; 
  charClass[CHR(10)] = 1;} /* ~NewStateTable */

void ScanStateItem(int st, int ch, int ac, int ns) {
  int ix = st*nCharSets+ch;
  if (ix>=topState) Abortit("Egad! Too few states");
  StateTable[ix].action = ac;
  StateTable[ix].nextState = ns;} /* ~ScanStateItem */

boolean MatchToken(int tkno) {
  if (tkno==NexToken) {
    Getoken();
    Taken++;
    return TRUE;}
  if (ReservedWord(RecentStrx)==tkno) {
    Getoken();
    Taken++;
    return TRUE;}
  return FALSE;} /* ~MatchToken */

int ReservedWord(int strix) {
  int here = ResWordPtr[0];
  if (NexToken>here) return 0;
  if (ResWordPtr[NexToken]==0) return 0;
  if (ResWordPtr[here-1]<strix) return 0;
  if (ResWordPtr[here-2]>strix) return 0;
  return ResWordPtr[ResWordPtr[here]+strix];} /* ~ReservedWord */

void AddResWord(int tokno, int wordty, char* theWord) {
  int here, strix;
  boolean ok = fninitbl();
  for (here=strlen(theWord); here>0; here--) ok = fnaddtbl((int)*theWord++);
  ok = fnstrindex(0,wordty,&strix);
  here = ResWordPtr[0];
  if (here==0) return;
  if (ResWordPtr[here-1]<strix) ResWordPtr[here-1] = strix;
  if (ResWordPtr[here-2]==0) {
    ResWordPtr[here] = here+1-strix;
    ResWordPtr[here-2] = strix;}
  ResWordPtr[ResWordPtr[here]+strix] = tokno;} /* ~AddResWord */

boolean fnStartScanner(boolean closeIt) {
  if (Debugging>0) sWriteS("*StartScanner*");
  if (closeIt) {      /* called by InitializeScanner and maybe user */
    FileClose(1);
    FileOpen(1,"");}
  ScanNextCh=Null_Char;
  LastCharacter=Null_Char;
  SourceHere=0;
  SourceThar=0;
  CurrentState=0;
  Getoken();
  return true;} /* ~fnStartScanner */

boolean fnsrctoklinch(VARi linch) {
  if (Debugging>1) {
    sWriteS("#");
    sWriteI(SourceThar>>12);
    sWriteC('.');
    sWriteI(SourceThar&0xFFF);
    sWriteC(' ');}
  *linch = SourceThar;
  return true;} /* ~fnsrctoklinch */

/**************************** String Table Procedures ****************************/

boolean InitStrings(void) {
  /* CharTable = new char[CharMax]; */
  /* if (CharTable==null) return false; */
  CharTable[0] = CHR(0);
  EndChars = 1;
  /* StringTable = new int[StringMax]; */
  if (StringTable==null) return false;
  EndStrings = 0;
  while (EndStrings<FirstString) {
    /* StringTable^.StrAry[EndStrings].sLink = 0; */
    /* StringTable^.StrAry[EndStrings].StrOffs = 0; */
    /* StringTable^.StrAry[EndStrings].StrType = 0; */
    StringTable[EndStrings] = 0;
    EndStrings = EndStrings+1;}
  StringTable[EndStrings+StrOffs] = EndChars;
  /* StringTable^.StrAry[EndStrings].StrOffs = EndChars; */
  WorkHash = -1; /* so we can see if it`s active or not */
  return true;} /* ~InitStrings; */

int myLength(int item) { /* returns length of string # item */
  if (item<FirstString) return 0;
  if (item>=EndStrings) return 0;
  if (item%StrRef !=0) return 0;
  return StringTable[item+3+StrOffs]-1-StringTable[item+StrOffs];
  } /* ~myLength; */

void ShowStrings(void) { /* not used locally */
  int ix;
  for (ix = FirstString; ix<EndStrings; ix = ix+StrRef) {
    sWriteLn();
    sWriteI(ix);
    sWriteS(":");
    sWriteI(StringTable[ix+StrType]);
    sWriteS(" ");
    SpellIt(ix,true);}
  sWriteLn();} /* ~ShowStrings(); */

void SpellIt(int thename, int showy) {
  int index, len;
  CHAR tempch;
  char* aline = &CharTable[StringTable[thename+StrOffs]];
  int quoter = false;
  CalledOften();
  len = myLength(thename);
  for (index = 0; index<len; index++) {
    tempch = CharTable[StringTable[thename+StrOffs]+index];
    if (showy) {
      if (tempch<'0') quoter = true;
      else if (tempch>'z') quoter = true;
      else if (tempch>'Z') {if (tempch<'a') if (tempch != '_') quoter = true;}
      else if (tempch>'9') if (tempch<'A') quoter = true;}}
  if (showy) {
    if (quoter) sWriteS("`");
    sWriteS(aline);
    if (quoter) sWriteS("`");}
  else WriteS(aline);} /* ~SpellIt; */

boolean fnaddtbl(int theChar) {
  /* if (theChar != ORD(CRLF)) if (theChar != ORD(EOL_Char)) */
  if (theChar != ORD(EOF_Char)) if (EndChars < CharMax) {
    CharTable[EndChars] = CHR(theChar);
    WorkHash = (WorkHash*2 + WorkHash/HashTop + theChar)%HashTop;
    EndChars = EndChars+1;
    return true;}
  return false; /* string table overflow or bogus char */
 } /* ~fnaddtbl; */

boolean fncharfrom(int index, int item, VARi thechar) {
  /* index:  0-based offset from beginning of string of the character we */
  /*   want (i.e., 0 gives first char, 1 gives second .. .) */
  if (myLength(item)>index) {
    *thechar = ORD(CharTable[StringTable[item+StrOffs]+index]);
    return true;}
  return false;} /* ~fncharfrom; */

boolean fncharval(VARi charNum) {
  *charNum = ORD(LastCharacter);
  return true;} /* ~fncharval; */

boolean fninitbl(void) {
  /* prepares the string table to accept characters of a new string */
  int retVal = false; /* possible string table overflow */
  if (EndChars < CharMax) if (EndStrings < StringMax) {
    StringTable[EndStrings+sLink] = 0;
    StringTable[EndStrings+StrType] = 0;
    EndChars = StringTable[EndStrings+StrOffs];
    WorkHash = 0;
    retVal = true;}
  if (Debugging>2) {
    sWriteS("*initbl ");
    sWriteB(retVal);
    sWriteS(" (EndChars=");
    sWriteI(EndChars);
    sWriteS(", EndStrs=");
    sWriteI(EndStrings);
    sWriteC(')');
    sWriteLn();}
  return retVal;} /* ~fninitbl; */

boolean fnlength(int item, VARi outLen) {
  *outLen = myLength(item);
  if (Debugging>0) {
    sWriteS("*length:");
    ShoInt(item);
    ShoInt(*outLen);
    sWriteC(' ');}
  return true;} /* ~fnlength; */

boolean fnstleng(int item, VARi outLen) {return fnlength(item,outLen);}

boolean fnstrindex(int equivclass, int stringType, VARi stringIndex) {
  /* Adds current string to the string table and returns its index */
  int addChar;  /* start of the string we're adding */
  int lookat; /* StringRange; // next string index in this hash bucket */
  int insideChar, len; /* its start char, length */
  int equiv = (equivclass-1)<<8;
  CharTable[EndChars] = CHR(0);
  StringTable[EndStrings+StrType] = stringType;
  addChar = StringTable[EndStrings+StrOffs];
  lookat = StrHashTbl[WorkHash];
  while (lookat>0) {
    len = myLength(lookat);
    if (StringTable[lookat+StrType] == stringType)
        if (EndChars-addChar==len) {
      insideChar = StringTable[lookat+StrOffs];
      if (equivclass>0) while (len>0) {
        len = len-1; 
        if (EquivClassPtr[equiv+(int)CharTable[addChar+len]]
          != EquivClassPtr[equiv+(int)CharTable[insideChar+len]]) len = -1;}
      else while (len>0) {
        len = len-1;
        if (CharTable[addChar+len] != CharTable[insideChar+len])
          len = -1;}
      if (len==0) { /* found it, back out of this temp.. */
        WorkHash = -1;
        EndChars = StringTable[EndStrings+StrOffs];
        RecentStrx = lookat;
        *stringIndex = RecentStrx;
        if (Debugging>0) {
          sWriteC('`');
          sWriteI(lookat);}
        return true;}}
    lookat = StringTable[lookat+sLink];}
  StringTable[EndStrings+sLink] = StrHashTbl[WorkHash];
  StrHashTbl[WorkHash] = EndStrings;
  RecentStrx = EndStrings;
  *stringIndex = RecentStrx;
  WorkHash = -1;
  if (Debugging>0) {
    sWriteC('`');
    sWriteI(EndStrings);}
  EndStrings = EndStrings+StrRef;
  EndChars = EndChars+1;
  StringTable[EndStrings+StrOffs] = EndChars;
  len = EndStrings+1;
  if ((len&-len) !=len) return true; /* skip over power of 2 -1 .. */
    /* rationale: 2**n-1 is a common constant; importree renumbers n*3 */
  StringTable[EndStrings+sLink] = 0;
  StringTable[EndStrings+StrType] = 0;
  EndStrings = EndStrings+StrRef;
  StringTable[EndStrings+StrOffs] = EndChars;
  return true;} /* ~fnstrindex; */

boolean fnuplino(int plus, VARi newly) {
  InLineNum = InLineNum+plus;
  *newly = InLineNum;
  return true;} /* ~fnuplino; */

/**************************** Symbol Table Procedures ****************************/

boolean fnixfrom(int indx, table oldtabTb, VARt theValue, VARi lexn, VARi thename) {
  int ix = indx;
  int retVal = false;  /* assume we won`t find it */
  table currTree = /* tab2tree */ (oldtabTb);
  *theValue = NIL;
  *lexn = 0;
  /* validptr(currTree); */
  if (Debugging>1) {
    sWriteS("*ixFrom:");
    ShoInt(ix);
    sWriteS(",");
    if (Debugging>2) if (currTree != null) ShoTable(oldtabTb);}
  while (currTree != null) { /* WITH currTree^ DO */
    if (ix == 0) {
      retVal = true;
      *theValue = currTree->subtrees[0];
      *lexn = currTree->intval;
      *thename = currTree->nodename;
      currTree = null;}
    else {
      currTree = (table) currTree->subtrees[1];
      ix--;}}
  if (Debugging>1) {
    sWriteS(" =");
    if (retVal) {
      ShoInt(*lexn);
      sWriteC(':');
      ix = *thename;
      ShoInt(ix);
      if (ix>=FirstString) if (ix<EndStrings) if (ix%StrRef ==0) {
        sWriteS(" '");
        SpellIt(ix,true);
        sWriteS("'");}
      if (Debugging>2) ShoTreDeep(*theValue, 9, 0);
        else ShoTreDeep(*theValue, 1, 99);}
    else sWriteS("F");
    sWriteLn();}
  return retVal;} /* ~fnixfrom; */

boolean fnfrmlex(int thename, table oldtabTb, VARt theValue, VARi lexn) {
  int tx, retVal = false;  /* assume we won`t find it */
  table currTree = /* tab2tree */ (oldtabTb);
  *theValue = NIL;
  *lexn = 0;
  /* validptr(currTree); */
  if (Debugging>1) {
    sWriteS("*From:");
    ShoInt(thename);
    if (thename>=FirstString) if (thename<EndStrings) if (thename%StrRef ==0) {
      sWriteS(" '");
      SpellIt(thename,true);
      sWriteS("'");}
    sWriteS(",");
    if (Debugging>2) if (currTree != null) ShoTable(oldtabTb);}
  while (currTree != null) { /* WITH currTree^ DO */
    tx = currTree->nodename;
    if (Debugging>4) {
      sWriteLn();
      sWriteS(" ");
      ShoInt(currTree->intval);
      sWriteC(',');
      ShoInt(tx);
      if (tx>=FirstString) if (tx<EndStrings) if (tx%StrRef ==0) {
        sWriteS(" '");
        SpellIt(tx,true);
        sWriteS("'");}}
    if (tx == thename) {
      retVal = true;
      *theValue = currTree->subtrees[0];
      *lexn = currTree->intval;
      currTree = null;}
    else currTree = (table) currTree->subtrees[1];}
  if (Debugging>1) {
    sWriteS(" =");
    if (retVal) {
      ShoInt(*lexn);
      sWriteS(", ");
      if (Debugging>2) ShoTreDeep(*theValue, 9, 0);
        else ShoTreDeep(*theValue, 1, 99);}
    else sWriteS("F");
    sWriteLn();}
  return retVal;} /* ~fnfrmlex; */

boolean fnfrom(int thename, table oldtabTb, VARt theValue) {
  int ix=0;
  return fnfrmlex(thename,oldtabTb,theValue, &ix);} /* ~fnfrom; */

boolean fnfroms(int thename, table oldtabTb, VARt theValue) {
  int ix=0;
  return fnfrmlex(thename,oldtabTb,theValue, &ix);} /* ~fnfroms; */

boolean fninto(int thename, table oldtabTb, tree theValue, VARa newtabtb) {
  int lex = treeint(oldtabTb);
  return fnlxinto(thename,oldtabTb,theValue,lex, newtabtb);} /* ~fninto; */

boolean fnsinto(int thename, table oldtabTb, tree theValue, VARa newtabtb) {
  int lex = treeint(oldtabTb);
  return fnlxinto(thename,oldtabTb,theValue,lex, newtabtb);} /* ~fnsinto; */

boolean fnslxinto(int thename, table oldtabTb, tree theValue, int lex, VARa newtabtb)
  {return fnlxinto(thename,oldtabTb,theValue,lex,newtabtb);}

boolean fnlxinto(int thename, table oldtabTb, tree theValue, int lex, VARa newtabtb) {
    /* dangerous to use lxinto directly, it corrupts the symbol table */
  table newtab;          /* not most efficient, but oh well... */
  int dupe = false; /* assume we won`t find it already there at this level */
  table temptr = oldtabTb;
  if (thename==0) lex = lex+1;
  else while (temptr != null) { /* WITH temptr^ DO */
    if (lex != temptr->intval) temptr = null;
    else if (thename != temptr->nodename) temptr = (table) temptr->subtrees[1];
    else {
      dupe = true;
      temptr = null;}}
  if (Debugging>0) {
    sWriteS("*into:");
    ShoInt(thename);
    if (thename>=FirstString) if (thename<EndStrings) if (thename%StrRef ==0) {
      sWriteS(" '");
      SpellIt(thename,true);
      sWriteS("'");}
    sWriteC(',');
    ShoInt(lex);
    if (dupe) sWriteS(",=F");
    sWriteS(", ");
    if (Debugging>2) ShoTreDeep(theValue, 9, 0);
      else ShoTreDeep(theValue, 1, 99);
    sWriteLn();}
  *newtabtb = null;
  if (dupe) return false;
  newtab = newtable(thename,lex,oldtabTb,theValue);
  if (newtab==null) MemBomb();
  if (Debugging>2) ShoTable(newtab);
  *newtabtb = newtab;
  return true;} /* ~fnlxinto; */

boolean fnnothere(int thename, table oldtabTb) {
  tree theValue = null;
  int lexn = 0;
  if (Debugging>0) sWriteS("*nothere:");
  if (fnfrmlex(thename,oldtabTb,&theValue,&lexn)) 
    return (treeint(oldtabTb) != lexn);
  return true;} /* ~fnnothere; */

boolean fnopen(table oldframe, VARa newFrame) {
  int retVal = fninto(0,oldframe,NIL,newFrame);
  ShoLexLvl = true;
  if (Debugging>0) {
    sWriteS("*open+");
    sWriteI(treeint(oldframe)+1);
    sWriteS("*");}
  return retVal;} /* ~fnopen; */


/**************************** Parser Stack Procs ****************************/

int pop(void) {
  if (stackTop == 0) Abortit(" TagStack underflow error");
  stackTop = stackTop - 1;
  if (Debugging>7) {
    sWriteS("^");
    sWriteI(theStack[stackTop]);
    sWriteS(". ");}
    return theStack[stackTop];} /* ~pop; */

void push(int value) {
  int ix;
  theStack[stackTop] = value;
  if (stackTop > StackMax - (StackMax>>3)) {
    if (Debugging==0) Debugging = PreMortemLevel;
    sWriteS("!");
    sWriteI(stackTop);
    sWriteS("=");
    sWriteI(value);
    sWriteS(". ");
    if (stackTop == StackMax) Debugging = 8;}
  if (Debugging>7) {
    ShoInt(stackTop);
    sWriteS("#");
    for (ix = 0; ix<=stackTop; ix++) ShoInt(theStack[ix]);
    sWriteS(".");
    sWriteLn();}
  if (stackTop == StackMax) Abortit(" TagStack overflow error");
  stackTop = stackTop + 1;
  if (stackTop>HiStack) HiStack = stackTop;} /* ~push; */


/**************************** Memory Image Functions ****************************/

boolean fnnewimage(int size) {
  int* here;
  if (MemImage != null) free(MemImage);
  MemImage = null;
  ImageSize = 0;
  if (size <= 0) return true;
  MemImage = (int*)malloc(size*sizeof(int));
  if (MemImage == null) return false;
  here = MemImage;
  ImageSize = size;
  for (; size>0; size--) *here++ = 0;
  return true;} /* ~fnnewimage */

boolean fnputat(int locn, int datum) {
  if (locn<0 || locn>=ImageSize) return false;
  MemImage[locn] = datum;
  return true;} /* ~fnputat */

boolean fngetat(int locn, VARi datum) {
  if (locn<0 || locn>=ImageSize) return false;
  *datum = MemImage[locn];
  return true;} /* ~fngetat */

/**************************** TAG File I/O Utilities ****************************/

boolean fnascii(int item) {
  if (item == 7) Sys_Beep(30);
  else if (item == 0) {
    CalledOften();
    WriteLn();}
  else WriteC(CHR(item));
  return true;} /* ~fnascii; */

void WriteB(int outBool) {
  /* Write a boolean to code, possibly echoing to debug */
  if (outBool) {
    fileWriteS(destFile, "true");
    if (Debugging>0) fileWriteS(DbugFile, "`true`");}
  else {
    fileWriteS(destFile, "false");
    if (Debugging>0) fileWriteS(DbugFile, "`false`");}} /* ~WriteB; */

void WriteC(CHAR outChar) {
  /* Write a character to code, possibly echoing to debug */
  fileWriteC(destFile, outChar);
  if (Debugging>0) fileWriteC(DbugFile, outChar);} /* ~WriteC; */

void WriteLn(void) {
  /* Write a CR (ASCII 13) to code, possibly echoing to debug */
  if (Debugging>0) sWriteLn();
  if (ErrLnCapIx >= 0) {
    ErrLnCapIx = 0;
    *ErrLnCapture = '\0';}
  fileWriteC(destFile, CRLF);} /* ~WriteLn(); */

void WriteS(char* outString) {
  /* Write a string to code, possibly echoing to debug */
  int lx = strlen(outString);
  if (ErrLnCapIx >= 0) if (ErrLnCapIx+lx<ErrLnCaptLen) {
    strcat(ErrLnCapture,outString);
    ErrLnCapIx = ErrLnCapIx+lx;}
  fileWriteS(destFile, outString);
  if (Debugging>0) {
    fileWriteS(DbugFile, "`");
    fileWriteS(DbugFile, outString);
    fileWriteS(DbugFile, "`");}} /* ~WriteS; */

void WriteI(int value) {
  /* Write an integer to code, possibly echoing to debug */
  WriteS(IntToStr(value));} /* ~WriteI; */

boolean fnnumber(int value) {
  WriteI(value);
  return true;} /* ~fnnumber; */

void WriteHexDigit(boolean logfi, int theH) {
  CHAR theCh = (CHAR)((theH&15)+48);
  if (theCh>'9') theCh = (CHAR)((theH&15)+55);
  if (logfi) sWriteC(theCh);
    else WriteC(theCh);} /* ~WriteHexDigit; */

boolean fnhexnum(int value, int ndigs) {
  while (ndigs-- >0) WriteHexDigit(false, value >> (ndigs*4));
  return true;} /* ~fnhexnum; */

boolean fnspam(tree thename) {
  if (thename==NIL) return false;
  return fnspell(treeint(thename));} /* ~fnspam; */

boolean fnspell(int thename) {
  if (thename<FirstString) return false;
  if (thename>=EndStrings) return false;
  if (thename%StrRef !=0) return false;
  /* if (((thename+1)&(-1-thename)) == thename+1) return false; */
  SpellIt(thename,false);
  return true;} /* ~fnspell; */

#ifdef LibryFileName
void WriteIs(int value) {WriteI(value); WriteC(' ');}

boolean fnexportree(int thename, tree theTree) {
  int ix;
  if (Debugging>0) sWriteS("+exportree*");
  ix = marker(theTree,0);
  WriteIs(ix);
  if (ix>0) {
    SpellIt(thename,false);
    ExportNode(theTree);}
  unmark(theTree);
  WriteLn();
  WriteIs(-6);
  WriteLn();
  if (Debugging>0) sWriteS("-exportree*");
  return (ix>0);} /* ~fnexportree; */

void unmark(tree theTree) {
  int ix, nval;
  trees stree;
  tree atree;
  if (theTree==NIL) return;
  ix = theTree->kind;
  if (ix<mark) return; /* already got this one */
  ix = ix & (mark-1);
  theTree->kind = ix;
  nval = theTree->intval;
  if (ix==symt) nval = 1;
  if (ix>=wood) for (ix = 0; ix<=nval /* nsbtrees */; ix++) {
    stree = (trees) theTree;
    atree = stree->subtrees[ix];
    if (atree != NIL) unmark(atree);}} /* ~unmark; */

int marker(tree theTree, int theint) {
  int ix, nval;
  trees stree;
  tree atree;
  /* typecaster cheat; */
  if (theTree==NIL) return theint;
  ix = theTree->kind;
  if (ix >= mark) return theint; /* already saw this one */
  theint = theint+1;
  theTree->kind = ix+(theint << mrkn);
  if (ix<wood) return theint;
  nval = theTree->intval;
  if (ix==symt) nval = 1;
  for (ix = 0; ix<=nval /* nsbtrees */; ix++) {
    stree = (trees) theTree;
    atree = stree->subtrees[ix];
    if (atree != NIL) theint = marker(atree,theint);}
  return theint;} /* ~marker; */

void ExportNode(tree theTree) {
  int ix, nval, nn;
  trees stree;
  tree atree;
  if (theTree == NIL) return;
  ix = theTree->kind;
  if ((ix & mark) !=0) return; /* already got this one */
  theTree->kind = ix | mark;
  nval = ix >> mrkn;
  ix = ix & (mark-1);
  if (Debugging>2) sWriteS("+ExportNode:");
  if (ix==symt && theTree->intval<0)
    ExportNode(theTree->subtrees[1]); /* omit symbols with lex<0 */
  else {
    WriteLn();
    WriteIs(nval); /* node # */
    WriteIs(ix); /* *ORD* /(theTree->kind)); */
    nval = theTree->intval;
    if (ix==seti) ExpoSet(nval);
    else if (ix>=wood) {
      if (Debugging>0) sWriteS("<");
      stree = (trees) theTree;
      nn = stree->nodename;
      WriteIs(nn);
      if (ix>symt) WriteC('.');
      else if (ix==wood) WriteS((char*) (NodeNameS+nn*3));
      else if (!fnspell(nn)) WriteC('?');
      WriteC(' ');
      WriteIs(nval); /* stree->intval / *nsbtrees* /); */
      if (ix==symt) nval = 1;
      /* ExpoNodNum(stree->subtrees[0]); */
      for (ix = 0; ix<=nval /* nsbtrees */; ix++) {
        /* ExpoNodNum(stree->subtrees[ix]); */
        atree = stree->subtrees[ix];
        while (atree != NIL && atree->intval<0 && ((atree->kind)&(mark-1)) == symt)
          atree = atree->subtrees[1]; /* bypass negative lex symbols */
        if (atree != NIL) {
          WriteIs(atree->kind >> mrkn);}
        else WriteIs(0);}
      if (Debugging>0) sWriteS(">");
      for (ix = 0; ix<=nval /* nsbtrees */; ix++) {
        atree = stree->subtrees[ix];
        if (atree != NIL) ExportNode(atree);}
      if (Debugging>0) sWriteS("_");}
    else if (ix==snam) {
      if (Debugging>2) sWriteS("+Str:");
      WriteIs(nval);
      if (fnspell(nval)) ;} /* to the end of line */
    else WriteIs(nval);}
  if (Debugging>2) sWriteS("-ExportNode.");} /* ~ExportNode; */

void ExpoSet(int theSet) { /* output the elts in increasing order */
  if (Debugging>2) sWriteS("+Set:");
  if (theSet>0) {
    ExpoSet(UniqueSets[theSet].NextElt);
    WriteIs(UniqueSets[theSet].ThisElt);}} /* ~ExpoSet; */

trees GetixTree(trees theTree, int inx) {
  trees stree = theTree;
  if (Debugging>2) {
    sWriteS("+GetixTree:");
    ShoInt(inx);}
  if (inx==0) return null;
  while (inx>1) {
    if (stree==null) return null;
      else stree = (trees) stree->subtrees[1];
    inx--;}
  if (stree==null) return null;
    else return (trees) stree->subtrees[0];} /* ~GetixTree; */

int Skip2Int(int theFile, boolean NLok) { /* skips comment lines in input theFile */
  CHAR ch;
  int theInt;
  while(true) {
    ch = PeekInChar(theFile);
    if (ch==EOF_Char) return 0;
    if (EndLn(theFile)) {
      if (Debugging>0) sWriteC('/');
      if (!NLok) return 0;
      ch = ReadInChar(theFile);}
    else if (ch<=' ') ch = ReadInChar(theFile);
    else if (ch<'-') {
      if (Debugging>0) sWriteC(ch);
      while(!EndLn(theFile)) ch = ReadInChar(theFile);}
    else if (ch>'9') while(ch>' ') ch = ReadInChar(theFile);
      else break;}
  theInt = ReadInt(theFile);
  if (Debugging==0) return theInt;
  sWriteS(" #");
  ShoInt(theInt);
  return theInt;} /* ~Skip2Int; */

boolean fnimportree(int thename, VARt theTree) {
  int ix, nx, nval, Nnodes;
  trees atree; /* , root; CHAR ch; */
  tree* TrList;
  int vix = thename; /* not needed, it just stops compiler warnings */
  int theFile = FileOpen(4,"Library file..");
  if (theFile==0 || FileOpLen[theFile]<=0) return false;
  if (Debugging>0) sWriteS("+importree: ");
  Nnodes = Skip2Int(theFile,true);
  if (Nnodes>=MaxImportTrees || Nnodes<=0) AbortInt("Invalid Library file",0,Nnodes);
  TrList = (tree*) malloc((Nnodes+3)*sizeof(tree*));
  if (TrList==null) return false;
  TrList[0] = null;
  /* root = null; -- build list of Nnodes numbered trees, use them for input #s.. */
  nx = Debugging;
  for (ix = 1; ix<=Nnodes; ix++) {
    TrList[ix] = build(TabNo,NIL,NIL,NIL,NIL,NIL,NIL,NIL,NIL,NIL,0);
    /* root = build(TabNo,atree,root,NIL,NIL,NIL,NIL,NIL,NIL,NIL,1); */
    Debugging = 0;}
  Debugging = nx;
  *theTree = TrList[1]; /* GetixTree(root,1); */
  for (ix = 1; ix<=Nnodes; ix++) {
    if ((ix&-1024)==0) if (fntcheckpoint()) ; /* pacifier for long imps */
    if (Debugging>0) {
      ShoInt(ix);
      sWriteC('+');}
    nx = Skip2Int(theFile,true);
    if (nx<0) break;
    if (nx>Nnodes) AbortInt("Oversize node in Library file",ix,nx);
    atree = TrList[nx]; /* fill in tree # ix.. */
    if (atree == null) AbortInt("Inconsistent Library file",ix,nx);
    nx = Skip2Int(theFile,false);
    if (nx>=mark || nx<=0) AbortInt("Bad node in Library file",ix,nx);
    atree->kind = (nx); /* nodeclass */
    if (nx>=wood) { /* it's a real (wood) tree node, or symtable, or pack.. */
      nval = Skip2Int(theFile,false);
      atree->nodename = nval;
      vix = 0;
      if (nx==symt) {
        vix = ImportString(theFile, 1);
        if (nval<FirstString) vix = 0;}
      if (vix>0) atree->nodename = vix;
      nval = Skip2Int(theFile,false);
      atree->intval = nval; /* nsbtrees */
      if (nx==symt) nval = 1; /* symbol table entries (symt=6).. */
      for (nx = 0; nx<=nval; nx++) { /* nsbtrees */
        vix = Skip2Int(theFile,false);
        if (vix>Nnodes || vix<0) vix = 0;
        atree->subtrees[nx] = TrList[vix];}}
    else if (/* nodeclass */ (nx) == seti) { /* set: ascending elts on this line.. */
      nval = 0;
      while (! EndLn(theFile)) nval = myaddset(Skip2Int(theFile,false),nval);
      atree->intval = nval;}
    else { 
      vix = 0;
      nval = Skip2Int(theFile,false);
      atree->intval = nval;
      if (nx != snam) continue; /* otherwise: string table entries (snam=3).. */
      vix = ImportString(theFile, -1);
      if (vix>0) atree->intval = vix;}}
    /* next line starts with Skip2In */
  FileClose(theFile);
  free(TrList);
  ShoLexLvl = true;
  if (Debugging>0) {
    sWriteS("*importree: ");
    ShoTreDeep(*theTree,999,0);
    sWriteLn();
    sWriteS("-importree ");
    sWriteLn();}
  return true;} /* ~fnimportree; */

int ImportString(int theFile, int kin) {
  CHAR ch;
  int vix = 0;
  if (!fninitbl()) return 0;
  if (EndLn(theFile)) return 0;
  ch = ReadInChar(theFile);
      /** if (Debugging>00) {sWriteS(" \""); ShoInt(kin); ShoInt(ch);} **/
  while (ch==' ') {
    if (EndLn(theFile)) return 0;
    ch = ReadInChar(theFile);
    if (kin==0) break;} /* skip exactly one space for string literals */
  while (true) {
    if (ch<' ' || ch==EOF_Char || ch==' ' && kin > 0) break;
    if (kin>0) if (ch=='?') return 0;
    if (kin<0) if (ch<'0' || ch>'z' || ch>'9' && ch<'A' || ch>'Z' && ch<'a' && ch!='_')
      kin = 0; /* auto-type to string literal */
    if (!fnaddtbl(ORD(ch))) break;
    vix++;
    if (EndLn(theFile)) break;
    ch = ReadInChar(theFile);}
      /** if (Debugging>00) {sWriteS(" ~"); ShoInt(vix); ShoInt(chx);} **/
  if (vix==0) return 0;
  if (kin<0) kin = -kin; /* auto-type to identifier */
  if (kin==1) vix = 1; else vix = 0; /* assume identifier equiv class =1 */
  if (fnstrindex(vix, kin, &vix)) {
    if (Debugging>0) {
      sWriteS(" ");
      SpellIt(vix,true);}
    return vix;}
  return 0;} /* ~ImportString */
#endif

/**************************** Debugging Utility Procedures ****************************/

/* -- see also ShowStrings */

void ShoTreDeep(tree theTree, int numLevels, int indent) {
  /* Shows contents at 1+ level, shows nulls or ..`s at 0 level or lower */
  int index, elt;
  int* here;
  trees stree;
  if (Debugging<0) return;
  for (index=0; index<16; index++) {
    if (MagiTrees[index] == null) continue;
    if (MagiTrees[index] == theTree) {
      sWriteS("{");
      sWriteI(index);
      sWriteS("}");
      if (Debugging<3 && SeenMagiTr[index]>DbgOutLns) {
        sWriteS(".."); return;}
      else SeenMagiTr[index] = DbgOutLns+5000;}}
  if (indent<60) {
    sWriteLn();
    for (index=0; index<indent; index++) {
      if (index<20) sWriteS(". "); else sWriteS(".");}}
  if (theTree != NIL) if (Debugging>1) sWriteI((((int)theTree)>>4)&8191);
  sWriteS("<");
  if (theTree == NIL) sWriteS(">");
  else { /* validptr(theTree); */
    here = &(theTree->kind);
    elt = *here;
    *here = (elt | mrkt);
    numLevels = numLevels-1;
    if (numLevels < 0) sWriteS("..>");
    else if ((elt & mrkt) != 0) {
      sWriteS("..#>");
      return;}
    else { /* WITH theTree^ DO */
      if (elt>pkat) {
        sWriteI(/* ORD */(elt));
        sWriteS("HUH?>");
        sWriteLn();
        sWriteS("=================================================");
        sWriteLn();}
      else switch (elt) {
        case bool:
          if (theTree->intval>0 /* boolval */) sWriteS("T>");
            else sWriteS("F>");
          break;
        case intt:
          /* sWriteS("i"); */
          index = theTree->intval;
          sWriteI(index);                                           if (false)
          if (index>=FirstString) if (index<EndStrings) if (index%StrRef ==0) {
            sWriteS(" '");
            SpellIt(index,true);
            sWriteS("'");}
          sWriteS(">");
          break;
        case snam:
          /* sWriteS("s"); */
          index = theTree->intval;
          sWriteI(index);
          if (index>=FirstString) if (index<EndStrings) {
            sWriteS("=");
            SpellIt(index,true);}
          sWriteS(">");
          break;
        case nono:
          sWriteS("?>");
          break;
        case seti:
          elt = theTree->intval;
          sWriteI(elt);
          if (Debugging>2) {
            sWriteS("{");
            index = 1;
            while (elt>0) {
              ShoInt(UniqueSets[elt].ThisElt);
              if (Debugging>6) {
                sWriteS("/"); sWriteI(elt);
                if (index>0) {
                  index = index+1;
                  if (index>7) {
                    index = theTree->intval;
                    if (myinset(index,UniqueSets[SawSetSet].SetPtr))
                      numLevels = 0;
                    else SawSetSet = myaddset(index,SawSetSet);
                    index = 0;}}}
              elt = UniqueSets[elt].NextElt;
              if (elt>0) if (numLevels < 2) {
                sWriteS("..");
                elt = 0;}}
            sWriteS("}");}
          sWriteS(">");
          break;
        case symt:
          /* sWriteS("40"); */
          stree = (trees) theTree;
          /* if (Line_Len>72) sWriteLn(); */
          if (numLevels<7777)
            ShoTreDeep(stree->subtrees[1], numLevels, 999);
          else ShoTreDeep(stree->subtrees[1], 0, 99);
          sWriteLn();
          if (ShoLexLvl) ShoInt(stree->intval);
          sWriteS(":");
          index = stree->nodename;
          if ((index>=FirstString) && (index<EndStrings) && (index%StrRef ==0)) {
            SpellIt(index,true);
            if (Debugging>2) ShoInt(index);}
          else ShoInt(index);
          sWriteS(" ");
          index = Debugging>4 ? numLevels :
            (Debugging>2 ? 9 : (Debugging<2 ? 1 : 3));
          ShoTreDeep(stree->subtrees[0], index<numLevels?index:numLevels,
            indent>888 ? 2 : indent+1);
          sWriteS(">");
          break;
        case pkat:
          stree = (trees) theTree;
          sWriteS("@");
          ShoInt(stree->intval);
          for (index = 1; index<=5; index++) {
            sWriteS("/");
            ShoTreDeep(stree->subtrees[index], numLevels, indent+1);}
          sWriteS(">");
          break;
        case wood:
          stree = (trees) theTree;
          index = stree->nodename;
          SpellNodeName(index);
          sWriteS(" ");
          for (index = 1; index<=stree->intval /* nsbtrees */; index++) {
            if (numLevels<7777) 
              ShoTreDeep(stree->subtrees[index], numLevels, indent+1);
            else {
              ShoTreDeep(stree->subtrees[index], 0, 99);
              numLevels = 1001;}
            if (index < stree->intval /* nsbtrees */) sWriteS(" ");}
          sWriteS(">");
          if (stree->subtrees[0] != NIL) {
            sWriteS("%");
            ShoTreDeep(stree->subtrees[0], 1, 99);}
      } }  /* CASE else */
      *here = ((*here) & (~mrkt));}} /* else // ~ShoTreDeep; */

void LogReplay (CHAR theCh) { /* log input&output for diagnostic purposes */
  theChecksum = ((theChecksum>>15)+theChecksum+theChecksum+((int)theCh))&65535;
  ReadReplay[ReplayHere&1023] = theCh;
  ReplayHere = ReplayHere+1;} /* ~LogReplay */

void SaveChecksum(int failed) { /* record it for next run, in case it fails again */
#ifdef FailMsFileName
  char* theString;
  if (CkSumHere<64) failed = 0;
  if (failed) failed = CkSumReplay[CkSumHere&63];
  theString = IntToStr(failed);
  Open_File(6, "");
  fileWriteS(6,theString);
  WriteOutChar(6,CRLF);
  Close_File(6);} /* ~SaveChecksum */
#else
  } /* ~SaveChecksum */
#endif

void GetChecksum(void) { 
#ifdef FailMsFileName
  Open_File(5, ""); /* hopefully fails gracefully if not there */
  PriorChksum = ReadInt(5);
  Close_File(5);
#endif
  CheckOff = -1;} /* ~GetChecksum */

void ChekChecksum(void) { /* called for each token or nth output */
  int herenow = theChecksum+((CkSumHere&65535)<<16);
  if (CheckOff>0) {
    CheckOff = CheckOff-1;
    if (CheckOff==0) if (Debugging==PreMortemLevel) {
      Debugging = 0;
      sWriteLn();
      sWriteS("** Stopping pre-log **");
      sWriteLn();}}
  else if (CheckOff<0) if (herenow==PriorChksum) if (PriorChksum !=0)
      if (Debugging<PreMortemLevel) {
    Debugging = PreMortemLevel;
    CheckOff = 99;
    sWriteLn();
    sWriteS("** Starting pre-mortem log ");
    sWriteI(Debugging);
    sWriteS(" **");
    sWriteLn();}
  CkSumReplay[CkSumHere&63] = herenow;
  CkSumHere = CkSumHere+1;} /* ~ChekChecksum */

void Abortit(char* ErrMsg) {
  int ix = InputLinePosn;
  CHAR ch;
  ErrLnCapIx = -ErrLnCapIx;
  KillingIt = 1;
  DbugFiLen = 0; /* so we don't repeat a runaway */
  SaveChecksum(true);
  if (EchoInputBuffer !=0) if (InputLinePosn>3) {
    WriteLn();
    EchoThisInput('\n');
    while (ix>0) {WriteS(". "); ix = ix-2;}
    WriteS("^");}
  WriteLn();
  WriteS(ErrMsg);
  sWriteLn();
  if (EchoInputBuffer==0) {
    ix = ReplayHere-1023;
    if (ix<0) ix = 0;
    while (ix<ReplayHere) {
      ch = ReadReplay[ix&1023];
      if (ch != Null_Char) fileWriteC(0,ch);
      ix = ix+1;}
    sWriteLn();}
  sWriteS(ErrMsg);
  TagEnd();
  if (ErrLnCapIx<0) ErrMsg = ErrLnCapture;
  Stop_Run(ErrMsg);} /* ~Abortit; */

void AbortInt(char* ErrMsg, int lino, int aNum) {
  DbugFiLen = 0; /* so we don't repeat a runaway */
  sWriteS("Died at line ");
  sWriteI(lino);
  sWriteS(" with #");
  sWriteI(aNum);
  Abortit(ErrMsg);} /* ~AbortInt; */

void MemBomb(void) {Abortit("***MEMERR***");} /* ~MemBomb(); */

void ShowTransList(void) { /* diagnostic trace of TransPtr & DetrnPtr */
  int ix;
  sWriteLn();
  sWriteS("*TransList ");
  sWriteI(transnumber);
  for (ix = 1; ix<=transnumber; ix++) {
    sWriteLn();
    sWriteS("*");
    sWriteI(ix);
    /* WITH TransPtr[ix] DO */
    ShoInt(TransList[ix].fromst);
    ShoInt(TransList[ix].tost);
    ShoInt(TransList[ix].actset);
    ShoSett(TransList[ix].onset);
    /* END WITH */}
  sWriteLn();
  sWriteS("+TransList ");
  sWriteI(detransnmber);
  for (ix = 1; ix<=detransnmber; ix++) {
    sWriteLn();
    sWriteS("*");
    sWriteI(ix);
    /* WITH DetrnPtr[ix] DO */
    ShoInt(DeTrnList[ix].fromst);
    ShoInt(DeTrnList[ix].tost);
    ShoInt(DeTrnList[ix].actset);
    ShoInt(DeTrnList[ix].onchar);
    /* END WITH */}
  sWriteLn();
  sWriteS("-TransList ");
  sWriteLn();} /* ~ShowTransList(); */

void ShowK(int value, int refer) {
  sWriteI(value/1024);
  sWriteS("/");
  sWriteI(refer/1024);} /* ~ShowK; */

void StatsReport(void) {
#ifdef FailMsFileName
  sWriteI(Tree_Count);
  sWriteS(" trees, Strs==");
  ShowK(EndStrings,StringMax);
  sWriteS("K, Chrs==");
  ShowK(EndChars,CharMax);
  sWriteS("K");
  sWriteLn();
  sWriteS("  Sets==");
  ShowK(LastUniSet,TopSetNo);
  sWriteS("K, Stk==");
  ShowK(HiStack,StackMax);
  sWriteS("K, Trn==");
  sWriteI(transnumber);
  if (transnumber+detransnmber>0) {
    sWriteS("/");
    sWriteI(TransMax);
    sWriteS(",");
    sWriteI(detransnmber);
    sWriteS("/");
    sWriteI(DetrnMax);}
#else
  int best = (EndStrings*100)/StringMax;
  int ix = (EndChars*100)/CharMax;
  KillingIt = 1;
  sWriteLn();
  if (ix>best) best = ix;
  ix = (LastUniSet*100)/TopSetNo;
  if (ix>best) best = ix;
  ix = (HiStack*100)/StackMax;
  if (ix>best) best = ix;
  ix = (transnumber*100)/TransMax;
  if (ix>best) best = ix;
  ix = (detransnmber*100)/DetrnMax;
  if (ix>best) best = ix;
  sWriteI(best);
  sWriteS("% tables used.");
#endif
  if (Debugging>0) {
    sWriteLn();
    sWriteS("  Pre-mortem logged \"");}
  sWriteLn();} /* ~StatsReport(); */

void validtree(tree thetree) { /* validptr(thetree); */
  if (thetree != NIL) if (thetree->kind != wood) Abortit("(invalid tree)");
  } /* ~validtree; */


/**************************** User Trace Utilities ****************************/

void ShoBool(int outBool) {
  sWriteC(' ');
  sWriteB(outBool);} /* ~ShoBool; */

void ShoInt(int outInteger) {
  sWriteC(' ');
  sWriteI(outInteger);} /* ~ShoInt; */

void ShoStrn(int outInteger) {
  sWriteC(' ');
  sWriteI(outInteger);
  if (outInteger>=FirstString) if (outInteger<EndStrings) {
    sWriteS("=");
    SpellIt(outInteger,true);}} /* ~ShoStrn; */

void ShoTable(table outTable) {
  /* validtable(outTable); */
  sWriteS(" ");
  if (outTable==RecenTable) ShoTreDeep(outTable,1,99);
  else if (Debugging>4) ShoTreDeep(outTable,999,0);
  else if (Debugging>1) ShoTreDeep(outTable,3,0);
  sWriteS(";");
  RecenTable = outTable;} /* ~ShoTable; */

void ShoTree(tree outTree) {
  sWriteC(' ');
  if (Debugging>4) ShoTreDeep(outTree, 999, 0);
  else if (Debugging>1) ShoTreDeep(outTree, 9, 0);
    else ShoTreDeep(outTree, 2, 99);
  sWriteS(",");} /* ~ShoTree; */

void ShoSett(settt outTree) {
  sWriteC(' ');
  if (Debugging>1) {
    Debugging++;
    ShoTreDeep(outTree, 9, 0);  /* is (tree) cast necessary?? */
    Debugging--;
    sWriteS(",");}
  else ShoTreDeep(outTree, 1, 99);} /* ~ShoSett; */

void sWriteB(int outBool) { /* Write a boolean to the debug "screen" */
  if (outBool) fileWriteS(DbugFile, "true");
    else fileWriteS(DbugFile, "false");} /* ~sWriteB; */

void sWriteC(CHAR outChar) { /* Write a character to the debug "screen" */
  fileWriteC(DbugFile, outChar);} /* ~sWriteC; */

void sWriteI(int value) { /* Write an integer to the debug "screen" */
  fileWriteS(DbugFile, IntToStr(value));} /* ~sWriteI; */

void sWriteLn(void) { /* Write a CR (ASCII 13) to the debug "screen" */
  Line_Len = 0; fileWriteC(DbugFile, CRLF); DbgOutLns++;} /* ~sWriteLn(); */

void sWriteS(char* outString) { /* Write a string to the debug "screen" */
  int now; /* Time_Tix; */
  if (Debugging<0) if (outString[0]=='+') { /* never happens */
    Debugging = NoisyTrace;
    sWriteLn();
    fileWriteS(DbugFile, "++++++++++++++++++++++++++++++++");
    sWriteLn();}
  fileWriteS(DbugFile, outString);
  if (Line_Len>80) if (outString[0]==')') sWriteLn();
  /* user pacifier occasionally writes to screen */
  if (DbugFile != 0) {
    now = Tick_Count();
    if (now > dbgNextTime) {
      if (dbgNextTime>0) fileWriteC(0, CRLF); /* CR */
      fileWriteS(0, outString);
      dbgNextTime = now + 120;}} /* two seconds later */
  } /* ~sWriteS; */

boolean fntpremolevel(int theLevel) { /* set pre-mortem trace level */
  PreMortemLevel = theLevel; return true;} /* ~fntpremolevel; */

boolean fntroff(void) { /* turn off trace */
  Debugging = 0; return true;} /* ~fntroff; */

boolean fntron(int theLevel) { /* turn trace on to spec`d level */
  Debugging = theLevel; return true;} /* ~fntron; */

boolean fntrlevel(VARi theLevel) { /* get current trace level */
  *theLevel = Debugging; return true;} /* ~fntrlevel; */

boolean fntrasci(int theChar) { /* show this char */
  if (theChar==0) sWriteLn();
    else fileWriteC(DbugFile, CHR(theChar));
  return true ;} /* ~fntrasci; */

boolean fntrnmbr(int theNum) { /* show this number */
  sWriteI(theNum); return true;} /* ~fntrnmbr; */

boolean fntrhexn(int value, int ndigs) { /* show this number in hex */
  while (ndigs-- >0) WriteHexDigit(true, value >> (ndigs*4));
  return true;} /* ~fntrhexn; */

boolean fntrspel(int theName) { /* show this name */
  if (theName%StrRef ==0) if (theName>=FirstString) if (theName<EndStrings) {
    SpellIt(theName,true);
    return true;}
  return false;} /* ~fntrspel; */

boolean fnttreedmp(tree theTree) { /* show this tree */
  int ix;
  if (fntrunflag(theTree, &ix)) if (ix>=0) fntreeflag(theTree, ix);
  ShoTreDeep(theTree, 999, 0); return true;} /* ~fnttreedmp; */

boolean fntsetdmp(settt theTree) { /* show this set */
  Debugging = Debugging+3;
  ShoTreDeep(theTree, 999, 0);
  Debugging = Debugging-3;
  return true;} /* ~fntsetdmp; */

boolean fnttabdmp(table theTree) { /* show this tree as table */
  Debugging = Debugging+0;
  ShoTreDeep(theTree, 9999, 0);
  Debugging = Debugging-0;
  if (theTree==null) return true;
    else return fnttabdmp((table) theTree->subtrees[1]);} /* ~fnttabdmp; */

boolean fnttoptree(tree theTree, int theChar) { /* show first tree in table */
  Debugging = Debugging+3;
  sWriteLn(); 
  if (theChar>0) {
    sWriteS("");
    if (theChar>32) fileWriteC(DbugFile, CHR(theChar));
    sWriteS(" ");} 
  ShoTreDeep(theTree, 9999, 0);
  Debugging = Debugging-3;
  return true;} /* ~fnttoptree; */

boolean fnttranslsts(tree theTree) { /* show both trans lists */
  ShowTransList(); return true; theTree = 0;} /* ~fnttranslsts; */

boolean fntcheckpoint(void) { /* I/O-less processing checkpoint, for crash tracing */
  sWriteS(" ."); LogReplay('.'); ChekChecksum(); return true;} /* ~fntcheckpoint; */

boolean fntreeflag(tree theTree, int theIndx) { /* set this tree's ref# for trace */
  if (theIndx<0) return false;
  if (theIndx>15) return false;
  MagiTrees[theIndx] = theTree;
  SeenMagiTr[theIndx] = 0;
  return true;} /* ~fntreeflag; */

boolean fntrunflag(tree theTree, VARi theIndx) { /* unset this tree's ref# */
  int ix;
  if (theTree !=null) for (ix=0; ix<16; ix++) if (MagiTrees[ix] == theTree) {
    *theIndx = ix;
    MagiTrees[ix] = null;
    return true;}
  *theIndx = -1;
  return true;} /* ~fntrunflag; */

void SpellNodeName(int theNode) {
  sWriteS((char*) (NodeNameS+theNode*3));} /* ~SpellNodeName; */


/**************************** Compiled TAG Procedures ****************************/

void ModuleInitialization(void) { /* module initialization */
  TabNo = 0;
  evtNextTime = 0;} /* ~ModuleInitialization(); */

int TagBegin(void) { /* might longjmp out, but can't return false */
  int ix;
  for (ix=0; ix<16; ix++) {MagiTrees[ix] = null; SeenMagiTr[ix] = 0;}
  IO_Startup();
  Debugging = NoisyTrace;
  Taken = 0;
  sourceFile = 0;
  destFile = 0;
  sourceFile = FileOpen(1, "Grammar input file..");
  if (sourceFile == 0) return false;
  destFile = FileOpen(2, "Save C code as...");
  if (destFile == 0) return false;
  DbugFile = FileOpen(3, "Debug (Cancel=screen)...");
  dbgNextTime = 0;
  GetChecksum();
  stackTop = 0;
  HiStack = 0;
  /* theStack = new int[StackMax]; */
  /* if (theStack==null) return false; */
  if (! InitStrings()) return false; /* can't fail */
  if (! InitSets()) return false; /* can't fail */
  RecenTable = null;
  ModuleInitialization();
  RunPhase = 1;
  return true;} /* ~TagBegin; */

void TagEnd(void) {
  CHAR donestate = SavedChar;
  if (RunPhase<0) return; /* already been here */
  RunPhase = -1;
  sWriteLn();
  if (EchoInputBuffer !=0) if (InputLinePosn !=0) EchoThisInput('\n');
  StatsReport();
  if (sourceFile != 0) FileClose(sourceFile);
  if (destFile != 0) FileClose(destFile);
  if (DbugFile != 0) {
    FileClose(DbugFile);
    DbugFile = 0;
    sWriteLn();
    if (donestate==CHR(83)) {
      SaveChecksum(false);
      sWriteS("Success ");}
    else sWriteS("Failed ");
    StatsReport();}} /* ~TagEnd(); */

void SyntaxError(void) {Abortit(" syntax error");} /* ~SyntaxError(); */

/************************** Revision History **************************/

/* 2003 May 20 -- first (working) public release */
/* 2003 June 27 -- added failure diagnostics */
/* 2003 July 1 -- added support for reserved words */
/* 2003 July 3 -- minor bug fix */
/* 2003 December 26 -- added some more predefined node names */
/* 2004 January 6 -- added memory image functions */
/* 2004 January 19 -- columnized ShoTreDeep, stopped tree recursion */
/* 2004 February 28 -- delayed token display in debug trace */
/* 2004 March 16 -- reduced runaway trace to 1MB */
/* 2004 March 27 -- fixed ExportNode to write to OutputFile */
/* 2004 April 9 -- removed node names from string table */
/* 2004 May 31 -- some bug fixes, enlarged MagiTrees */
/* 2004 June 24 -- added StackMax diagnostic, fixed library import */
/* 2004 July 9 -- bigger; trunflag no longer fails on <> */
/* 2004 July 28 -- needed CHAR so addtbl wouldn't disallow \r or \n */
/* 2004 September 15 -- new function srctoklinch() gets source line+char */
/* 2004 September 24 -- bigger StackMax */
/* 2004 November 3 -- AbortInt gives more info when fnimportree dies */
/* 2005 November 3 -- added fnhexnum() for hex text output */
/* 2005 November 3 -- if EchoInputBuffer>0, echo input lines to output */
/* 2004 November 9 -- added str type (snam) for string exports */
/* 2005 December 20 -- added ErrLnCapture, to show error msg on console */
/* 2006 January 11 -- fixed nodeTsize bug; doubled sets; tab to space */
/* 2006 January 14 -- added fntrhexn() for hex text in trace */

/**************************** To-Do List ****************************/

