// TurkFrame.h, Win32 threads version -- 2010 January 14

#define Debug_Trace "C:\\AllDocs\\DebugLog.txt" // enables DebugLogFile
#define _Crash_and_Burn_ true // div/0 in Array_Bounds_Fail and elsewhere
#define SigWait_Trace 0 // C++ doesn't like: true; =2 to div/0
#define Malloc_Log 0 // =3 logs malloc cache for leakage; =5 logs Res; =0 normal

#define Drag_Trace true
#define Draw_Trace false // traces image items in DrawImag
#define Sys_Trace 0 // turns message traffic on/off
#define Post_Trace 0 // logs Post_Ev (Send/Async) only
#define Panic_Log 0 // 1024; =2^n; >0 logs recent stuff (evts only)
#define Event_Log 2048 // =2^n; >1 logs recent Win32 evts

namespace  _System_ {int System_Peek_Address(int ix);} // end  _System_
// #define PEEK_OFTEN _System_::System_Peek_Address(81)

#ifndef MOS_FrameWork  // _FrameWork_::xxx
#define MOS_FrameWork
#define FrameWork_Body

#define CR_LF "\r\n" // "\r" for Mac, "\n" for eunuchs, "\r\n" for Windoze
#define CRLF_LEN 2 // 1 for Mac/eunuchs, 2 for Windoze -- 2010 January 14

#include "SysNames.h" // all the MOS framework prototypes

#define _BASE_ 0x10000

// This symbol needs to be defined before this header is #included:
//  char* ProgramName = "IttyBittyComputer"; // or whatever
//
// StartNewProg needs to be defined somewhere:
//  DWORD StartNewProg(void* whom); // whom is really a Process

#define _String_Log_ 256 // =256 if using MonitorStrings(), else =1

// bool My_Hold_Off = false, My_Test_Flag = false; // for testing D&D

// #define _w32RECT_ RECT
// #define _wPOINT_ POINT
#define wPOINT POINT*
#define w32RECT RECT*

// _wPOINT_ _1pint_, *w1pint = &_1pint_;
// _w32RECT_ _1rect_, *w1rect = &_1rect_, _2rect_, *w2rect = &_2rect_;

inline w32RECT Win32Rect(RectAngle rx) { // trims off size,ref-count
  int* rw = (int*)rx; ++rw; return (w32RECT)(++rw);}

inline wPOINT Win32Point(RectAngle rx) { // trims off size,ref-count,rTop
  int* rw = (int*)rx; ++rw; ++rw; return (wPOINT)(++rw);}

#define PathSize (MAX_PATH*4)

#define App_Icon 1001
#define Ask_Text 129
#define IDC_EdLn 1000
#define IDC_AskLabel 1002

char MsgBoxTitle[sizeof(ProgramName)+16];

////////////////////////////////////////////////////////////////////////

// Thread synchronization, implemented as a Win32 semaphore object

#define NewSem32(sem,pre,snam) sem = CreateSemaphore(NULL,pre,0xFFFFF,snam)
#define Signal32(sem) ReleaseSemaphore(sem,1,NULL)
#define Wait32(sem) WaitForSingleObject(sem,INFINITE)

// #define Str_Signal(sem) Signal32(sem) // could also try it without these..
// #define Str_Wait(sem) Wait32(sem) // (insignificantly faster)

#define DebugLogFile _FrameWork_::DebugLogFileOpen // used as a boolean
#define SysDebugLogging() (_FrameWork_::DebugLogFileOpen)

namespace  _FrameWork_ { // for private library functions..

int Mouse_Buttons = 0; // captured from recent button messages; see ControlKey
int Modifier_Keys = 0; // modifiers: lock:0, shft:1, ctrl:2, cmd:3, opt:4

File DebugLogFileOpen = NULL;

void DebugTracer(char* str, int whom, int lino) { // logs process ID
  if (DebugLogFile) // only from Log_Sig/Log_Wai if SigWait_Tr=true
    SystemDebugLog(String(str)+whom+" "+FormInt(lino)+" "+WhoAmI());}

Sem32 MoveOK = NULL, ResFiLok = NULL, DebugLok = NULL, StrnLok = NULL,
    GooLok = NULL, EvenLok = NULL, PanicLok = NULL;
  // MoveOK protects my_malloc's Chunk_List
  // ResFiLok protects PIBfiles in _File_ constructor/destructor,
  //   also in System.h, Buf_Root and ResCache in resource manager
  // DebugLok protects DebugLogFile in SystemDebugLog
  // StrnLok protects String ref count in UpDownStr, also the_String_Log
  // GooLok protects Xfer_xx through DoGooey
  // EvenLok protects T2F_Process _event_list_ and n_Timers
  // PanicLok protects Panic_Log trace data (if enabled)

int SemRefLockt = 0, SemRefCount = 0, SemRefHisto = 0; // used only here..

void Log_Signal(Sem32& s, int whom, int lino) { // only by System.h; whom: 1..7
  SemRefLockt--;
  if (SigWait_Trace) { // enable logging, to help find mismatches..
    SemRefCount = SemRefCount+100; // was odd 10 if locked
    if (SemRefLockt<0) { // lost Wait or excess Signal..
      if (DebugLogFile)
        DebugTracer("~~Signal ",SemRefCount+whom+(SemRefLockt&7)*10,lino);
      if (_Crash_and_Burn_) if (SigWait_Trace>1) whom = whom/(whom>>4);} // div/0
    else if (DebugLogFile)
      DebugTracer("~ Signal ",SemRefCount+whom+(SemRefLockt&7)*10,lino);}
  SemRefHisto = (SemRefHisto<<4)+whom; // recent calls, even if not logging
  Signal32(s);}
void Log_Wait(Sem32& s, int whom, int lino) { // only by System.h; whom: 1..7
  if (SigWait_Trace) {
    SemRefCount = SemRefCount+100; // lost Signal just hangs
    if (DebugLogFile)
      DebugTracer("~ WaitSm ",SemRefCount+whom+(SemRefLockt&7)*10,lino);}
  SemRefHisto = (SemRefHisto<<4)+whom+8;
  Wait32(s);
  SemRefLockt++;}

} // (temply) end namespace  _FrameWork_

DWORD StartNewProg(void* whom); // forward decl'n for funct in main prog
DWORD Gooey_Thread_ID; // was: GooeyThreadID;

////////////////////////////////////////////////////////////////////////

// Standard type conversions and other built-ins [moved to SysNames.h]..

namespace  _FrameWork_ { // resume

Process MainProgram = NULL, Ready_List = NULL; // we need this soon
bool LogTextWrites = false; // initially disabled

int Logged_Events[Event_Log];
int Ev_Log_Ix = 0;

void FILL0(int* ary, int offs, int nwds) { // adjusts for array size+cnt words
  ++offs; while (nwds-- >0) ary[++offs] = 0;} // offs in words

void copy_bytes(char* frm, char* too, int nby) {
  while (nby-- >0) *too++ = *frm++;}
inline void BlockMove(int frm, int too, int nby) {
  copy_bytes((char*)frm,(char*)too,nby);}

void Fill_Block(int valu, int* addr, int nwds) { // skips over size,ref wrds
  addr++;
  *(++addr) = valu;
  char* thar = (char*)addr;
  _FrameWork_::copy_bytes(thar,thar+4,nwds*4-4);}

inline int PeekByte(int cPtr) {
  char* xptr = (char*)cPtr;
  return ((int)*xptr)&255;}
inline void PokeByte(int cPtr, int val) {
  char* xptr = (char*)cPtr;
  *xptr = (char)(val&255);}

// My Own Memory Allocation Manager (cuz Microsoft's bombs)..

#ifdef Debug_Trace
#define fronx (Malloc_Log+2) // links to prior w/odometer
//#define fronx 2
#else
#define fronx 0
#endif

#define Chunk_Tag 0x804D4D00 // "`MM#"
//        each size:  16  32  64 128 256 512 1K 2K 4K 8K 16 32 64 // max save
int Enuff_Chunx[] = {200,100,100,100,100,100,80,50,20,20,10,10,10};
#define Num_Chunks 12 // 13 chunk sizes
#define Min_Chunk 0x10 // 16 ints in smallest chunk
#define Max_Chunk (Min_Chunk<<Num_Chunks) // = 64K
int* Chunk_Linx[Num_Chunks+1]; // all alloc list heads, if Malloc_Log !=0
int* Chunk_List[Num_Chunks+1]; // free lists
int Chunk_Free[Num_Chunks+1]; // how many we have of that size free
int Chunk_Count[Num_Chunks+1]; // how many of that size in use
int Chunk_Nops[Num_Chunks+1]; // actual high idle
int Chunk_Tops[Num_Chunks+1]; // actual high usages
#define Chunk_Hist 512
#define Chunk_Mast (Chunk_Hist-1)
int Chunk_Acts[Chunk_Hist];  int Recent_Chunk = 0;  int Net_Chunx = 0;
  int Chunx_Odom = 0;  int Chunx_Stop = 0;
#define Chunk_Freeze 0x5100 // >0 stops Chunk_Act when Chunx_Odom=Freeze
#define Chunk_Bitz 32
int Chunk_Bits[Chunk_Bitz+4]; // each bit: 1MB in memory space
#define Chunk_Edge 1

void Init_MOMAM() {
  int ix;
  for (ix=0;ix<Num_Chunks;ix++) {
    Chunk_List[ix] = NULL;
    Chunk_Linx[ix] = NULL;
    Chunk_Count[ix] = 0;
    Chunk_Free[ix] = 0;
    Chunk_Nops[ix] = 0;
    Chunk_Tops[ix] = 0;}
  for (ix=0; ix <= Chunk_Bitz; ix++) Chunk_Bits[ix] = 0;
  for (ix=0;ix<Chunk_Hist;ix++) Chunk_Acts[ix] = 0;} //~Init_MOMAM

int* Record_Chunk(int* where) {
  int aBit = (int)where, here = aBit>>25;
  if (aBit<0) here = Chunk_Bitz+3; // shouldn't hap
  else if (here>Chunk_Bitz) here = Chunk_Bitz;
  aBit = 1 << ((aBit>>20)&31);
  Chunk_Bits[here] = Chunk_Bits[here] | aBit;
  return where;} //~Record_Chunk

int* my_malloc(int lnx, int whom) { // _FrameWork_::my_malloc()
  int abit = Min_Chunk, chn = 0, size = ((lnx+11)>>2)+fronx;
  int* thePtr = NULL;            // size is fronx+2 ints bigger than req'd
  if (lnx==0) return NULL;
  Net_Chunx++;
  try {
    if (size >= Max_Chunk) {
      thePtr = (int*)malloc(size<<2);
      if ((Chunx_Odom<Chunk_Freeze) || (Chunk_Freeze==0)) {
        Chunk_Acts[(Recent_Chunk++)&Chunk_Mast] = size+0x51000000;
        Chunk_Acts[(Recent_Chunk++)&Chunk_Mast]
            = ((int)thePtr)&0xFFFFFFC | 0xF0000000;                      // F..0
        if (whom !=0) Chunk_Acts[(Recent_Chunk++)&Chunk_Mast] = whom;
        Chunk_Acts[Recent_Chunk&Chunk_Mast] = -1;}
      if (Malloc_Log) {
        *thePtr++ = 0;
        *thePtr++ = whom;
        *thePtr++ = ++Chunx_Odom;
        if (Malloc_Log==5) {*thePtr++ = 0; *thePtr++ = 0;}}
      if (fronx) {
        Fill_Block(0xDEADBEEF,--thePtr,size-2); // fills all but link and fence
        thePtr[size] = 0xBAD1DEA5; // sets the no-touchum fence
        thePtr++;}
      *thePtr++ = Chunk_Tag+Num_Chunks;
      if (fronx) {
        *thePtr++ = size;
        *thePtr++ = 0xBAD1DEA5;}
      return Record_Chunk(thePtr);}
    while (abit<size) {
      abit = abit<<1;
      chn++;}
    if (chn <= Num_Chunks) {
      Wait32(MoveOK);
      Chunk_Count[chn]++;
      Chunk_Tops[chn] = MAX(Chunk_Tops[chn],Chunk_Count[chn]);
      thePtr = Chunk_List[chn]; // non-raw pointer
      if (Chunk_Free[chn]>0 && thePtr != NULL) {
        if ((Chunx_Odom<Chunk_Freeze) || (Chunk_Freeze==0))
            if (chn >= Chunk_Edge) {
          Chunk_Acts[(Recent_Chunk++)&Chunk_Mast]
              = ((int)thePtr)&0xFFFFFFC | (chn<<28) | 1;                 // n..1
          if (whom !=0) Chunk_Acts[(Recent_Chunk++)&Chunk_Mast] = whom;
          Chunk_Acts[Recent_Chunk&Chunk_Mast] = -1;}
        Chunk_List[chn] = (int*)(*thePtr);
        Chunk_Free[chn]--;
        Signal32(MoveOK);
        if (Malloc_Log) {
          if (Malloc_Log==5) {thePtr--;  thePtr--;}
          thePtr--;  thePtr--;
          *thePtr++ = whom;
          *thePtr++ = ++Chunx_Odom;
          if (Malloc_Log==5) {*thePtr++ = 0; *thePtr++ = 0;}}}
      else {
        lnx = Recent_Chunk++;
        Recent_Chunk++;
        Signal32(MoveOK);
        thePtr = Record_Chunk((int*)malloc(abit<<2));
        if ((Chunx_Odom<Chunk_Freeze) || (Chunk_Freeze==0)) {
          Chunk_Acts[(lnx++)&Chunk_Mast]
              = ((int)thePtr)&0xFFFFFFC | (chn<<28);                     // n..0
          if (whom !=0) Chunk_Acts[(Recent_Chunk++)&Chunk_Mast] = whom;
          Chunk_Acts[lnx&Chunk_Mast] = -1;}
        if (Malloc_Log) {
          *thePtr = (int)Chunk_Linx[chn]; // other thread could race, so what?
          Chunk_Linx[chn] = thePtr++;
          *thePtr++ = whom;
          *thePtr++ = ++Chunx_Odom;
          if (Malloc_Log==5) {*thePtr++ = 0; *thePtr++ = 0;}}}
      if (fronx) {
        Fill_Block(0,--thePtr,abit-2); // unused part of block
        Fill_Block(0xDEADBEEF,thePtr,size-2); // bogus but recognizable values
        thePtr[size] = 0xBAD1DEA5; // sets the no-touchum fence
        thePtr[abit] = 0xBAD1DEA5; // another fence at physical end
        thePtr++;}
      *thePtr++ = Chunk_Tag+chn;
      if (fronx) {
        *thePtr++ = size;
        *thePtr++ = 0xBAD1DEA5;}}
    else if (_Crash_and_Burn_) abit = abit/(chn>>6); // div/0
      else throw abit;}
    catch (...) {
      if (_Crash_and_Burn_) abit = abit/(chn>>6); // div/0: I still want to see
      thePtr = NULL;}
  return thePtr;} //~my_malloc

boolean Die__Politely(int why, int* ary, int whom) { // get out & quit, if log..
  int rept = Chunx_Stop;
  if (Malloc_Log != 5) return false;
  Chunx_Stop = why;
  Chunk_Acts[(Recent_Chunk++)&Chunk_Mast]
      = ((int)ary)&0xFFFFFFC | 0xF0000000 | why;                     // F..1/2/3
  if (whom !=0) Chunk_Acts[(Recent_Chunk++)&Chunk_Mast] = whom;
  Chunk_Acts[Recent_Chunk&Chunk_Mast] = -1;
  Chunx_Odom++;
  if (rept==0) AsyncMsg("FAIL",why,whom,(int)ary,MainProgram);
  return true;}

void my_free(int* ary, int whom) { // does not look at user data space
  int tag = (int)(ary--), az = tag-tag;
  if ((tag&-32)==0) return; // huh??
  if ((tag&3) !=0) return; // huh??
  Net_Chunx--;
  if (fronx) {
    tag = *ary; // front fence, at [-1]
    *ary-- = whom; // record deletor there
    ary--; // now at actual block front
    if (tag != 0xBAD1DEA5) { // validate front fence
      if (Die__Politely(1,ary,whom)) return;                             // F..1
        else az = tag/az;} // div/0
    tag = ary[1]; // allocation size
    if (ary[tag-1] != 0xBAD1DEA5) { // back fence
      if (Die__Politely(2,ary,whom)) return;                             // F..2
        else az = tag/az;}} // div/0
  tag = (*ary)^Chunk_Tag; // if unchanged, this is chunk # in low byte
  whom = whom+0x87000000; // file ID,line#
  if ((tag&-256)==0) {
    if (tag >= Num_Chunks) {
      if (Malloc_Log) {
        if (Malloc_Log==5) {ary--;  ary--;}
        ary--;  ary--;  ary--;}
      if ((Chunx_Odom<Chunk_Freeze) || (Chunk_Freeze==0)) {
        Chunk_Acts[(Recent_Chunk++)&Chunk_Mast]
            = ((int)ary)&0xFFFFFFC | 0xF0000003;                         // F..3
        if (whom !=0) Chunk_Acts[(Recent_Chunk++)&Chunk_Mast] = whom;
        Chunk_Acts[Recent_Chunk&Chunk_Mast] = -1;
        Chunx_Odom++;}
      free(ary);
      return;}
    if (fronx) if (ary[(Min_Chunk<<tag)-1] != 0xBAD1DEA5) az = tag/az;
    Wait32(MoveOK);
    Chunk_Count[tag]--;
    if (Chunk_Free[tag]<Enuff_Chunx[tag]) {
      ary[2] = whom; // file ID,line#
      *ary = (int)Chunk_List[tag];
      Chunk_List[tag] = ary;
      if (Malloc_Log) {
        if (Malloc_Log==5) {ary--;  ary--;}
        ary--;  ary--;  ary--;}
      if ((Chunx_Odom<Chunk_Freeze) || (Chunk_Freeze==0)) if (tag>=Chunk_Edge) {
        Chunk_Acts[(Recent_Chunk++)&Chunk_Mast]
            = ((int)ary)&0xFFFFFFC | (tag<<28) | 2;                      // n..2
        if (whom !=0) Chunk_Acts[(Recent_Chunk++)&Chunk_Mast] = whom;
        Chunk_Acts[Recent_Chunk&Chunk_Mast] = -1;
        Chunx_Odom++;}
      Chunk_Free[tag]++;
      Chunk_Nops[tag] = MAX(Chunk_Nops[tag],Chunk_Free[tag]);
      Signal32(MoveOK);}
    else {
      if ((Chunx_Odom<Chunk_Freeze) || (Chunk_Freeze==0)) {
        Chunk_Acts[(Recent_Chunk++)&Chunk_Mast]
            = ((int)ary)&0xFFFFFFC | (tag<<28) | 3;                      // n..3
        if (whom !=0) Chunk_Acts[(Recent_Chunk++)&Chunk_Mast] = whom;
        Chunk_Acts[Recent_Chunk&Chunk_Mast] = -1;
        Chunx_Odom++;}
      Signal32(MoveOK);
      if (Malloc_Log) return; // shouldn't happen in test
      free(ary);}}
  else if (_Crash_and_Burn_) az = tag/az;} // div/0 //~my_free

String KorN(int n) {
  if (n<1024) return String(n)+String(":");
  return String(n>>10)+String("K:");} //~KorN

void Log_MOMAM() { // [only if lock: shft+1, ctrl+2, menu+4]
  int nx = Num_Chunks, ix; // = (Modifier_Keys>>1)&3; // =0??
  String aLine; // = String("-- ")+FormInt(ix)+String(" ");
  int* here; // = Chunk_Linx[1];
  int ChunkActs[Chunk_Hist];
  for (ix=0;ix<Chunk_Hist;ix++)
    ChunkActs[ix] = Chunk_Acts[(Recent_Chunk-Chunk_Hist+ix)&Chunk_Mast];
  if (Net_Chunx>666) nx = 1;
  for (nx=nx;nx>=0;nx--) {
    aLine = FormHexNum(nx,1)+String("+");
    here = Chunk_Linx[nx];
    while (here) {
      aLine = aLine+FormHexNum((int)here,8)+String(":");
      for (ix=1;ix<16;ix++) aLine = aLine+String(" ")+FormInt(here[ix]);
      SystemDebugLog(aLine);
      aLine = String("  ");
      here = (int*)(*here);} //~while
    if (Net_Chunx>666) break;} //~for
  aLine = String("++");
  for (ix=0;ix<Chunk_Hist;ix++) {
    aLine = aLine+String(" ")+FormHexNum(ChunkActs[ix],8);
    if ((ix&7)<7) continue;
    SystemDebugLog(aLine);
    aLine = String("  ");}} //~for //~Log_MOMAM

void Exit_MOMAM() {
  int ix, zx = 0, abit = Min_Chunk;
  String aLine = "*MOMAM +";
  aLine = aLine+String(Net_Chunx);
  for (ix=0;ix<Num_Chunks;ix++) {
    if (Chunk_Tops[ix]+Chunk_Nops[ix]>0)
      aLine = aLine+String(" ")+KorN(abit)+String(Chunk_Tops[ix])
          + String("/")+String(Chunk_Nops[ix]);
    abit = abit<<1;}
  SystemDebugLog(aLine);
  abit = Min_Chunk;
  aLine = "    ";
  for (ix=0;ix<Num_Chunks;ix++) {
    if (Chunk_Count[ix]+Chunk_Free[ix]>0)
      aLine = aLine+String(" ")+KorN(abit)+String(Chunk_Count[ix])
          + String("/")+String(Chunk_Free[ix]);
    abit = abit<<1;}
  SystemDebugLog(aLine);
  abit = 0;
  for (ix=Chunk_Bitz+3;ix>=0;ix--) {
    if (Chunk_Bits[ix] == 0) continue;
    if (abit==0) abit = ix|3;
    zx = ix&-4;} //~for
  aLine = " ";
  for (ix=abit;ix>=zx;ix--) {
    aLine = aLine+" "+FormHexNum(ix<<5,3)+"M: "+FormHexNum(Chunk_Bits[ix],8);
    if ((ix&3) !=0) continue;
    SystemDebugLog(aLine);
    aLine = " ";} //~for
  if (Malloc_Log) Log_MOMAM();} //~Exit_MOMAM

void Log_Res_Info(int rez, int rty, int FiNo) {
  int* aptr = (int*)(rez-16);
  if (Malloc_Log<5) return;
  *aptr-- = FiNo;
  *aptr = rty;}

#undef fronx
#undef Chunk_Tag
#undef Num_Chunks
#undef Min_Chunk
#undef Max_Chunk

#if _String_Log_>1 // enable string log for unit tests..
void Log_Str_Cash(int whom, bool doWhat);
#define _Str_Logit_(whom) Log_Str_Cash(whom,true)
#define _Str_unLog_(whom) Log_Str_Cash(whom,false)
#else
#define _Str_Logit_(whom)
#define _Str_unLog_(whom)
#endif

int _Strings_Count_ = -1; // initially disabled
int the_String_Log[_String_Log_];

 } // (temply) end namespace  _FrameWork_

// array/pointer management routines..

int* my_int_ary(int lnx, int whom) { // int* new_int_ary(int lnx) {
  int* ptr = (int*)_FrameWork_::my_malloc((lnx+2)*sizeof(int),whom);
  if (ptr==NULL) return NULL;
  _FrameWork_::FILL0(ptr,0,lnx);
  *ptr++ = lnx;
  *ptr-- = 1; // always precounted
  return ptr;}

void set_ary_size(void* aptr, int siz) { // siz<0 for static alloc (uncounted)
  int isiz = ABS(siz); // data size in int's (not including size,ref-count)
  int* ptr = (int*)aptr;
  _FrameWork_::FILL0(ptr,0,isiz);
  *ptr++ = isiz;
  if (siz<0) *ptr = -256;
    else *ptr = 1;}

boolean Valid_Ptr_Verify(void* ptr) { // check for memory corruption often..
#ifdef PEEK_OFTEN
  int* aptr = (int*)PEEK_OFTEN;
  if (aptr) if (*(--aptr) != 0xBAD1DEA5) if (_FrameWork_::Chunx_Stop==0)
    if (!_FrameWork_::Die__Politely(3,(int*)(++aptr),_BASE_+__LINE__))   // F..3
      aptr = NULL;
  aptr = (int*)ptr;
#else
  int* aptr = (int*)ptr;
#endif
  --aptr;
  if (*aptr-- == 0xBAD1DEA5) {
    if (aptr[(*aptr)-2] == 0xBAD1DEA5) return false;}
  else if (*aptr-- == 0xBAD1DEA5) // Text_Ary_Str ptr is offset +1..
    if (aptr[(*aptr)-2] == 0xBAD1DEA5) return false;
  if (!_FrameWork_::Die__Politely(3,(int*)ptr,_BASE_+__LINE__))          // F..3
    aptr = NULL;
  return true;} // retn true if failed

#if Malloc_Log == 5
#define Verify_Valid_Ptr(ptr) Valid_Ptr_Verify(ptr)
#else
#define Verify_Valid_Ptr(ptr)
#endif

void Stop_n_Look(int* ptr) { // can set breakpoint here, cuz shouldn't happen..
  ptr--;}

int* up__ptr(void* aptr) {
  int* ptr = (int*)aptr;
  if ((((int)ptr)&-32) !=0) if (*(++ptr) >= 0) { // not if stack alloc
    Verify_Valid_Ptr(aptr);
    if ((*ptr)==12) Stop_n_Look(ptr);
    (*ptr)++;}
  return (int*)aptr;}

int* dn__ptr(void* aptr, int whom) { // normally called as void, except in expn
  int* ptr = (int*)aptr;
  if ((((int)ptr)&-32)==0) return (int*)0; // huh?
  if (*(++ptr)<0) return (int*)aptr; // stack allocated: do nothing
  Verify_Valid_Ptr(aptr);
  if (*ptr) (*ptr)--; // could have 2 threads looking at same ptr if res,
  if ((*ptr)==0)             // but cache holds a copy, so never hits 0
    _FrameWork_::my_free(--ptr,whom);
  return (int*)aptr;}

////////////////////////////////////////////////////////////////////////

// 4-character-as-int (Char4) class, intrinsic to Turk/2..

// class String; // forward ref in Char4

Char4::Char4() {cccc = 0;}
Char4::Char4(int x) {cccc = x;}
bool Char4::operator<(Char4 st) {return (bool)(this->cccc < st.cccc);}
bool Char4::operator>(Char4 st) {return (bool)(this->cccc > st.cccc);}
bool Char4::operator<=(Char4 st) {return (bool)(this->cccc <= st.cccc);}
bool Char4::operator>=(Char4 st) {return (bool)(this->cccc >= st.cccc);}
bool Char4::operator==(Char4 st) {return (bool)(this->cccc == st.cccc);}
bool Char4::operator!=(Char4 st) {return (bool)(this->cccc != st.cccc);}
Char4::Char4(const char* st) {
  int ix = 0, ch;
  cccc = 0;
  if (st) while (ix<4) {
    ch = ((int)st[ix++])&255;
    if (ch<32) ch = ch+32;
    cccc = (cccc<<8)+ch;}}
Char4 Char4::operator=(const int x) {
  *this = Char4(x);
  return *this;}
Char4 Char4::operator=(const char* st) {
  *this = Char4(st);
  return *this;}
bool Char4::operator<(const char* st) {
  Char4 x(st);
  return (bool)(*this < x);}
bool Char4::operator>(const char* st) {
  Char4 x(st);
  return (bool)(*this > x);}
bool Char4::operator<=(const char* st) {
  Char4 x(st);
  return (bool)(*this <= x);}
bool Char4::operator>=(const char* st) {
  Char4 x(st);
  return (bool)(*this >= x);}
bool Char4::operator==(const char* st) {
  Char4 x(st);
  return (bool)(*this == x);}
bool Char4::operator!=(const char* st) {
  Char4 x(st);
  return (bool)(*this != x);}
Char4 CHR4(int x) {return Char4(x);}
int ORD(Char4 x) {return x.cccc;}

// Char4 Null4 = CHR4(0);
Char4 NoName = "    ";

Char4 Up4Char(Char4 Num4) { // increment Num4 decimally
  int pnam = ORD(Num4), digit = pnam&127;
  if (digit <= ORD('0')) return CHR4((pnam&-256)+ORD('1'));
  if (digit>ORD('9')) return CHR4((pnam&-256)+ORD('1'));
  if (digit<ORD('9')) return CHR4(pnam+1);
  digit = (pnam>>8)&127;
  if (digit <= ORD('0')) return CHR4((pnam&~0xFFFF)|ORD("  10")&0xFFFF);
  if (digit>ORD('9')) return CHR4((pnam&~0xFFFF)|ORD("  10")&0xFFFF);
  if (digit<ORD('9')) return CHR4(pnam+247);
  digit = (pnam>>16)&127;
  if (digit <= ORD('0')) return CHR4((pnam&-0x1000000)|ORD(" 100")&0xFFFFFF);
  if (digit>ORD('9')) return CHR4((pnam&-0x1000000)|ORD(" 100")&0xFFFFFF);
  if (digit<ORD('9')) return CHR4(pnam+65536-0x909);
  return CHR4((pnam&-0x1000000)+(ORD("1000")&0x1FFFFFF));}

boolean Is4chars(int val) {
  if ((val&0x80808080) !=0) return false; // disallows option chars
  if (((val+0x01010101)&0x80808080) !=0) return false; // disallows del
  return (((val+0x60606060)&0x80808080) == 0x80808080);} // disallow ctls

////////////////////////////////////////////////////////////////////////

// Global declarations used in WinMain/InitializeAl..

#define Cast_Proc(p) p
#define Proc_Cast(p) p

namespace  _FrameWork_ { // resume name space protection..

Char4 DocX4 = "DocX", Imag4 = "Imag", PrgB4 = "PrgB", TxtD4 = "TxtD";
Char4 Next_ProcID = "@000", Typ4 = "Type";

// Process MainProgram = NULL, Ready_List = NULL; // was: AllProgP
HINSTANCE Main_Context = NULL; // context passed to main from Win32
HWND GooeyWin = NULL, HasSplash = NULL; // something to send messages to
HWND FrontedWin = NULL; // force this to the front if not NULL
HWND theFrontWin = NULL; // this win is on top (sometimes we care)
HBRUSH StockWhiteBrush;  HBRUSH StockBlackPen;
HBRUSH PreProgBrush;     HBRUSH ProgressBrush;
HCURSOR HourGlass = NULL; HCURSOR OldCursor = NULL; // used in win redraw delays
HCURSOR CrossHair = NULL; HCURSOR Ibeam = NULL; HCURSOR PlainArrow = NULL;
HCURSOR EastWest = NULL; HCURSOR NorthSouth = NULL; HCURSOR DiagonalAro = NULL;
int my_Cursor = 0, Tenth_Seconds = 0; // used to blink caret
unsigned char eTrsCan[268], FullCan[268]; 
boolean Drag_Is_Busy = false;

struct {
  BITMAPINFOHEADER BitMapStuff;
  int ColorTbl[216];} MyBitMap; // RGBQUAD
BITMAPINFO* theBitMap = (BITMAPINFO*)(&MyBitMap);

int ScreenTall, ScreenWide; // used in BitMapStuff and for pixelating
unsigned char* theScreen = NULL; // 8 bytes of IntArray o'hed at front

WNDCLASS Wind_type;

int FlushClix = 0; // >0 to flush spurious clicks from file dlg
Point ClikPt, ScroRefLoc; // used to reconstruct over-range scroll thumb
Point Mouse_Locn = Nowhere; // captured from recent WM_MOUSEMOVE
bool Fresh_Mods = false; // true if might be stale
bool New_MsgBx = false; // true when open MsgBx, to set FrontedWin
bool Menu_Is_Open = false; // to override mouse in win cursor display
Char4 Next_Win_ID = "#000";
int Window_Count = 0; // quit when it reaches 0 again // Popup_Result = 0,

// Some timers..
int Tix_Now = 0, Tix_100ms = 0; // <0 while splash win is visible
int n_Timers = 0; // number of processes in Delay
int Zap_Time = 0; // zap drag when decrements to 0 (ref'd but unused)

// early forward declarations..

void Down_Win_Cnt(void);
String FormTimeMS(void); // in FrameWork.h

Window PIB_wins = NULL, Free_Wins = NULL; // master window list, closed spares
Window Progress_o = NULL; // cached progress bar win, cuz deleting it bombs
Window Pacify_wins = NULL; // window(s) being redrawn
Window Cursor_Win = NULL; // last known window showing cursor

#define Menu_LineHeight 20
#define Menu_Comm_base 3333
#define Max_Menus 256
// struct Menu_Com_ {Process whose; Char4 say;};
Char4 Menu_Comms[Max_Menus];
int Num_Menus = 0;
// #define BaseM_LineHgt 17 // ad hoc as observed, to be adjusted (NOT)
// int Menu_Hi_Chek = 2222, Menu_Preselect = -1, Menu_LineHeight = BaseM_LineHgt;

#if Panic_Log

int the_Panic[Panic_Log];
int Log_at = 0;

void Panic_See(int whom) {
  Wait32(PanicLok);
  the_Panic[Log_at&(Panic_Log-1)] = whom;
  Log_at++;
  Signal32(PanicLok);}

void Log_Panic(String aLine) {if (DebugLogFile) {
  int here;
  SystemDebugLog(aLine+" "+FormHexNum(_FrameWork_::SemRefHisto,8));
  Wait32(PanicLok);
  here = MAX(Log_at-Panic_Log,0);
  while (here<Log_at) {
    if (((Log_at-here)&7)==0) aLine = "  ";
    aLine = aLine + " " + FormInt(the_Panic[here&(Panic_Log-1)]);
    here++;
    if (((Log_at-here)&7)==0) SystemDebugLog(aLine);}
  Signal32(PanicLok);}} //~Log_Panic

#define See_Panic(x) Panic_See(x)
#else
void Log_Panic(String aLine) {}
#define See_Panic(x)
#endif // Panic_Log

 } // (temply) end namespace  _FrameWork_

////////////////////////////////////////////////////////////////////////

// ref-counted String class, intrinsic to Turk/2..

//  String is a 1-word reference to the string value, =0 is ""
//  String compare in T2 is not case-sensitive in the ASCII range

namespace  _FrameWork_ { // private library functions needed for String..

int FindText(char* xPtr, int xLen, char* aPtr, int aLen) { // find xPtr in aPtr
  int ix, offs = 0, xCh;
  while (aLen >= xLen) {
    xCh = 0;
    for (ix=0;ix<xLen;ix++) {
      xCh = ((int)xPtr[ix])&255;
      if (xCh==0) break; // end of lookfor, must be =
      xCh = (((int)Capitize(aPtr[offs+ix]))&255)-((int)Capitize((char)xCh));
      if (xCh !=0) break;} // not equal here
    if (xCh==0) return offs;
    offs++;
    aLen--;}
  return -1;} // ~FindText

int CompareText(char* xPtr, int xLen, const char* aPtr, int aLen) { // xPtr:aPtr
  int ix, offs = MIN(xLen,aLen), xCh = 0;
  for (ix=0;ix<offs;ix++) {
    xCh = ((int)aPtr[ix])&255;
    if (xCh==0) return ((int)xPtr[ix])&255;
    xCh = (((int)Capitize(xPtr[ix]))&255)-((int)Capitize((char)xCh));
    if (xCh !=0) return xCh;} // not equal
  if (xLen>offs) return ((int)xPtr[ix])&255;
  if (aLen>offs) return -(((int)aPtr[ix])&255);
  return 0;} // ~CompareText

char* AllocStr(int len) {
  int nby = (len+16)&-4;
  int* ipt = my_malloc(nby,0);
  if (ipt==NULL) return NULL;
  *ipt = len; // string len
  ipt[1] = -1; // temp ref count
  _Str_Logit_((int)ipt); // ref-count in wrong place
  nby = (nby>>2)-3;
  *(++ipt) = nby; // int array size
  *(++ipt) = 1; // string ref count
  *(++ipt) = 0; // initialize null
  ipt[--nby] = 0;
  if (nby) ipt[--nby] = 0;
  return (char*)ipt;} //~AllocStr

char* NewStrText(const char* st) {
  char* sptr = NULL;
  if (st) if (*st) {
    sptr = AllocStr(strlen(st));
    if (sptr) strcpy(sptr,st);}
  return sptr;} //~NewStrText

char* UpDownStr(char* sptr, int upby) {
  int* ipt = (int*)sptr;
  if (((int)ipt)&-32) {
    ipt--;
    Wait32(StrnLok);
    ipt[0] = ipt[0]+upby;
    Signal32(StrnLok);
    if (*(ipt--)>0) return sptr; // no other threads have a ref if =0
    upby = *(ipt--); // int length, now "ref-count"
    _Str_unLog_((int)ipt);
    // ipt[0] = ++upby; // corrected int length (nobody looks)
    my_free(ipt,_BASE_+__LINE__);}
  return NULL;} //~UpDownStr

void SetActualStrLen(char* sptr) {
  int* ipt = (int*)sptr;
  if (((int)ipt)&-32) {
    ipt--;  ipt--;  ipt--;
    ipt[0] = strlen(sptr);}} //~SetActualStrLen

 } // (temporarily) end namespace  _FrameWork_

String::String() {
  stp = NULL;}
String::String(const char* st) {
  stp = _FrameWork_::NewStrText(st);}
String::String(const char* st, int len) {
  stp = _FrameWork_::AllocStr(len);
  if (stp==NULL) return;
  if (st) {
    while (len-- >0) stp[len] = st[len];
    _FrameWork_::SetActualStrLen(stp);}}
String::String(const String& st) {
  stp = _FrameWork_::UpDownStr(st.stp,1);}
String::String(int i) {
  int dg = -1, ix = 0;
  char cxi[16], cxs[16];
  char* cp = (char*)(14+cxi);
  if (i<0) {
    cxs[ix++] = '-';
    i = -i;
    if (i<0) {
      cxs[ix++] = '2';
      i = i-2000000000;}
    dg = 0;}
  cxs[ix] = '\0';
  *cp = '\0';
  while (i>0) {
    dg = i%10;
    i = (i-dg)/10;
    *(--cp) = (char)(dg+48);}
  if (dg<0) *(--cp) = '0';
  strcat(cxs,cp);
  stp = _FrameWork_::NewStrText(cxs);}
String::String(char c) {
  char cxc[2];
  cxc[0] = c;
  cxc[1] = '\0';
  stp = _FrameWork_::NewStrText(cxc);}
String::String(Char4 x) {
  int ix = ORD(x);
  char cxc[5];
  cxc[0] = (char)MAX((ix>>24)&127,32);
  cxc[1] = (char)MAX((ix>>16)&127,32);
  cxc[2] = (char)MAX((ix>>8)&127,32);
  cxc[3] = (char)MAX(ix&127,32);
  cxc[4] = '\0';
  stp = _FrameWork_::NewStrText(cxc);}
String::String(bool b) {
  char* cp;
  if (b) cp = "true";
    else cp = "false";
  stp = _FrameWork_::NewStrText(cp);}
String String::operator=(const char* st) {
  _FrameWork_::UpDownStr(stp,-1);
  stp = _FrameWork_::NewStrText(st);
  return *this;}
String String::operator=(String st) {
  if (stp==st.stp) return *this;
  _FrameWork_::UpDownStr(stp,-1);
  stp = _FrameWork_::UpDownStr(st.stp,1);
  return *this;}
String String::operator=(char c) {
  char cxc[2];
  cxc[0] = c;
  cxc[1] = '\0';
  _FrameWork_::UpDownStr(stp,-1);
  stp = _FrameWork_::NewStrText(cxc);
  return *this;}
String String::operator=(int i) {
  String st = i;
  _FrameWork_::UpDownStr(stp,-1);
  stp = _FrameWork_::UpDownStr(st.stp,1);
  return *this;}
String String::operator+(const char* st) {
  int sz = 0, len = this->length_();
  String res;
  if (st) sz = strlen(st);
  if (sz==0) return *this;
  if (len==0) {
    res.stp = _FrameWork_::NewStrText(st);
    return res;}
  sz = this->length_()+sz;
  if (sz==0) return Null_Str;
  res.stp = _FrameWork_::AllocStr(sz);
  if (res.stp==NULL) return *this;
  strcpy(res.stp,this->stp);
  strcat(res.stp,st);
  return res;}
String String::operator+(String st) {
  int sz = st.length_(), len = this->length_();
  String res;
  if (sz==0) return *this;
  if (st.stp==NULL) return *this; // shoulda: sz=0
  if (len==0) return st;
  sz = sz+len;
  res.stp = _FrameWork_::AllocStr(sz);
  strcpy(res.stp,this->stp);
  strcat(res.stp,st.stp);
  return res;}
String String::operator+(int i) {
  String res, st = i;
  int sz = this->length_();
  if (sz==0) return st;
  res.stp = _FrameWork_::AllocStr(st.length_()+sz);
  if (res.stp) {
    strcpy(res.stp,this->stp);
    if (st.stp) strcat(res.stp,st.stp);}
  return res;}
String::~String() { // we have a problem with near-null pointers..
  stp = _FrameWork_::UpDownStr(stp,-1);}

void String::valid_() {_FrameWork_::SetActualStrLen(stp);}
int String::length_() {
  int* ipt = (int*)stp;
  if (((int)ipt)&-32) {
    ipt--;  ipt--;  ipt--;
    return *ipt;}
  return 0;}
char* String::text_() {
  if (stp==NULL) return "";
  return stp;}
int String::cmpr_(const char* st) {
  int len = this->length_();
  if (len==0) {
    if (st==NULL) return 0;
    if (*st == '\0') return 0;
    return -1;}
  if (st==NULL) return 1;
  if (*st == '\0') return 1;
  return _FrameWork_::CompareText(stp,len,st,~HiBit);}
int String::cmpr_(String st) {
  int len = this->length_(), sz = st.length_();
  if (len==0) {
    if (sz==0) return 0;
    return -1;}
  if (sz==0) return 1;
  return _FrameWork_::CompareText(this->stp,len,st.stp,sz);}

bool String::operator<(const char* st) {return (bool)(this->cmpr_(st) < 0);}
bool String::operator>(const char* st) {return (bool)(this->cmpr_(st) > 0);}
bool String::operator<=(const char* st) {return (bool)(this->cmpr_(st) <= 0);}
bool String::operator>=(const char* st) {return (bool)(this->cmpr_(st) >= 0);}
bool String::operator==(const char* st) {return (bool)(this->cmpr_(st) == 0);}
bool String::operator!=(const char* st) {return (bool)(this->cmpr_(st) != 0);}
bool String::operator<(String st) {return (bool)(this->cmpr_(st) < 0);}
bool String::operator>(String st) {return (bool)(this->cmpr_(st) > 0);}
bool String::operator<=(String st) {return (bool)(this->cmpr_(st) <= 0);}
bool String::operator>=(String st) {return (bool)(this->cmpr_(st) >= 0);}
bool String::operator==(String st) {return (bool)(this->cmpr_(st) == 0);}
bool String::operator!=(String st) {return (bool)(this->cmpr_(st) != 0);}

String GetPstr(int* ary) { // P-string operators: no memory allocation..
  if (ary==NULL) return Null_Str;
  ary++; // PC is Little-Endian, but no matter..
  char* cptr = (char*)(++ary);
  int stl = ((int)(*cptr++))&255;
  return String(cptr,stl);} // not a cast, so OK
void AppPstr(int* ary, const char* st) {
  if (ary==NULL) return;
  int alx, aln = (*ary++)*4-2;
  if ((*ary)>0) (*ary)--; // should be static, but just in case
  int asl = (*(++ary))&255; // Little-Endian
  char* cptr = (char*)ary;
  if (st==NULL) return;
  if (*st == '\0') return;
  alx = asl+strlen(st);
  if (alx>aln) return;
  if (alx>254) return;
  cptr[0] = (char)alx;
  cptr[++asl] = '\0';
  strcat(cptr,st);}
void AppPstr(int* ary, String st) {
  if (st.length_()>0) AppPstr(ary,st.text_());}
void AppChar(int* ary, char chx) {
  if (ary==NULL) return;
  int aln = (*ary++)*4-2;
  if ((*ary)>0) (*ary)--;
  int asl = ((*(++ary))&255)+1; // Little-Endian
  char* cptr = (char*)ary;
  if (asl>aln) return;
  if (asl>254) return;
  cptr[0] = (char)asl;
  cptr[asl++] = chx;
  cptr[asl] = '\0';}

// char* TextOf(String st) {return st.text_();} // unused

// String Manager Operations..
// int LENGTH(String st) {return st.length_();} // in SysNames.h

Char4 Chr4Str(String st) {
  if (st.length_() == 0) return CHR4(0);
  if (st.length_() < 4) st = st+String("    ");
  return Char4(st.text_());}

char CharAt(int i, String st) {
  char* here = st.text_();
  if (i<0) return '\0';
  if (st.length_() <= i) return '\0';
  return here[i];}

String Substring(int offs, int len, String st) {
  String nu;
  if (offs<0) return Null_Str;
  len = MIN(st.length_()-offs,len);
  if (len<=0) return Null_Str;
  nu = String((char*)(st.text_()+offs),len);
  return nu;}

String TxtBlok2Str(int offs, int nch, int* ary) { // nch<0 omits copy
  // PRE: offs >=0, nch >=0, offs+nch <= ary.len  // nch=0 takes all
  if (ary==NULL) return Null_Str;
  if (offs<0) offs = 0;
  char* chad = (char*)(((char*)ary)+offs+sizeof(int)*2);
  offs = (*ary)*sizeof(int)-offs;
  if (offs>0) {
    if (nch<0) { // don't copy (nor even look at) array contents..
      offs = MIN(offs,-nch);
      chad = NULL;}
    else {
      offs = MIN((int)strlen(chad),offs);
      if (nch>0) offs = MIN(offs,nch);}}
  if (offs <= 0) {
    if (ary[1]>0) ary[1]--;
    return Null_Str;}
  String res(chad,offs);
  if (ary[1]>0) ary[1]--;
  if (nch >= 0) return res;
  if (LENGTH(res)==0) return Null_Str;
  return res;}

void Str2TxtBlok(String st, int offs, int mxln, int* ary) {
  // PRE: offs >=0, offs+MIN(st.len,mxln-1) < ary.len (both checked)
  if (ary==NULL) return;
  char* chad = (char*)(((char*)ary)+offs+sizeof(int)*2);
  int nby = (*ary)*sizeof(int)-offs;
  if (offs >= 0) if (nby>0) while (true) { // funnel all exits thru down-ref..
    if (mxln>0) nby = MIN(nby,mxln); // mxln>0 truncates or fills to mxln
    if (st.length_() >= nby) {
      _FrameWork_::copy_bytes(st.text_(),chad,nby);
      break;}
    strcpy(chad,st.text_());
    if (mxln==0) break; // mxln<0 fills to end of array with nulls
    nby = nby-1-st.length_(); // if length is given, fill it out with nulls..
    chad = (char*)(chad+st.length_());
    if (nby>0) _FrameWork_::copy_bytes(chad,(char*)(chad+1),nby);
    break;}
  if (ary[1]>0) ary[1]--;}

int FindStTxBlk(String st, int offs, int mxln, int* ary) {
  // PRE: offs >=0, offs < MIN(ary.len,mxln)
  if (ary==NULL) return -1;
  int aoff = ABS(offs), nby = (*ary)*sizeof(int)-aoff;
  char* here = (char*)(((char*)ary)+aoff+sizeof(int)*2);
  if (mxln>0) nby = MIN(nby,mxln);
  mxln = st.length_();
  if (mxln==0) {
    if (ary[1]>0) ary[1]--;
    if (offs<0) if (nby>0) return -1;
    return 0;}
  if (nby <= 0) nby = -1;
  else if (offs >= 0) {
    nby = _FrameWork_::FindText(st.text_(),mxln,here,nby);
    if (nby<0) nby = -1;
      else nby = nby+aoff;}           // special case, must =
  else nby = -ABS(_FrameWork_::CompareText(st.text_(),mxln,here,nby));
  if (ary[1]>0) ary[1]--;
  return nby;}


int* Text_Ary_Str(String st) {
  if (st.length_()==0) return NULL;
  return (int*)up__ptr((int*)(st.text_()-8));}

// Turk/2 compiled code access..
#define C_string char*
#define EXTRACT_Cstr(s) s.text_()
#define MAKE_Cstr(s) String(s)
#define Literal_Cstr(s) s

////////////////////////////////////////////////////////////////////////

// Opaque File type, implemented as a pointer to a _File_ class..

namespace  _FrameWork_ { // resume

int FilnIx = 511;
bool IsValidFile(File aFile); // fwd, defined in FrameWork.t2..
File Open_In_Dir(String name, int opt, File dir);

File PIBfiles = NULL, PIBwDir = NULL;
File Using_ResFi = NULL; // used in System.h, but checked by Close2Ref

File Make_File(String nam, String pth, WinFile wfi, int flg) {
  File aFile;
  if (PIBfiles==NULL) if (flg != Directory) return NULL;
  try {aFile = new _File_(pth,wfi,flg);} catch (...) {aFile = NULL;}
  return aFile;}

File Next_File(File aFile) { // PRE: aFile is valid or null
  if (aFile==NULL) return PIBfiles;
  return aFile->GetLink();}

WinFile Get_File_w32(File aFile) {return aFile->GetWn32_File();} // valid aFile

void Set_File_Flagz(File aFile, int flagz) {
  if (IsValidFile(aFile)) aFile->SetFlagz(flagz);}

int Write_Data(File aFile, char* ary, int nby) {
  // write nby bytes of data from ary to open file fi, return actual # written
  int bitz = aFile->GetFlagz(); // =0 if invalid
  if (bitz==0) return 0;
  if (nby>0) if ((bitz&FlagMask) != 0) if ((bitz&DirecOnly) != Directory)
    if (((bitz&(MakeFiRef+Append))^WriteOnly) < Append)
      return (int)fwrite(ary,1,nby,aFile->GetWn32_File());
  return 0;} //~Write_Data

int Read_Data(File aFile, char* ary, int nby) {
  // read up to nby bytes of data into ary from open file fi, return actual #
  int bitz = aFile->GetFlagz(); // =0 if invalid
  if (bitz==0) return 0;
  if (nby>0) if ((bitz&FlagMask) != 0) if ((bitz&DirecOnly) != Directory)
    if ((bitz&(MakeFiRef+WriteOnly)) ==0)
      return (int)fread(ary,1,nby,aFile->GetWn32_File());
  return 0;} //~Read_Data

void Win32FindClose(File aFile) {
  if (aFile->GetFiHand() != null) FindClose(aFile->GetFiHand());
  aFile->SetFiHand(NULL);}

File OpenFileDialog(String Prompt, File iDir, String finam, String filter,
      String ext3, bool newly) {
  OPENFILENAME theData;
  DWORD bitz;
  char flt[256];
  char* fixx = flt;
  char cbf[PathSize+1];
  String Dir_Path = Null_Str;
  if (iDir != NULL) Dir_Path = iDir->myFullPath();
  strcpy(cbf,finam.text_());
  strcpy(flt,filter.text_());
  while (true) {
    if (*fixx == '\t') *fixx++ = '\0';
    else if (*fixx++ == '\0') break;}
  theData.lStructSize = sizeof(OPENFILENAME);
  theData.hwndOwner = theFrontWin;
  theData.hInstance = NULL;
  theData.lpstrFilter = flt;
  theData.lpstrCustomFilter = NULL;
  theData.nMaxCustFilter = 0;
  theData.nFilterIndex = 0;
  theData.lpstrFile = cbf;
  theData.nMaxFile = PathSize;
  theData.lpstrFileTitle = NULL;
  theData.nMaxFileTitle = 0;
  if (Dir_Path.length_()==0) theData.lpstrInitialDir = NULL;
    else theData.lpstrInitialDir = Dir_Path.text_();
  theData.lpstrTitle = Prompt.text_();
  theData.nFileOffset = 0;
  theData.nFileExtension = 0;
  theData.lpstrDefExt = ext3.text_();
  theData.lCustData = 0;
  theData.lpfnHook = NULL;
  theData.lpTemplateName = NULL;
  if (newly) {
    theData.Flags = OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY
        | OFN_NOREADONLYRETURN;
    newly = (GetSaveFileName(&theData) !=0);}
  else {
    theData.Flags = OFN_FILEMUSTEXIST;
    newly = (GetOpenFileName(&theData) !=0);}
  Fresh_Mods = true;
  if (!newly) {
    if (DebugLogFile) SystemDebugLog(String("OpFiDlg '")+Prompt+"' -");
    return NULL;}
  if (DebugLogFile) SystemDebugLog(String("OpFiDlg '")+Prompt+"' = '"+cbf+"'");
  bitz = GetFileAttributes(cbf);
  if (bitz<0) return NULL; // no such file
  Dir_Path = cbf;
  if (bitz&FILE_ATTRIBUTE_DIRECTORY)
    return Open_In_Dir(Dir_Path,Directory+MakeFiRef,NULL);
  else return Open_In_Dir(Dir_Path,MakeFiRef,NULL);} // ~OpenFileDialog

 } // (temporarily) end namespace  _FrameWork_

#define Zap_File_w32(aFile) aFile->File_w32_Zap()

_File_::~_File_() { // Note: #define File_Destructor(f) delete f
  File lnk, prior = NULL;
  if (DebugLogFile) if (DebugLogFile != this)
    SystemDebugLog(String("FiDestruct -")+String(FileRefNum));
  if (Find_Hand != NULL) FindClose(Find_Hand);
  Find_Hand = NULL;
  Wait32(_FrameWork_::ResFiLok);
  lnk = _FrameWork_::PIBfiles;
  while (lnk != NULL) {
    if (lnk==this) {
      if (prior == NULL) _FrameWork_::PIBfiles = lnk->FileLink;
        else prior->FileLink = lnk->FileLink;
      break;}
    prior = lnk;
    lnk = lnk->FileLink;}
  Signal32(_FrameWork_::ResFiLok);}

_File_::_File_() {
  FileRefNum = ++_FrameWork_::FilnIx;
  Find_Hand = NULL;
  FileLink = NULL;
  FileInodN = NULL;
  ParDirLnk = NULL;
  FiFlagCnt = 0;
  FilePosnO = 0;}

_File_::_File_(String pth, WinFile wfi, int flg) {
  // _File_();
  FileRefNum = ++_FrameWork_::FilnIx;
  Find_Hand = NULL;
  FileLink = NULL;
  ParDirLnk = NULL;
  FilePosnO = 0;
  Wait32(_FrameWork_::ResFiLok);
  if (_FrameWork_::PIBfiles != NULL) {
    FileLink = _FrameWork_::PIBfiles->FileLink;
    _FrameWork_::PIBfiles->FileLink = this;}
  else _FrameWork_::PIBfiles = this;
  Signal32(_FrameWork_::ResFiLok);
  FileInodN = wfi;
  FiFlagCnt = flg&FlagMask;
  DirPath = pth;
  FilNamLnk = LastItem(pth, '\\');}

int GetFileFlags(File aFile) { // see bits above
  if (!_FrameWork_::IsValidFile(aFile)) return 0;
  return aFile->GetFlagz();}
String GetFileName(File aFile) {
  if (!_FrameWork_::IsValidFile(aFile)) return Null_Str;
  return aFile->myFileName();}
String FullPath(File aFile) {
  if (!_FrameWork_::IsValidFile(aFile)) return Null_Str;
  return aFile->myFullPath();}
int GetFiRefNum(File aFile) {
  if (!_FrameWork_::IsValidFile(aFile)) return 0;
  return aFile->GetRefNum();}

void SystemDebugLog(String theSt) {if (DebugLogFile) { // adds \r\n at end
  char* here = theSt.text_();
  char* crlf = CR_LF; // char* crlf = "\r\n";
  int ix = theSt.length_();
  // if (DebugLogFile) {
  // Wait32(_FrameWork_::DebugLok);
  _FrameWork_::Write_Data(DebugLogFile,here,ix);
  _FrameWork_::Write_Data(DebugLogFile,crlf,CRLF_LEN);
  // Signal32(_FrameWork_::DebugLok);
  }} //~SystemDebugLog
    // DebugLogFile->flush_(); // incredibly slow!!

int ReadData(File aFile, int* ary) { // returns a byte count
  if (ary==NULL) return 0;
  int nby = *ary++;
  if ((*ary)>0) (*ary)--; // doesn't dispose, but shouldn't need to
  return _FrameWork_::Read_Data(aFile,(char*)++ary,nby<<2);}
String ReadText(File aFile, int nby) { // read nby bytes of text, dele \0,\r
  int ix = 0, sz = FiSize_If_0(aFile,nby); // nby=0 gets all; never: sz<0
  if (sz==0) return Null_Str;
  String res(NULL,sz);
  char* here = res.text_();
  char ch;
  sz = _FrameWork_::Read_Data(aFile,here,sz);
  if (sz==0) return Null_Str;
  nby = 0;
  while (ix<sz) {
    ch = here[ix++];
    if (ch == '\r') {
      if (here[ix] == '\n') ix++;
      here[nby++] = '\n';}
    else if (ch != '\0') here[nby++] = ch;}
  if (nby==res.length_()) return res;
  return String(here,nby);}
String ReadLine(File aFile) { // read text to end of line
  int nby = 0;
  char ch, bf[260];
  // Fi_Find_32 FindInfo;
  String stx;
  if (!_FrameWork_::IsValidFile(aFile)) return Null_Str;
  // if (aFile->GetFiHand() != NULL) { // directory search already open..
  //   if (FindNextFile(aFile->GetFiHand(),&FindInfo))
  //     return String(FindInfo.cFileName);}
  // else if ((aFile->GetFlagz()&DirecOnly) == Directory) {
  //   stx = aFile->myFullPath()+"\\*";
  //   aFile->SetFiHand(FindFirstFile(stx.text_(),&FindInfo));
  //   if (aFile->GetFiHand() == INVALID_HANDLE_VALUE) aFile->SetFiHand(NULL);
  //     else return String(FindInfo.cFileName);}
  // else while (_FrameWork_::Read_Data(aFile,&ch,1)>0) {
  while (_FrameWork_::Read_Data(aFile,&ch,1)>0) {
    if (ch == '\0') break;
    if (ch == '\r') break;
    if (ch == '\n') {
      if (nby==0) continue;
      break;}
    bf[nby++] = ch;
    if (nby>255) break;}
  if (nby==0) return Null_Str;
  return String(bf,nby);} // ~ReadLine

int WriteData(File aFile, int* ary) { // returns a byte count
  if (ary==NULL) return 0;
  int nby = *ary++;
  if (nby>0) {
    nby = _FrameWork_::Write_Data(aFile,(char*)++ary,nby<<2);
    --ary;}
  down__ptr(--ary); // dispose if done
  return nby;}
void WriteBytes(File aFile, int* ary, int bgn, int ln) {
  if (ary==NULL) return;
  int nby = MAX(MIN(((*ary++)<<2)-bgn,ln),0);
  if (nby>0) {
    nby = _FrameWork_::Write_Data(aFile,(char*)(((int)(++ary))+bgn),nby<<2);
    --ary;}
  down__ptr(--ary);} // dispose if done
void WriteChar(File aFile, char ch) {
  _FrameWork_::Write_Data(aFile,&ch,1);}
void WriteLn(File aFile) {
  char* crlf = CR_LF;
  _FrameWork_::Write_Data(aFile,crlf,CRLF_LEN);}
void WriteText(File aFile, String theSt) { // write text or line break..
  char* here = theSt.text_();
  char* crlf = CR_LF;
  int ix = 0, mx = theSt.length_(), zx = -1;
  if (mx==0) return;
  if (DebugLogFile) if (_FrameWork_::LogTextWrites) if (aFile != DebugLogFile) {
    WriteChar(DebugLogFile,'\'');
    zx = 80;}
  while (ix<mx) {
    if (here[ix] == '\n') {
      if (ix>0) {
        if (zx>0) {
          _FrameWork_::Write_Data(DebugLogFile,here,MIN(ix,zx));
          if (ix <= zx) WriteChar(DebugLogFile,'\\');
          zx = MAX(zx-1-ix,0);}
        if (_FrameWork_::Write_Data(aFile,here,ix)<ix) break;}
      if (_FrameWork_::Write_Data(aFile,crlf,CRLF_LEN)<CRLF_LEN) break;
      here = (char*)(here+ix+1);
      mx = mx-1-ix;
      ix = 0;}
    else ix++;}
  if (ix>0) {
    if (zx>0) _FrameWork_::Write_Data(DebugLogFile,here,MIN(ix,zx));
    _FrameWork_::Write_Data(aFile,here,ix);}
  if (zx<0) return;
  WriteChar(DebugLogFile,'\'');
  WriteLn(DebugLogFile);}
  // DebugLogFile->flush_(); // incredibly slow!!
void WriteChr4(File aFile, Char4 cx) {WriteText(aFile,String(cx));} //~WriteText

String GetCurWinDirName(void) {
  char cbf[PathSize+1];
  int nn = GetCurrentDirectory(PathSize,cbf);
  String theName = String(cbf,nn);
  if (nn >= PathSize)
    nn = GetCurrentDirectory(nn+1,theName.text_());
  return theName;}

String ProgFileName(void) { // returns full path name
  char thename[2222]; // plenty big
  if (GetModuleFileName(NULL,thename,2220)<2222) return String(thename);
  return Null_Str;} //~ProgFileName

////////////////////////////////////////////////////////////////////////

// Forward function declarations, mostly defined in System.h ..

#ifdef MOS_System

// namespace hack into System.h (which mostly follows after this file)..
namespace  _System_ { // so we can see these fwds..
  void Async_Msg(boolean SysEv, Char4 EvNo, int eInfo, int EvLoc, int EvDat,
      Process toPro);
  WinWindow Create_Splash(IntArray Splashy);
  void System_Inits(IntArray eTrs, IntArray FulCn);
  void CacheInvalid8(File theFi, int rty, int rno, IntArray AryPtr);
  void Zap_TheDrag(void); // gets used soonly
  bool Analyze_DragStr(WinWindow theWin, String theData);
  void Drop_Here(WinWindow theWin); // sends "Drop" to app
  int PkgDataOffset(DraCliPkg thePkg, int theType);
  void HidenDie(int why, Window aWin);
  void Frame_Cursor(WinWindow theWin, int here);
  void Do_Menu_Comm(Char4 EvNo, Window myWin, Process toPro);
  boolean Do_KeyDown(WinWindow theWin, char theKey); // true if we took it
  boolean Mouse_InDocX(WinWindow theWin, int mov);
  boolean Scroll_Some(WinWindow theWin, int whom, int whar);
  Window Rectify_MyWin(WinWindow theWin, int fixt);
  boolean Windo_Resized(WinWindow theWin);
  boolean Windo_Painting(WinWindow theWin, RectAngle wnRect);
  void Windo_Activ8(Window theWin, int ax);
  void Carat_Blink_Chk(boolean ckt);
  int Do_WinEvent(int msg, Event ev);
  boolean Get_HotRect(WinWindow theWin, Widgit theWg, RectAngle wnRect);
  void unPacifyWin(Window theWn);
  int Get_Progress_Pix(Window theWin, int wide);
  void Clear_Hilites();
  boolean TrakDrag(WinWindow theWin, Point Locn, Widgit TrkWig, int taker);
  void Clean_Unpend(Process whom);
  void PopRelease(int zap);
  // void RollDate();
  void Delaying();
  void Quit_Process(Process whom);

 } // end namespace  _System_

namespace  _FrameWork_ { // resume

void Asyn_Msg(Char4 EvNo, int eInfo, int EvLoc, int EvDat, Process toPro) {
  _System_::Async_Msg(true,EvNo,eInfo,EvLoc,EvDat,toPro);}
void Hiden_Die(Window aWin) {_System_::HidenDie(0,aWin);}
void Do_Menu_Comm(Char4 EvNo, Window myWin, Process toPro) {
  _System_::Do_Menu_Comm(EvNo,myWin,toPro);}
boolean Do_KeyDown(WinWindow theWin, char theKey) {
  return _System_::Do_KeyDown(theWin,theKey);} // true if we took it
boolean Mouse_InDocX(WinWindow theWin, int mov) {
  return _System_::Mouse_InDocX(theWin,mov);}
boolean Scroll_Some(WinWindow theWin, int whom, int whar) {
  return _System_::Scroll_Some(theWin,whom,whar);}
Window Rectify_MyWin(WinWindow theWin, int fixt) {
  return _System_::Rectify_MyWin(theWin,fixt);}
boolean Windo_Resized(WinWindow theWin) {
  return _System_::Windo_Resized(theWin);}
boolean Windo_Painting(WinWindow theWin, RectAngle wnRect) {
  return _System_::Windo_Painting(theWin,wnRect);}
void Windo_Activ8(Window theWin, int ax) {_System_::Windo_Activ8(theWin,ax);}
int Do_WinEvent(int msg, Event ev) {return _System_::Do_WinEvent(msg,ev);}
boolean Get_HotRect(WinWindow theWin, Widgit theWg, RectAngle wnRect) {
  return _System_::Get_HotRect(theWin,theWg,wnRect);}
void unPacifyWin(Window theWn) {_System_::unPacifyWin(theWn);}
int Get_Progress_Pix(Window theWin, int wide) {
  return _System_::Get_Progress_Pix(theWin,wide);}
void Clean_Un_pend(Process whom) {_System_::Clean_Unpend(whom);}
void PopRelease(int zap) {_System_::PopRelease(zap);}
void Quit_Proc(Process whom) {_System_::Quit_Process(whom);}

 } // (temply) end namespace  _FrameWork_

String FormInt(int theNum); // these also declared in SysNames.h (used above)
String FormHexNum(int num, int dgs);
String FormBibleRef(int ref);
String FormDateTime(int theNum, int opt);
void CloseWindow(Window aWin); // used in FW

#else // if no MOS_System, no '#include System.h'..

// namespace  _System_ { // redirect compiler-generated pointer-down-count..
// void Cache_Done(void* aptr, int whom) {dn__ptr(aptr);}
//  } // end namespace  _System_

namespace  _FrameWork_ { // resume

void Asyn_Msg(Char4 EvNo, int eInfo, int EvLoc, int EvDat, Process toPro) {}
void Hiden_Die(Window aWin) {}
void Do_Menu_Comm(Char4 EvNo, Window myWin, Process toPro) {}
boolean Do_KeyDown(WinWindow theWin, char theKey) {return false;}
boolean Mouse_InDocX(WinWindow theWin, int mov) {return false;}
boolean Scroll_Some(WinWindow theWin, int whom, int whar) {return false;}
Window Rectify_MyWin(WinWindow theWin, int fixt) {return NULL;}
boolean Windo_Resized(WinWindow theWin) {return false;}
boolean Windo_Painting(WinWindow theWin, RectAngle wnRect) {return false;}
int Do_WinEvent(int msg, Event ev) {return 0;}
boolean Get_HotRect(WinWindow theWin, Widgit theWg, RectAngle wnRect)
  {return false;}
void unPacifyWin(Window theWn) {}
int Get_Progress_Pix(Window theWin, int wide) {return 0;}
boolean TrakDrag(WinWindow theWin, Point Locn, Widgit TrkWig, int taker)
  {return false;}
void Clean_Un_pend(Process whom) {}
void PopRelease(int zap) {}
void Quit_Proc(Process whom) {whom->T2F_Quit();} // if (Ready_List=MainProgram)

void Windo_Activ8(Window theWin, int ax) {
  if (LoHalf(ax) == WA_ACTIVE) theFrontWin = GetWgtWinWin(theWin);
    else theFrontWin = NULL;}

 } // (temply) end namespace  _FrameWork_

String FormInt(int theNum) {return String(theNum);}
String FormHexNum(int num, int dgs) {return String(num);}
String FormDateTime(int theNum, int opt) {return _FrameWork_::FormTimeMS();}
void CloseWindow(Window aWin) {}
void AsyncMsg(Char4 EvNo, int eInfo, int EvLoc, int EvDat, Process toPro) {}

#endif // (not) MOS_System

// MsgBox constants and fwds..
Char4 QuestIcon = "Qstn";
Char4 InfoIcon = "Bang";
Char4 StopIcon = "Stop";

// array bounds check..

int* _tmp_0_; int* _tmp_00_;
int __aix__, __aiy__, __aiz__;

int _Array_Bounds_Fail_(int aix, int aiz) { // never returns
  int aiy = 0;
  MsgBox(String("Array bounds check ")+String(aix)+"<"+String(aiz)
      +" {"+ProcID(NULL)+"}",StopIcon,0);
  // Asyn_Msg("AB_Q",aix,aiz,0,_FrameWork_::MainProgram);
  if (_Crash_and_Burn_) {aiy = 0; aiy = aix/aiy;}
  _FrameWork_::Asyn_Msg("Died",aix,aiz,0,_FrameWork_::MainProgram);
  ExitProcess(0); // Quit();
  return 0;} // ~_Array_Bounds_Fail_

// #define _Bounds_ChkL_(ary,aix) ary[2+(aix)] // moved to SysNames.h
// #define _Bounds_ChkR_(ary,aix) ary[2+(aix)] // moved to SysNames.h

void Valid8_Event(Event theEv) {
  int az = 0;
  if (theEv->_00_ == 7) if ((theEv->_ev_state_&-4)==0) return;
  if (_Crash_and_Burn_) az = (++az)/az; // div/0: I want to know about this
  // _FrameWork_::FILL0((int*)theEv,0,7); // set_ary_siz does it
  set_ary_size((void*)theEv,7);
  theEv->_ev_link_ = NULL;}

////////////////////////////////////////////////////////////////////////

// Process management stuff..

_T2F_Process_::_T2F_Process_(String pname = Null_Str, Char4 pnam4 = Null4) {
  set_ary_size((void*)this,-(sizeof(_T2F_Process_)/sizeof(int)-1));
  Char4 pn4 = pnam4; // should be already valid & unique, from NewProgr
  char* Qname = NULL; char* Wname = NULL;
  _event_list_ = NULL;
  Thread_32 = NULL;
  WinThread_ID = 0;
  if (pn4 <= NoName) if (pname != Null_Str) pn4 = pname.text_();
  if (pn4 > NoName) if (Proc4ch(pn4) != NULL) pn4 = NoName;
  if (pn4 <= NoName) {
    _FrameWork_::Next_ProcID = Up4Char(_FrameWork_::Next_ProcID);
    pn4 = _FrameWork_::Next_ProcID;}
  if (pname != Null_Str) {
    if (ProcByName(pname) != NULL) pname = pname+" "+pn4;
    _my_name_ = pname+" +";
    Wname = _my_name_.text_();
    Qname = pname.text_();}
  NewSem32(_wait_event_,0,Qname);
  NewSem32(_reply_wait_,0,Wname);
  _my_name_ = pname;
  _my_ID_ = pn4;
  _Pend_Bitz_ = 0;
  _Delay_Time_ = 0;
  _link_ = _FrameWork_::Ready_List;
  _FrameWork_::Ready_List = this;
  if (DebugLogFile) SystemDebugLog(String("*** NEW ") + pname
      + String(" = ") + pnam4 + _FrameWork_::FormTimeMS());} //~T2F_Process

Process NewProgram(String pname, File dir, Char4 pnam4) {
  // open program file pname in dir, give it (suggested) process ID pnam4
  Process whom;
  DWORD aThreadID = GetCurrentThreadId();
  PTHREAD_START_ROUTINE starter = (PTHREAD_START_ROUTINE)StartNewProg;
  if (_FrameWork_::Ready_List != NULL) {
    if (pname == Null_Str) return NULL;
    if (CharAt(0,pname) <= ' ') return NULL;}
  if (pnam4 <= NoName) {
    do {_FrameWork_::Next_ProcID = Up4Char(_FrameWork_::Next_ProcID);}
      while (Proc4ch(_FrameWork_::Next_ProcID) != NULL);
    pnam4 = _FrameWork_::Next_ProcID;}
  else while (Proc4ch(pnam4) != NULL) pnam4 = Up4Char(pnam4);
  // if (DebugLogFile) SystemDebugLog(String("+ NewProg ")+pnam4+" "+pname);
  try {whom = new _T2F_Process_(pname,pnam4);} catch (...) {whom = NULL;}
  if (whom==NULL) return NULL;
  if (pname == Null_Str) return whom; // System process is Gooey thread
  if (CharAt(0,pname) > ' ') {
    whom->Thread_32 = CreateThread(NULL,0,starter,whom,
        CREATE_SUSPENDED,&aThreadID);
    whom->WinThread_ID = aThreadID;
    ResumeThread(whom->Thread_32);}
  else whom->WinThread_ID = aThreadID;
  if (DebugLogFile)
    SystemDebugLog(String("  = ")+ProcID(whom)+" "+String((int)aThreadID));
  return whom;} // ~NewProgram

Process MySelf() {
  int az = 0;
  DWORD myID = GetCurrentThreadId();
  Process here = _FrameWork_::Ready_List;
  while (here != NULL)
    if (here->WinThread_ID == myID) return here;
      else here = here->_link_;
  if (_Crash_and_Burn_) az = (++az)/az; // div/0
  return NULL;} //~MySelf

boolean Valid_Process(Process whom);

////////////////////////////////////////////////////////////////////////

// Predefined Event class, implemented as a pointer to class _Event_

//_Event_::_Event_() {}
//  set_ary_size((void*)this,7);
//  _ev_link_ = NULL;}

int Evp_cnt_Chk = 0;
Event Evp_Chk(Event evp) {
  int az = 0;
  if (((int)evp) == 0xcccccccc) {
    _FrameWork_::Log_Panic(String("*** Evp-Chk *** ")+String(Evp_cnt_Chk));
    if (_Crash_and_Burn_) az = (++az)/az; // div/0
    evp = NULL;}
  Evp_cnt_Chk++;
  return evp;}

void _Event_::EnQueue(Process whom, bool state) {
  Event here;
  if (!Valid_Process(whom)) return;
  Wait32(_FrameWork_::EvenLok);
  here = Evp_Chk(whom->_event_list_);
  _ev_link_ = NULL;
  _ev_state_ = ORD(state)&1;
  if (here != NULL) {
    // Valid8_Event(here);
    while (here->_ev_link_ != NULL) {
      here = Evp_Chk(here->_ev_link_);
      if (here==NULL) break;}
      // Valid8_Event(here);}
    if (here != NULL) here->_ev_link_ = this;}
  else whom->_event_list_ = this;
  Signal32(_FrameWork_::EvenLok);
  Signal32(whom->_wait_event_);} //~EnQueue

Event De_Queue(Process whom) { // leaves it there if state=1 (now 2)
  Event here;
  if (!Valid_Process(whom)) return NULL;
  Wait32(_FrameWork_::EvenLok);
  here = Evp_Chk(whom->_event_list_);
  if (here != NULL) {
    // Valid8_Event(here);
    if (here->_ev_state_ != 1) {
      whom->_event_list_ = here->_ev_link_;
      here->_ev_link_ = NULL;}
    if (here->_ev_state_ > 0) here->_ev_state_++;}
  Signal32(_FrameWork_::EvenLok);
  return here;} //~De_Queue

// Event processing functions..

boolean I_am_Gooey() {return GetCurrentThreadId() == Gooey_Thread_ID;}

namespace  _FrameWork_ { // more

Event Pending_Event = NULL;

int PkgPending = 0; // see also Process->_Pend_Bitz_, Pend_Ex()
// 0=idle, 1=WM_MOUSEMOVE, 3=Analyze_DragStr, 4=Drop_Here.
//  -3: "AcTk" sent, -2: Hotclick sent, -1: Fetch_Ev after -2/-3
// Five busy states for PkgPendin/PendBitz:            -- 5=DragClipEvent
// 1. Start drag, reserved for "Drag" evt, in which NewDrag builds "Drag"
// 2. Drag event accepted, upgrades PkgPendin from 1 to 2, sets PendBitz
//   and sends it off by PkgDone->StartDrago, which releases (or else
//   Drag evt exit releases) if not taken by #4.
// 3. Drag enter keeps #2, or else AnalDragStr makes new "Drop" pkg #3
//   which is released on drag leave.
// 4. Drop takes over #2 or #3 for "Drop" evt as #4, to release on exit.
// 5. Drop event accepted, upgrades PkgPendin from 4 to 5.
// 7. DroPastEv makes "Past" from system clip, or NewClip builds "Clip",
//   both released on exit.

inline void Copy_Event(Event too, Event frm) {
  too->EventNo = frm->EventNo;
  too->EvInfo = frm->EvInfo;
  too->EvLocn = frm->EvLocn;
  too->EvDatum = frm->EvDatum;
  too->EvSend = frm->EvSend;}

void Post_Event(Event evp, Process toProc, Process fromm, bool wait4) {
  Event evx;
  if (evp==NULL) return;
  if (!Valid_Process(toProc)) return;
  if (toProc->_my_name_ == Null_Str) return; // it's quitting
  // Valid8_Event(evp);
  if (fromm==NULL) fromm = MySelf();
  else if (!Valid_Process(fromm)) return;
  else if (wait4) if (fromm != MySelf()) return;
  if (!wait4) { // if wait4=false, evp will be disposed after delivery..
    evx = evp;
    evp = (Event)new_int_ary(7); // evp = new _Event_(); // so make a copy
    if (evp==NULL) return;
    // Valid8_Event(evp);
    Copy_Event(evp,evx);} // evp = evx
    // evp->_ev_state_ = 0; // EnQueue does it
  else if (toProc==fromm) return;
  See_Panic(ORD("+Ev+"));
  See_Panic(ORD(evp->EventNo));
  See_Panic(ORD(ProcID(toProc)));
  if (Sys_Trace|Post_Trace) if (DebugLogFile) {
    SystemDebugLog(String("--Post ")+ProcID(fromm)+" > "+ProcID(toProc)+" "
        + String(wait4)+" '"+(evp->EventNo)+"' "+FormInt(evp->EvInfo)+", "
        + FormInt(evp->EvLocn)+", "+FormInt(evp->EvDatum)+FormTimeMS());}
  evp->EvSend = fromm;
  evp->EnQueue(toProc,wait4);
  // Signal_Sem(&toProc->_wait_event_); // no, EnQue did
  if (!wait4) return;
  DWORD why = Wait32(fromm->_reply_wait_);
  if (why==WAIT_OBJECT_0) return;
  evx = evp;} //~Post_Event

bool Fetch_Event(Event evp, Process whom) { // false if flushit && NULL
  Event evx;
  boolean flushit = false;
  Process fromm;
  if (evp==NULL) if (whom==NULL) return false; // caller checks
  if (whom==NULL) whom = MySelf(); // now assumed
  else if (!Valid_Process(whom)) return false;
    else flushit = true;
  if ((PkgPending|whom->_Pend_Bitz_) !=0) Clean_Un_pend(whom); // dispose drag
  evx = Evp_Chk(whom->_event_list_);
  if (flushit) if (evx==NULL) return false;
  if (evx != NULL) {
    // Valid8_Event(evx);
    if (evx->GetState() == 2) { // we have a reply to send..
      fromm = evx->EvSend;
      evx = De_Queue(whom);
      if (evx != NULL) if (Valid_Process(fromm)) {
        if (evp != NULL) Copy_Event(evx,evp); // *evx = *evp;
        evx->EvSend = fromm;
        Signal32(fromm->_reply_wait_);}}}
  while (true) {
    if (!flushit) Wait32(whom->_wait_event_);
    else if (whom->_event_list_ == NULL) return false;
    evx = De_Queue(whom);
    if (evx==NULL) { // huh??
      if (flushit) return false;
      continue;}
    See_Panic(ORD("-Ev-"));
    See_Panic(ORD(evx->EventNo));
    See_Panic(ORD(ProcID(whom)));
    if (evp != NULL) Copy_Event(evp,evx); // *evp = *evx;
    if (Sys_Trace) if (DebugLogFile) {
      SystemDebugLog(String("--GetEv ")+ProcID(whom)+" < "+ProcID(evx->EvSend)
          + " '"+(evx->EventNo)+"' "+FormInt(evx->EvInfo)+", "
          + FormInt(evx->EvLocn)+", "+FormInt(evx->EvDatum)+FormTimeMS());}
    if (evx->GetState() ==0) {
      if (whom->_Pend_Bitz_ ==2) if (evx->EventNo == "Drag")
        if (PkgPending==1) PkgPending = 2; // make drag disposable
      if (whom->_Pend_Bitz_ ==5) if (PkgPending==4)
        if (evx->EventNo == "Drop" || evx->EventNo == "DrTr")
          PkgPending = 5; // ditto
      down__ptr((int*)Evp_Chk(evx));} // delete evx;}
    return true;}} // ~Fetch_Event

////////////////////////////////////////////////////////////////////////

// DragonDrop data and functions (still inside _FrameWork_ namespace)..

#ifdef MOS_System // not needed otherwise

// Drag stuff..
int* theClipDrag = null;
Char4 ClipDragOwn = Null4;
String ClipDragStr = Null_Str;

#define ClpDrgPkSz 44
int ClipDragPkg[2+ClpDrgPkSz+1]; // pkg positions:
  //  +0 length = 44 // ClipDragPkg[0] = 44;
  // 2+0 length (dupe); -> pkg if Signature<0
  // 2+1 Signature "Drag" or "Drop"
  // 2+2 current size
  // 2+3 Sender's procID
  // 2+4 LENGTH(theDrag.DrDataStr), for validation
  // 2+5 Chr4Str(theDrag.DrDataStr), ditto
  // 2+6 current (potential) drop loc, -1 if trash can
  // 2+7 modifiers
  // 2+8 start of real data

/////////////////////////////////////////////// J.Brown's 2004 www.catch22.net

// D&D code adapted from <http://www.catch22.net/tuts/dragdrop.asp> [Part 1]

// Implementation of the IDropSource COM interface

class CDropSource : public IDropSource {
public: // IUnknown members
  HRESULT __stdcall QueryInterface(REFIID iid, void ** ppvObject);
  ULONG   __stdcall AddRef(void);
  ULONG   __stdcall Release(void);
// IDropSource members
  HRESULT __stdcall QueryContinueDrag(BOOL fEscapePressed, DWORD grfKeyState);
  HRESULT __stdcall GiveFeedback(DWORD dwEffect);
// Constructor / Destructor
  CDropSource();
  ~CDropSource();
private: LONG DS_RefCount;}; // ~CDropSource

CDropSource::CDropSource() {DS_RefCount = 1;}
CDropSource::~CDropSource() {}
ULONG __stdcall CDropSource::AddRef(void) {
  return InterlockedIncrement(&DS_RefCount);} // increment object ref count
ULONG __stdcall CDropSource::Release(void) {
  LONG count = InterlockedDecrement(&DS_RefCount); // decr object ref count
  if (count == 0) {
    delete this;
    return 0;}
  else return count;}

HRESULT __stdcall CDropSource::QueryInterface(REFIID iid, void **ppvObject) {
  // check to see what interface has been requested
  if (iid == IID_IDropSource || iid == IID_IUnknown) {
    AddRef();
    *ppvObject = this;
    return S_OK;}
  else {
    *ppvObject = 0;
    return E_NOINTERFACE;}}

// CDropSource::QueryContinueDrag
//   Called by OLE whenever Escape/Control/Shift/Mouse buttons have changed
HRESULT __stdcall CDropSource::QueryContinueDrag(
    BOOL fEscapePressed, DWORD grfKeyState) {
  // if the <Escape> key has been pressed since the last call, cancel the drop
  if (fEscapePressed == TRUE) return DRAGDROP_S_CANCEL;  
  // if the <LeftMouse> button has been released, then do the drop!
  if ((grfKeyState & MK_LBUTTON) == 0) return DRAGDROP_S_DROP;
  return S_OK;} // continue with the drag-drop

// CDropSource::GiveFeedback
//   Return either S_OK, or DRAGDROP_S_USEDEFAULTCURSORS to instruct OLE
//     to use the default mouse cursor images
HRESULT __stdcall CDropSource::GiveFeedback(DWORD dwEffect) {
  return DRAGDROP_S_USEDEFAULTCURSORS;}

//  Helper routine to create an IDropSource object
HRESULT CreateDropSource(IDropSource** ppDropSource) {
  if (ppDropSource == 0) return E_INVALIDARG;
  *ppDropSource = new CDropSource();
  return (*ppDropSource) ? S_OK : E_OUTOFMEMORY;}

// Implementation of the IEnumFORMATETC interface
 // For Win2K and above look at the SHCreateStdEnumFmtEtc API call!!
 //  Apparently the order of formats in an IEnumFORMATETC object must be
 //  the same as those that were stored in the clipboard

class CEnumFormatEtc : public IEnumFORMATETC {
public:
  // IUnknown members
  HRESULT __stdcall QueryInterface (REFIID iid, void** ppvObject);
  ULONG __stdcall AddRef (void);
  ULONG __stdcall Release (void);

  // IEnumFormatEtc members
  HRESULT __stdcall Next (ULONG celt, FORMATETC* rgelt, ULONG* pceltFetched);
  HRESULT __stdcall Skip (ULONG celt); 
  HRESULT __stdcall Reset (void);
  HRESULT __stdcall Clone (IEnumFORMATETC** ppEnumFormatEtc);

  CEnumFormatEtc(FORMATETC* pFormatEtc, int nNumFormats);
  ~CEnumFormatEtc();

private:
  LONG EF_RefCount;           // Reference count for this COM interface
  ULONG EF_nIndex;            // current enumerator index
  ULONG EF_nNumFormats;       // number of FORMATETC members
  FORMATETC* EF_pFormatEtc;}; // array of FORMATETC objects

HRESULT CreateEnumFormatEtc(UINT nNumFormats, FORMATETC* pFormatEtc,
      IEnumFORMATETC** ppEnumFormatEtc) {
  // "Drop-in" replacement for SHCreateStdEnumFmtEtc. Called by EnumFormatEtc
  if (nNumFormats == 0 || pFormatEtc == 0 || ppEnumFormatEtc == 0)
    return E_INVALIDARG;
  *ppEnumFormatEtc = new CEnumFormatEtc(pFormatEtc, nNumFormats);
  return (*ppEnumFormatEtc) ? S_OK : E_OUTOFMEMORY;}

static void DeepCopyFormatEtc(FORMATETC* dest, FORMATETC* source) {
  // Helper function to perform a "deep" copy of a FORMATETC
  *dest = *source; // copy the source FORMATETC into dest
  if (source->ptd) {
    // allocate memory for the DVTARGETDEVICE if necessary
    dest->ptd = (DVTARGETDEVICE*)CoTaskMemAlloc(sizeof(DVTARGETDEVICE));
    // copy the contents of the source DVTARGETDEVICE into dest->ptd
    *(dest->ptd) = *(source->ptd);}}

CEnumFormatEtc::CEnumFormatEtc(FORMATETC *pFormatEtc, int nNumFormats) {
  EF_RefCount = 1;
  EF_nIndex = 0;
  EF_nNumFormats = nNumFormats;
  EF_pFormatEtc = new FORMATETC[nNumFormats];
  for (int i = 0; i < nNumFormats; i++) // copy the FORMATETC structures
    DeepCopyFormatEtc(&EF_pFormatEtc[i], &pFormatEtc[i]);}

CEnumFormatEtc::~CEnumFormatEtc() { // Destructor
  if (EF_pFormatEtc) {
    for (ULONG i = 0; i < EF_nNumFormats; i++)
      if (EF_pFormatEtc[i].ptd) CoTaskMemFree(EF_pFormatEtc[i].ptd);
    delete[] EF_pFormatEtc;}}

ULONG __stdcall CEnumFormatEtc::AddRef(void) {
  return InterlockedIncrement(&EF_RefCount);} // increment object ref count

ULONG __stdcall CEnumFormatEtc::Release(void) {
  LONG count = InterlockedDecrement(&EF_RefCount); // decrement object ref count
  if (count == 0) {
    delete this;
    return 0;}
  else return count;}

HRESULT __stdcall CEnumFormatEtc::QueryInterface(REFIID iid, void** ppvObject) {
  // check to see what interface has been requested
  if (iid == IID_IEnumFORMATETC || iid == IID_IUnknown) {
    AddRef();
    *ppvObject = this;
    return S_OK;}
  else {
    *ppvObject = 0;
    return E_NOINTERFACE;}}

HRESULT __stdcall CEnumFormatEtc::Next(ULONG celt, FORMATETC* pFormatEtc,
      ULONG* pceltFetched) {
  // If the returned FORMATETC structure contains a non-null "ptd" member, then
  // the caller must free this using CoTaskMemFree (stated in the COM doc)
  ULONG copied  = 0;
  if (celt == 0 || pFormatEtc == 0) return E_INVALIDARG; // validate arguments
  // copy FORMATETC structures into caller's buffer
  while (EF_nIndex < EF_nNumFormats && copied < celt) {
    DeepCopyFormatEtc(&pFormatEtc[copied], &EF_pFormatEtc[EF_nIndex]);
    copied++;
    EF_nIndex++;}
  if (pceltFetched != 0) *pceltFetched = copied; // store result
  return (copied == celt) ? S_OK : S_FALSE;} // did we copy all requested?

HRESULT __stdcall CEnumFormatEtc::Skip(ULONG celt) {
  EF_nIndex += celt;
  return (EF_nIndex <= EF_nNumFormats) ? S_OK : S_FALSE;}

HRESULT __stdcall CEnumFormatEtc::Reset(void) {
  EF_nIndex = 0;
  return S_OK;}

HRESULT __stdcall CEnumFormatEtc::Clone(IEnumFORMATETC ** ppEnumFormatEtc) {
  HRESULT hResult // make a duplicate enumerator..
    = CreateEnumFormatEtc(EF_nNumFormats, EF_pFormatEtc, ppEnumFormatEtc);
  if (hResult == S_OK) // manually set the index state
    ((CEnumFormatEtc*) *ppEnumFormatEtc)->EF_nIndex = EF_nIndex;
  return hResult;}

//	Implementation of the IDataObject COM interface

class CDataObject : public IDataObject {
public: // IUnknown members
  HRESULT __stdcall QueryInterface (REFIID iid, void ** ppvObject);
  ULONG   __stdcall AddRef (void);
  ULONG   __stdcall Release (void);
  // IDataObject members
  HRESULT __stdcall GetData(FORMATETC* pFormatEtc, STGMEDIUM* pMedium);
  HRESULT __stdcall GetDataHere(FORMATETC* pFormatEtc, STGMEDIUM *pMedium);
  HRESULT __stdcall QueryGetData(FORMATETC* pFormatEtc);
  HRESULT __stdcall GetCanonicalFormatEtc (FORMATETC* pFormatEct,
        FORMATETC* pFormatEtcOut);
  HRESULT __stdcall SetData(FORMATETC* pFormatEtc, STGMEDIUM* pMedium,
        BOOL fRelease);
  HRESULT __stdcall EnumFormatEtc(DWORD dwDirection,
        IEnumFORMATETC** ppEnumFormatEtc);
  HRESULT __stdcall DAdvise(FORMATETC* pFormatEtc, DWORD advf,
        IAdviseSink* pAdvSink, DWORD* pdwConnection);
  HRESULT __stdcall DUnadvise(DWORD dwConnection);
  HRESULT __stdcall EnumDAdvise(IEnumSTATDATA** ppEnumAdvise);
  
  CDataObject(FORMATETC* fmt, STGMEDIUM* stgmed, int count); // Constructor
  ~CDataObject();
  
private:
  int LookupFormatEtc(FORMATETC* pFormatEtc);
  LONG DO_RefCount;
  FORMATETC* DO_pFormatEtc;
  STGMEDIUM* DO_pStgMedium;
  LONG DO_nNumFormats;}; // ~CDataObject

CDataObject::CDataObject(FORMATETC* fmtetc, STGMEDIUM* stgmed, int count) {
  DO_RefCount = 1;
  DO_nNumFormats = count;
  DO_pFormatEtc = new FORMATETC[count];
  DO_pStgMedium = new STGMEDIUM[count];
  for (int i = 0; i < count; i++) {
    DO_pFormatEtc[i] = fmtetc[i];
    DO_pStgMedium[i] = stgmed[i];}}

CDataObject::~CDataObject() { // Destructor
  if (DO_pFormatEtc) delete[] DO_pFormatEtc; // cleanup..
  if (DO_pStgMedium) delete[] DO_pStgMedium;
  /* OutputDebugString("oof\n"); */}

ULONG __stdcall CDataObject::AddRef(void) {
  return InterlockedIncrement(&DO_RefCount);} // increment object ref count

ULONG __stdcall CDataObject::Release(void) {
  LONG count = InterlockedDecrement(&DO_RefCount); // decrement object ref count
  if (count == 0) {
    delete this;
    return 0;}
  else return count;}

HRESULT __stdcall CDataObject::QueryInterface(REFIID iid, void** ppvObject) {
  // check to see what interface has been requested
  if (iid == IID_IDataObject || iid == IID_IUnknown) {
    AddRef();
    *ppvObject = this;
    return S_OK;}
  else {
    *ppvObject = 0;
    return E_NOINTERFACE;}}

HGLOBAL DupMem(HGLOBAL hMem) {
  // lock the source memory object
  DWORD len = GlobalSize(hMem);
  PVOID source = GlobalLock(hMem);
  PVOID dest = GlobalAlloc(GMEM_FIXED, len); // create a fixed "global" block
  memcpy(dest, source, len);  // i.e. just a regular lump of our process heap
  GlobalUnlock(hMem);
  return (HGLOBAL)dest;}

int CDataObject::LookupFormatEtc(FORMATETC* pFormatEtc) {
  for (int i = 0; i < DO_nNumFormats; i++) 
    if ((pFormatEtc->tymed & DO_pFormatEtc[i].tymed)
      && pFormatEtc->cfFormat == DO_pFormatEtc[i].cfFormat
      && pFormatEtc->dwAspect == DO_pFormatEtc[i].dwAspect) return i;
  return -1;}

HRESULT __stdcall CDataObject::GetData (FORMATETC* pFormatEtc,
      STGMEDIUM* pMedium) {
  int idx = LookupFormatEtc(pFormatEtc);
  // try to match the requested FORMATETC with one of our supported formats
  if (idx == -1) return DV_E_FORMATETC;
  // found a match! transfer the data into the supplied storage-medium
  pMedium->tymed = DO_pFormatEtc[idx].tymed;
  pMedium->pUnkForRelease = 0;
  switch (DO_pFormatEtc[idx].tymed) {
  case TYMED_HGLOBAL:
    pMedium->hGlobal = DupMem(DO_pStgMedium[idx].hGlobal);
    // return S_OK;
    break;
  default: return DV_E_FORMATETC;}
  return S_OK;}

HRESULT __stdcall CDataObject::GetDataHere (FORMATETC* pFormatEtc,
      STGMEDIUM* pMedium) {
  // GetDataHere is only required for IStream and IStorage mediums
  // It is an error to call GetDataHere for things like HGLOBAL
  //  and other clipboard formats
  return DATA_E_FORMATETC;} // OleFlushClipboard

HRESULT __stdcall CDataObject::QueryGetData (FORMATETC* pFormatEtc) {
  // Called to see if the IDataObject supports the specified format of data
  return (LookupFormatEtc(pFormatEtc) == -1) ? DV_E_FORMATETC : S_OK;}

HRESULT __stdcall CDataObject::GetCanonicalFormatEtc (FORMATETC* pFormatEct,
      FORMATETC* pFormatEtcOut) {
  pFormatEtcOut->ptd = NULL; // Apparently we have to set this field to NULL
  return E_NOTIMPL;} // .. even though we don't do anything else

HRESULT __stdcall CDataObject::SetData (FORMATETC* pFormatEtc,
      STGMEDIUM* pMedium,  BOOL fRelease) {
  return E_NOTIMPL;}

HRESULT __stdcall CDataObject::EnumFormatEtc (DWORD dwDirection,
      IEnumFORMATETC** ppEnumFormatEtc) {
  // for Win2k+ you can use the SHCreateStdEnumFmtEtc API call, however
  // to support all Windows platforms we need to implement IEnumFormatEtc
  if (dwDirection == DATADIR_GET) 
    return CreateEnumFormatEtc(DO_nNumFormats, DO_pFormatEtc, ppEnumFormatEtc);
  else return E_NOTIMPL;} // the direction specified is not support for drag+drop

HRESULT __stdcall CDataObject::DAdvise (FORMATETC* pFormatEtc, DWORD advf,
      IAdviseSink* pAdvSink, DWORD *pdwConnection) {
  return OLE_E_ADVISENOTSUPPORTED;}

HRESULT __stdcall CDataObject::DUnadvise (DWORD dwConnection) {
  return OLE_E_ADVISENOTSUPPORTED;}

HRESULT __stdcall CDataObject::EnumDAdvise (IEnumSTATDATA** ppEnumAdvise) {
  return OLE_E_ADVISENOTSUPPORTED;}

HRESULT CreateDataObject (FORMATETC* fmtetc, STGMEDIUM* stgmeds, UINT count,
      IDataObject** ppDataObject) { // Helper function
  if (ppDataObject == 0) return E_INVALIDARG;
  *ppDataObject = new CDataObject(fmtetc, stgmeds, count);
  return (*ppDataObject) ? S_OK : E_OUTOFMEMORY;}

HANDLE StringToHandle(char* szText) {
  void* ptr;
  int nTextLen = lstrlen(szText)+1;
  HGLOBAL ClipData = GlobalAlloc(GHND|GMEM_SHARE,nTextLen);
  // allocate and lock a global memory buffer. Make it fixed
  // data so we don't have to use GlobalLock
  // ptr = (void*)GlobalAlloc(GMEM_FIXED, nTextLen);
  ptr = (void*)GlobalLock(ClipData);
  memcpy(ptr, szText, nTextLen); // copy the string into the buffer
  GlobalUnlock(ClipData);
  return (HANDLE)ClipData;}

void RegisterDropWindow(HWND theWin, IDropTarget** ppDropTarget);
void UnregisterDropWindow(HWND theWin, IDropTarget* pDropTarget);

#endif // MOS_System

/////////////////////////////////////////// end of J.Brown's code (320 LoC)

// Graphics-related operations..

void FlipFlopPixels(int* ptr, int iTall, int iWide) {
  int ix, zx, tmp, ro, co, wid = iWide>>2, bak = (iTall-1)*wid, haf = iTall>>1;
  ptr++;
  if (*ptr) (*ptr)--; // doesn't dispose, but shouldn't need to
  ptr++;
  for (ro=0; ro<haf; ro++) { // goofy win95 pixels broken
    ix = ro*wid;
    zx = bak-ix;
    for (co=0; co<wid; co++) {
      tmp = ptr[zx];
      ptr[zx++] = ptr[ix];
      ptr[ix++] = tmp;}}
  ix = 0;} // ~FlipFlopPixels

// Selector constants for posting window manager events to foreground thread
#define Windo_Event (WM_USER*2) // WM_USER=0x0400
#define WEV_HidenShoWgt (Windo_Event+2)
#define WEV_Delay     (Windo_Event+27)
#define WEV_MsgBx     (Windo_Event+28)
#define WEV_AskTx     (Windo_Event+29)
#define WEV_CheckMenu (Windo_Event+43)
#define WEV_AddMenu   (Windo_Event+44)
#define WEV_UndoMenu  (Windo_Event+45)
#define WEV_NewMenBar (Windo_Event+46)
#define WEV_PopMenu   (Windo_Event+47)
#define WEV_NewWindow (Windo_Event+48)
#define WEV_CloseWind (Windo_Event+49) // CloseWindow
#define WEV_ShowWind  (Windo_Event+50) // Win2Behind/ShowWindow
#define WEV_WindowAcc (Windo_Event+51) // WindowAccess
#define WEV_HideWind  (Windo_Event+53)
#define WEV_InvalRect (Windo_Event+54)
#define WEV_DocScroll (Windo_Event+59) // Get-/SetDocScroll
#define WEV_WigitDraw (Windo_Event+60) // WidgitDraw
#define WEV_IniWidgit (Windo_Event+62)
#define WEV_PkgDone   (Windo_Event+85)
#define WEV_Splsh     (Windo_Event+88)
#define WEV_NoHigher  (Windo_Event+99) // protect from spurious WinXP garbage

_Event_ xev;
int DoGooey(int msg, Char4 EvNo, int eInfo, int EvLoc, int EvDat) {
  Process usn = MySelf();
  _Event_ ev;
  set_ary_size((void*)&ev,-7);
  // Valid8_Event(&ev);
  // Valid8_Event(&xev);
  ev.EventNo = EvNo;
  ev.EvInfo = eInfo;
  ev.EvLocn = EvLoc;
  ev.EvDatum = EvDat;
  ev.EvSend = usn;
#if Panic_Log
  See_Panic(ORD("Goo+"));
  See_Panic(ORD(EvNo));
  if (EvNo == "Wacc") See_Panic(eInfo);
  See_Panic(ORD(ProcID(usn)));
#endif
  if (Sys_Trace) if (DebugLogFile) SystemDebugLog(String("-->Gooy ")
      + ProcID(usn)+" "+msg+", '"+EvNo+"' "+FormInt(eInfo)+" "
      + FormInt(EvLoc)+" "+FormInt(EvDat)+FormTimeMS());
  if (msg == WEV_PkgDone) {
    set_ary_size((void*)&xev,-7);
    xev = ev;
    PostMessage(GooeyWin,msg,0,(LPARAM)(&xev));
    return 0;}
  if (I_am_Gooey()) Do_WinEvent(msg,&ev);
    else SendMessage(GooeyWin,msg,0,(LPARAM)(&ev));
  if (msg == WEV_PopMenu) while (ev.EvInfo == HiBit)
    if (Pending_Event == NULL) break; // huh??
      else SendMessage(GooeyWin,WEV_Delay,0,(LPARAM)(&ev));
  if (Sys_Trace) if (DebugLogFile) SystemDebugLog(String("<--Goox ")
      + ProcID(usn)+" "+FormInt(ev.EvInfo)+FormTimeMS());
  See_Panic(ORD("Goo-"));
  return ev.EvInfo;} // ~DoGooey

 } // (temp'ly) end namespace  _FrameWork_

#define Do_Gooey(ms,evn,inf,loc,dat) _FrameWork_::DoGooey(ms,evn,inf,loc,dat)
#define Gooey_Do(ms,evn,inf,loc,dat) _FrameWork_::DoGooey(ms,evn,inf,loc,dat)

////////////////////////////////////////////////////////////////////////

// Opaque Window type, implemented as a pointer to a WindInfo class..

// field names for Win_Set/Get:
#define _rTop 0
#define _rLeft 1
#define _rBotm 2
#define _rRite 3
#define _wKind 4
#define _wName 5
#define _wFlags 6
#define _wLink 7
#define _wParent 8
#define _wChild 9
#define _wWindo 10
#define _FoBgColo 11
#define _theText 12
#define _MoInfo 13
#define _wOwner 15
#define _wNxID 16
#define _wFocus 17
#define _wTextVH 18
#define _wUserData 19
#define _wMoreInfo 20
#define WinFlagsOffs _wFlags

Widgit _WindInfo_::Get_Wgt(int parm) { // hopefully inline
  if (parm==_wFocus) return wFocus;
    else return (Widgit)up__ptr(wChild);}
int _WindInfo_::Win_Get(int parm) { // hopefully inline
  //Point pt;
  if (parm>99) return (wFlags>>(parm-100))&1; // get flag bit
  switch (parm) {
  case _rTop: case _rLeft:
    return Pack2ints(rTop,rLeft);
  case _rBotm: case _rRite:
    return Pack2ints(rBotm-rTop,rRite-rLeft);
  case _wKind: return ORD(wKind);
  case _wName:
    if (Is4chars(ORD(wName))) return ORD(wName);
    return ORD("????");
  case _wFlags: return wFlags;
  case _wWindo:
    return 0;
  case _FoBgColo: return FoBgColo;
  case _MoInfo: return MoInfo;
  //case _theText:
  //  return (int)theText;
  case _wOwner: return ORD(wOwner);
  case _wNxID: return ORD(wNxID);
  case _wTextVH: return (int)wTextVH;
  case _wUserData: return wUserData;
  case _wMoreInfo: return wMoreInfo;}
  return 0;} // ~Win_Get

void _WindInfo_::Set_Wgt(Widgit datum, int parm) { // hopefully inline
  if (parm==_wFocus) wFocus = datum;
  else if (parm==_wChild) wChild = datum;}
void _WindInfo_::Set_Title(String title) {
  theText = title;
  if (HostInfo != 0) SetWindowText(Win32Win(),title.text_());}

void _WindInfo_::Win_Set(int datum, int parm) {
  int abit, vv, hh;
  POINT wpt;
  _RectAngle_ wnRect; // RECT wnRect;
  set_ary_size((void*)(&wnRect),-4);
  if (parm>99) { // set flag bit..
    parm = parm-100;
    datum = datum&1;
    abit = 1<<parm;
    vv = wFlags;
    wFlags = (wFlags&(-1-abit))+(datum<<parm);
    if (parm==0) if ((vv&wiVisble) != datum) { // changed wiVisble
      if (datum==0) {
        ShowWindow(Win32Win(),SW_HIDE);
        if (_FrameWork_::FrontedWin==Win32Win())
          _FrameWork_::FrontedWin = NULL;
        _FrameWork_::Down_Win_Cnt();}
      else {
        _FrameWork_::Window_Count++;
        ShowWindow(Win32Win(),SW_SHOW);}}
    return;}
  switch (parm) {
  case _rTop: case _rLeft:
    if (HostInfo==0) return;
    abit = wFlags&wiVisble;
    // rTop = HiWord(datum); // Rectify_MyW also does it from WM_MOVE
    // rLeft = LoWord(datum);
    wpt.x = 0; // H
    wpt.y = 0; // V
    ClientToScreen(Win32Win(),&wpt);
    GetWindowRect(Win32Win(),Win32Rect(&wnRect));
    if (abit !=0) ShowWindow(Win32Win(),SW_HIDE);
    MoveWindow(Win32Win(),LoWord(datum)-wpt.x+wnRect.rTop, // Win32 reversed T/L
        HiWord(datum)-wpt.y+wnRect.rLeft,
        wnRect.rBotm-wnRect.rTop,wnRect.rRite-wnRect.rLeft,false);
    _FrameWork_::Rectify_MyWin(Win32Win(),0);
    if (abit !=0) ShowWindow(Win32Win(),SW_SHOW);
    return;
  case _rBotm: case _rRite:
    if (HostInfo==0) return;
    abit = wFlags&wiVisble;
    wpt.x = 0;
    wpt.y = 0;
    ClientToScreen(Win32Win(),&wpt);
    vv = HiWord(datum);
    hh = LoWord(datum);
    // rBotm = vv+wpt.y; // actual (new) coordinates
    // rRite = hh+wpt.x; // NO, Rectify_MyWi does it from WM_SIZE
    // rTop = wpt.y;
    // rLeft = wpt.x;
    GetClientRect(Win32Win(),Win32Rect(&wnRect)); // retns TL=0,0; BR = act'l sz
    vv = vv-wnRect.rRite; // how much larger it needs to get..
    hh = hh-wnRect.rBotm; // W32 reversed B/R
    if ((vv|hh) ==0) return;
    GetWindowRect(Win32Win(),Win32Rect(&wnRect));
    if (abit !=0) ShowWindow(Win32Win(),SW_HIDE);
    MoveWindow(Win32Win(),wnRect.rTop,wnRect.rLeft, // W32 wnRect rev'd T/L,B/R
        wnRect.rBotm-wnRect.rTop+hh,wnRect.rRite-wnRect.rLeft+vv,false);
    _FrameWork_::Rectify_MyWin(Win32Win(),0);
    if (abit !=0) ShowWindow(Win32Win(),SW_SHOW);
    return;
  case _wKind:
    wKind = CHR4(datum);
    return;
  case _wName:
    wName = CHR4(datum);
    return;
  case _wFlags:
    wFlags = datum;
    return;
  case _FoBgColo:
    FoBgColo = datum;
    return;
  case _MoInfo:
    MoInfo = datum;
    return;
  case _wOwner:
    wOwner = CHR4(datum);
    return;
  case _wNxID:
    wNxID = CHR4(datum);
    return;
  case _wTextVH:
    wTextVH = datum;
    return;
  case _wUserData:
    wUserData = datum;
    return;
  case _wMoreInfo:
    wMoreInfo = datum;
    return;}} // ~Win_Set

// typedef _WindInfo_* Window;

_WindInfo_::_WindInfo_() {
  set_ary_size((void*)this,-(sizeof(_WindInfo_)/sizeof(int)-1));
  _FrameWork_::Next_Win_ID = Up4Char(_FrameWork_::Next_Win_ID);
  wName = _FrameWork_::Next_Win_ID;
  wNxID = "0000";
  wTextVH = Nowhere;
  wCoverR = Nowhere;}
_WindInfo_::_WindInfo_(String tx, Char4 kin, Char4 owner,
      WinWindow aWin = NULL, RectAngle rc = NULL) {
  set_ary_size((void*)this,-(sizeof(_WindInfo_)/sizeof(int)-1));
  _FrameWork_::Next_Win_ID = Up4Char(_FrameWork_::Next_Win_ID);
  wName = _FrameWork_::Next_Win_ID;
  wTextVH = Nowhere;
  wKind = kin;
  wOwner = owner;
  wNxID = "0000";
  wTextVH = Nowhere;
  theText = tx;
  Win32set(aWin);
  if (rc != NULL) {
    rTop = rc->rTop;
    rLeft = rc->rLeft;
    rBotm = rc->rBotm;
    rRite = rc->rRite;}
  else {
    rTop = 0;
    rLeft = 0;
    rBotm = 0;
    rRite = 0;}
  wWindo = NULL;
  wCoverR = Nowhere;
  if (aWin != NULL) Link_Me_In(kin);}
_WindInfo_::~_WindInfo_() { // never
  WinWindow aWin = Win32Win();
  Un_Link_Me(false);
  if (aWin) if (GetWindowLong(aWin,GWL_USERDATA)==0)
    DestroyWindow(aWin);}

void _WindInfo_::Link_Me_In(Char4 kin) {
  if (_FrameWork_::PIB_wins == NULL) {
    wWindo = NULL;
    _FrameWork_::PIB_wins = (Window)this;}
  else if (kin == "MBar") {
    wWindo = (Window)_FrameWork_::PIB_wins;
    _FrameWork_::PIB_wins = (Window)this;}
  else {
    wWindo = (Window)_FrameWork_::PIB_wins->wWindo;
    _FrameWork_::PIB_wins->wWindo = (Window)this;}} // ~Link_Me_In

void _WindInfo_::Un_Link_Me(bool Add2Free) {
  Window lnk = _FrameWork_::PIB_wins, prior = NULL;
  while (lnk != NULL) {
    if (lnk==this) {
      if (prior == NULL) _FrameWork_::PIB_wins = lnk->wWindo;
        else prior->wWindo = lnk->wWindo;
      break;}
    prior = lnk;
    lnk = lnk->wWindo;}
  wWindo = NULL;
  if (Add2Free) {
    wUserData = GetTime()+5;
    prior = _FrameWork_::Free_Wins;
    if (prior != NULL) {
      while (prior->wWindo != NULL) prior = prior->wWindo;
      prior->wWindo = (Window)this;}
    else _FrameWork_::Free_Wins = (Window)this;}
  _FrameWork_::unPacifyWin(this);} // ~Un_Link_Me

namespace  _FrameWork_ { // resume hiding..

_WindInfo_ _theDrag_; const Window theDrag = &_theDrag_; // hijack some fields

void Down_Win_Cnt(void) {
  // if (--Window_Count <=0) PostQuitMessage(0);
  if (--Window_Count > 0) return;
  if (!Sys_Trace) if (DebugLogFile) SystemDebugLog(String("-- 0Win"));
  Asyn_Msg("0Win",0,0,0,_FrameWork_::MainProgram);}

Window Next_Windo(Window theWin) {
  if (theWin==NULL) return PIB_wins;
  return theWin->wWindo;}

bool Valid_Window(Window theWin) {
  Window here = PIB_wins;
  while (here != NULL) {
    if (here==theWin) {
      if (here->wKind != "Wind") if (here->wKind != "Dlog") return false;
      return (here->HostInfo != 0);}
    here = here->wWindo;}
  return false;}

Window FindWin_by32(WinWindow w32) {
  Window here = PIB_wins;
  while (here != NULL) {
    if (here->Win32Win() == w32) return here;
    here = here->wWindo;}
  return NULL;}

WinWindow GetWgtWinWin(Widgit wgt) {
  if (wgt) return wgt->Win32Win();
  return NULL;} // ~GetWgtWinWin

void LinkMe_In(WindInfo wPtr, Char4 kin) {
  wPtr->Link_Me_In(kin);}
void UnLink_Me(WindInfo wPtr, bool A2F) { // called from WEV_CloseWind
  wPtr->Un_Link_Me(A2F);}
void Set_WinData(WindInfo wPtr, int datum, int here) {
  wPtr->Win_Set(datum,here);}
int Get_WinData(WindInfo wPtr, int here) {
  int datum;
  datum = wPtr->Win_Get(here);
  return datum;}

Widgit Iter8Wigit(Widgit theWg) { // next wigit in L-R depth-first order
  Widgit lnk = theWg->wChild; // theWg known !=0
  if (lnk==NULL) while (theWg != NULL) {
    lnk = theWg->wLink;
    if (lnk != NULL) break;
    theWg = theWg->wParent;}
  if (theWg != NULL) down__ptr(theWg); // caller up-ref'd
  if (lnk != NULL) return (Widgit)up__ptr(lnk);
  return NULL;} // ~Iter8Wigit

Widgit FindWgt_by32(Window theWin, WinWindow w32, Widgit sot = NULL) {
  Widgit theW = (Widgit)up__ptr(theWin->wChild);     // sot=NULL in only call
  while (theW != NULL) {
    if (sot==theW) return theW; // never hap
    if (theW->Win32Win() == w32) return theW;
    theW = Iter8Wigit(theW);}
  return NULL;} // ~FindWgt_by32

 } // (temply) end namespace  _FrameWork_

Char4 GetWindID(Window theWin) {
  if (theWin==NULL) return NoName;
  return theWin->wName; // if (Valid_Window(Window theWin)
  return NoName;}

String GetWinTitle(Window theWin) {
  if (theWin==NULL) return Null_Str;
  return theWin->theText;}

void SetWinTitle(Window theWin, String theTx) {
  if (theWin==NULL) return;
  if (theWin->theText != theTx) theWin->Set_Title(theTx);}

_Widgit_::_Widgit_(Window inW, Widgit par, String tx, Char4 kin,
      RectAngle rc = NULL, WinWindow aWin = NULL) {
  set_ary_size((void*)this,15);
  theText = tx;
  if (par != NULL) {
    if (inW == NULL) inW = par->wWindo;
    wParent = par;
    wLink = par->wChild;
    par->wChild = this;}
  if (inW != NULL) {
    inW->wNxID = Up4Char(inW->wNxID);
    wName = inW->wNxID;
    if (par == NULL) {
      wLink = inW->wChild;
      inW->wChild = this;}}
  wKind = kin;
  if (rc != NULL) {
    rTop = rc->rTop;
    rLeft = rc->rLeft;
    rBotm = rc->rBotm;
    rRite = rc->rRite;}
  else if (inW != NULL) {
    rBotm = inW->rBotm - inW->rTop;
    rRite = inW->rRite - inW->rLeft;}
  Win32set(aWin);
  wWindo = inW;}
_Widgit_::~_Widgit_() {
  Widgit par = NULL, lnk = wParent;
  if (wParent==NULL) lnk = (Widgit) wWindo;
  if (lnk) par = lnk->wChild;
  while (par) {
    if (par==this) {
      if (par==lnk->wChild) lnk->wChild = par->wLink;
        else lnk->wLink = par->wLink;
      break;}
    lnk = par;
    par = par->wLink;}} // _Widgit_

////////////////////////////////////////////////////////////////////////

// Utility Functions not elsewhere defined..

Char4 Unique_ProcessID(Char4 pnam4) { // used by compiled NewProgr
  Process allProgs = _FrameWork_::Ready_List;
  int bitz = ORD(pnam4), mask = (bitz&0x60606060)+0xE0E0E0E0
      | (((bitz^0xDFDFDFDF)+1)&0x7F7F7F7F)+0x1010101 | bitz+0x1010101 | bitz;
    mask = ((mask&0x80808080)<<1)-((mask>>7)&0x1010101); // FF in each non-char
  if (mask+1 ==0) pnam4 = "@000";
  else if (mask !=0) {
    if (mask<0) {
      bitz = (bitz&0xFFFFFF)+0x40000000; // "@" at top, if needed
      mask = mask&0xFFFFFF;}
    if ((mask&1) !=0) {
      bitz = MaskInts(bitz,0x303030,((mask+1)&(-mask-1))-1); // "000" at low end
      mask = mask&-((mask+1)&(-mask-1));}
    pnam4 = CHR4(MaskInts(bitz,0x5F5F5F5F,mask));} // "_" inside where needed
  if (allProgs != NULL) while (true) {
    if (allProgs->_my_ID_ == pnam4) {
      pnam4 = Up4Char(pnam4);
      allProgs = _FrameWork_::Ready_List;
      continue;}
    allProgs = allProgs->_link_;
    if (allProgs == _FrameWork_::Ready_List) break;}
  return pnam4;} //~Unique_ProcessID

int GetTime(void) { // returns seconds since midnight 2000 Jan 01
  int secs, days;
  SYSTEMTIME now;
  char* mo = " 01/001123344"; // cumulative, relative to '0' (not leap year)
    // assuming a 30-day month model, each character of mo[] encodes
    // the number of days it differs from the model; '0' exactly matches
  GetLocalTime(&now);
  days = (now.wYear-2001);
  days = (days+1)*365+(days>>2); // days since Y2K to Jan this year
  days = days+((int)mo[now.wMonth])+now.wMonth*30-78+now.wDay;
    // add +30 for each prior month in year, + current day + diff in mo[],
    // -48 [which is (int)'0'] -30 because wMonth is 1-based
    // -1 because wDay is 1-based, +1 after 2000 for Y2K leap day
  if (now.wMonth>2) if ((now.wYear&3)==0) days++; // correcting for leap year
  secs = days*86400+now.wHour*3600+now.wMinute*60+now.wSecond;
  return secs;}

int GetMilliSecs(void) { // returns milliseconds since midnight today
  SYSTEMTIME now;
  GetLocalTime(&now);
  return (now.wHour*3600+now.wMinute*60+now.wSecond)*1000+now.wMilliseconds;}

namespace  _FrameWork_ { // resume

#ifdef MOS_System // these following not needed otherwise

Process Next_Prog(Process whom) {
  if (whom==NULL) return Ready_List;
  return whom->_link_;} //~Next_Prog

// low-level functions used by Sortit.. -- this is #ifdef'd out if no System.h

inline void Exchange(int& Ap, int& Bp) {int med = Ap; Ap = Bp; Bp = med;}
inline void Byte_Store(int* ary, int Ap, char xCh) {
  char* theStr = (char*)ary;
  if (ary==NULL) return;
  if (ary[1]>0) ary[1]--; // doesn't dispose, but shouldn't need to
  theStr[Ap+8] = xCh;}
inline bool Dis_Order(int* ary, int Ap, int Bp) { // true if theStr[a]>theStr[b]
  unsigned char* theStr = (unsigned char*)ary;
  unsigned char xCh, zCh;
  if (ary==NULL) return false;
  if (ary[1]>0) ary[1]--; // doesn't dispose, but shouldn't need to
  Ap = Ap+8;
  Bp = Bp+8;
  while (true) {
    xCh = theStr[Ap++];
    if (xCh=='\0') return false;
    zCh = theStr[Bp++];
    if (xCh == zCh) continue;
    xCh = Capitize(xCh);
    zCh = Capitize(zCh);
    if (xCh != zCh) return (xCh>zCh);}}

////////////////////////////////////////////////////////////////////////

// [Part 2] D&D code adapted from <http://www.catch22.net/tuts/dragdrop.asp>

String GetW32DrData(IDataObject* pDataObject) {
  STGMEDIUM stgmo; // adapted from J.Brown's DropData in DropTarget.cpp
  FORMATETC fmtetc = { CF_TEXT, 0, DVASPECT_CONTENT, -1, TYMED_HGLOBAL };
  String strn; // See if the dataobject contains any TEXT stored as a HGLOBAL
  if (pDataObject->QueryGetData(&fmtetc) != S_OK) return Null_Str;
  if (pDataObject->GetData(&fmtetc, &stgmo) != S_OK) return Null_Str;
  // we asked for the data as a HGLOBAL, so access it appropriately
  strn = String((char*)GlobalLock(stgmo.hGlobal));
  GlobalUnlock(stgmo.hGlobal);
  ReleaseStgMedium(&stgmo); // release the data using the COM API
  return strn;}

void DropData(HWND theWin, IDataObject* pDataObject) {
  FORMATETC fmtetc = { CF_TEXT, 0, DVASPECT_CONTENT, -1, TYMED_HGLOBAL };
  STGMEDIUM stgmed;
  char* theData;
  // See if the dataobject contains any TEXT stored as a HGLOBAL
  if (pDataObject->QueryGetData(&fmtetc) == S_OK)
      if (pDataObject->GetData(&fmtetc, &stgmed) == S_OK) { // get the data
    // we asked for the data as a HGLOBAL, so access it appropriately
    theData = (char*)GlobalLock(stgmed.hGlobal);
    ///SetWindowText(theWin, (char*)theData);
    GlobalUnlock(stgmed.hGlobal);
    ReleaseStgMedium(&stgmed);}} // release the data using the COM API

class CDropTarget : public IDropTarget { // implements the IDropTarget interface
public:
  // IUnknown implementation
  HRESULT __stdcall QueryInterface (REFIID iid, void** ppvObject);
  ULONG __stdcall AddRef (void);
  ULONG __stdcall Release (void);

  // IDropTarget implementation
  HRESULT __stdcall DragEnter (IDataObject* pDataObject,
      DWORD grfKeyState, POINTL pt, DWORD* pdwEffect);
  HRESULT __stdcall DragOver (DWORD grfKeyState, POINTL pt, DWORD* pdwEffect);
  HRESULT __stdcall DragLeave (void);
  HRESULT __stdcall Drop (IDataObject* pDataObject, DWORD grfKeyState,
      POINTL pt, DWORD* pdwEffect);

  CDropTarget(HWND theWin); // Constructor
  ~CDropTarget();

private: // internal helper functions..
  DWORD DropEffct(DWORD grfKeyState, POINTL pt, DWORD dwAllowed);
  bool QueryDataObject(IDataObject* pDataObject);
  LONG DT_RefCount; // Private member variables..
  HWND DT_hWind;
  // Window DT_myWin;
  bool DT_fAllowDrop;
  IDataObject* DT_pDataObject;}; // ~CDropTarget

CDropTarget::CDropTarget(HWND theWin) {
  DT_RefCount = 1;
  DT_hWind = theWin;
  // DT_myWin = FindWin_by32(theWin);
  DT_fAllowDrop = false;}
CDropTarget::~CDropTarget() {}

void PositionCaret(HWND hwndEdit, POINTL pt) {
  // Position the edit control's caret under the mouse
  DWORD curpos; 
  ScreenToClient(hwndEdit, (POINT*)&pt); // get character position of mouse
  curpos = (DWORD) SendMessage(hwndEdit,EM_CHARFROMPOS,0,MAKELPARAM(pt.x,pt.y));
  // set caret position
  SendMessage(hwndEdit, EM_SETSEL, LOWORD(curpos), LOWORD(curpos));}

HRESULT __stdcall CDropTarget::QueryInterface (REFIID iid, void** ppvObject) {
  if (iid == IID_IDropTarget || iid == IID_IUnknown) {
    AddRef();
    *ppvObject = this;
    return S_OK;}
  else {
    *ppvObject = 0;
    return E_NOINTERFACE;}}

ULONG __stdcall CDropTarget::AddRef(void) {
  return InterlockedIncrement(&DT_RefCount);}  

ULONG __stdcall CDropTarget::Release(void) {
  LONG count = InterlockedDecrement(&DT_RefCount);
  if (count == 0) {
    delete this;
    return 0;}
  else return count;}

bool CDropTarget::QueryDataObject(IDataObject* pDataObject) { // helper
  FORMATETC fmtetc = { CF_TEXT, 0, DVASPECT_CONTENT, -1, TYMED_HGLOBAL };
  // does the data object support CF_TEXT using a HGLOBAL?
  return (pDataObject->QueryGetData(&fmtetc) == S_OK);}

// from <oleidl.h> :
//   #define	DROPEFFECT_NONE	( 0 )
//   #define	DROPEFFECT_COPY	( 1 )
//   #define	DROPEFFECT_MOVE	( 2 )
//   #define	DROPEFFECT_LINK	( 4 )

DWORD CDropTarget::DropEffct(DWORD grfKeyState, POINTL pt, DWORD dwAllowed) {
  DWORD dwEffect = 0; // helper function
  // 1. check "pt" -> do we allow a drop at the specified coordinates?
  //   --> determined by caller, true if we got here
  // 2. work out that the drop-effect should be based on grfKeyState
  if (grfKeyState & MK_CONTROL) dwEffect = dwAllowed & DROPEFFECT_COPY;
  else if (grfKeyState & MK_SHIFT) dwEffect = dwAllowed & DROPEFFECT_MOVE;
  // 3. no key-modifiers were specified (or drop effect not allowed),
  //    so base the effect on those allowed by the dropsource
  if (dwEffect == 0) {
    if (dwAllowed & DROPEFFECT_COPY) dwEffect = DROPEFFECT_COPY;
    if (dwAllowed & DROPEFFECT_MOVE) dwEffect = DROPEFFECT_MOVE;}
  return dwEffect;} // ~DropEffct

HRESULT __stdcall CDropTarget::DragEnter(IDataObject* pDataObject,
      DWORD grfKeyState, POINTL pt, DWORD* pdwEffect) { // pt is abs coord
  // does the dataobject contain data this window wants?
  DT_fAllowDrop = _System_::Analyze_DragStr(DT_hWind,GetW32DrData(pDataObject));
  if (DT_fAllowDrop)
      if (_System_::TrakDrag(DT_hWind,(Point)Pack2ints(pt.y,pt.x),null,0))
      if (ClipDragPkg[2+7]&1) { // get the drop effect based on keyboard state..
    *pdwEffect = DropEffct(grfKeyState, pt, *pdwEffect);
    SetFocus(DT_hWind);
    // PositionCaret(DT_hWind, pt);
    return S_OK;}
  *pdwEffect = DROPEFFECT_NONE;
  return S_OK;}

HRESULT __stdcall CDropTarget::DragOver(DWORD grfKeyState,
      POINTL pt, DWORD* pdwEffect) {
  if (DT_fAllowDrop)
      if (_System_::TrakDrag(DT_hWind,(Point)Pack2ints(pt.y,pt.x),null,0))
      if (ClipDragPkg[2+7]&1) {
    *pdwEffect = DropEffct(grfKeyState, pt, *pdwEffect);
    // PositionCaret(DT_hWind, pt);
    return S_OK;}
  *pdwEffect = DROPEFFECT_NONE;
  return S_OK;}

HRESULT __stdcall CDropTarget::DragLeave(void) {
  if (!DT_fAllowDrop) return S_OK;
  IgnoreBool(_System_::TrakDrag(DT_hWind,Nowhere,null,0));
  if (PkgPending==3) _System_::Zap_TheDrag();
  return S_OK;}

HRESULT __stdcall CDropTarget::Drop(IDataObject* pDataObject,
      DWORD grfKeyState, POINTL pt, DWORD* pdwEffect) {
  PositionCaret(DT_hWind, pt);
  if (DT_fAllowDrop)
      if (_System_::TrakDrag(DT_hWind,(Point)Pack2ints(pt.y,pt.x),null,0))
      if (ClipDragPkg[2+7]&1) {
    _System_::Drop_Here(DT_hWind); // DropData(DT_hWind, pDataObject);
    *pdwEffect = DropEffct(grfKeyState, pt, *pdwEffect);
    ClipDragPkg[2+7] = (ClipDragPkg[2+7]&~14)|((*pdwEffect&7)<<1);}
  else *pdwEffect = DROPEFFECT_NONE;
  return S_OK;}

void RegisterDropWindow(HWND theWin, IDropTarget** ppDropTarget) {
  CDropTarget* pDropTarget = new CDropTarget(theWin);
  CoLockObjectExternal(pDropTarget, TRUE, FALSE); // acquire a strong lock
  RegisterDragDrop(theWin, pDropTarget); // tell OLE the window is a drop target
  *ppDropTarget = pDropTarget;}

void UnregisterDropWindow(HWND theWin, IDropTarget* pDropTarget) {
  RevokeDragDrop(theWin); // remove drag+drop
  CoLockObjectExternal(pDropTarget, FALSE, TRUE); // remove the strong lock
  pDropTarget->Release();} // release our own reference

 } // (temporarily) end namespace  _FrameWork_ // end of J.Brown's code (140 LoC)

////////////////////////////////////////////////////////////////////////

// Win32 Registry Access.. -- note: this is #ifdef'd out if no System.h

String GetRegistryItem(String RegName, String ItemName) {
  HKEY RegItem = (HKEY)HKEY_CURRENT_USER;      // RegName: "\ProgName\path"
  bool ok = false;
  String datum;
  DWORD RegSize;
  if (CharAt(0,RegName) == '\\') {
    RegItem = (HKEY)HKEY_LOCAL_MACHINE;
    RegName = String("SOFTWARE") + RegName;}
  RegSize = RegOpenKeyEx(RegItem,RegName.text_(),0,KEY_ALL_ACCESS,&RegItem);
  if (RegSize == ERROR_SUCCESS) {
    if (RegQueryValueEx(RegItem,ItemName.text_(),0,NULL,NULL,&RegSize)
        == ERROR_SUCCESS) {
      datum = String(NULL,RegSize);
      ok = (RegQueryValueEx(RegItem,ItemName.text_(),0,NULL,
          (LPBYTE)datum.text_(),&RegSize) == ERROR_SUCCESS);}
    RegSize = RegCloseKey(RegItem);
    datum.valid_(); // make it so.
    if (ok) {
      if (CRLF_LEN == 2) datum = ReplacAll("","\r",datum);
      if (DebugLogFile) SystemDebugLog(String(" <GetReg> ")+RegName+","+ItemName
          + " = '"+Substring(0,777,datum)+"'");
      return datum;}}
  if (DebugLogFile) SystemDebugLog(String(" <GetReg> ")+RegName+","+ItemName
          + " (failed)");
  return Null_Str;}

void SaveRegistryItem(String datum, String RegName, String ItemName) {
  HKEY RegItem = (HKEY)HKEY_CURRENT_USER;      // RegName: "\ProgName\path"
  DWORD RegSize, Rx;
  if (CharAt(0,RegName) == '\\') {
    RegItem = (HKEY)HKEY_LOCAL_MACHINE;
    RegName = String("SOFTWARE") + RegName;}
  if (CRLF_LEN == 2) datum = ReplacAll(CR_LF,"\n",datum);
  if (DebugLogFile) SystemDebugLog(String(" <SaveReg> ")+RegName+","+ItemName
      +" = '"+datum+"'");
  Rx = RegCreateKeyEx(RegItem,RegName.text_(),0,NULL,0,KEY_ALL_ACCESS,
      NULL,&RegItem,&RegSize);
  if (Rx == ERROR_SUCCESS) {
    RegSize = RegSetValueEx(RegItem,ItemName.text_(),0,
        REG_SZ,(LPBYTE)datum.text_(),datum.length_()+1);
    RegSize = RegCloseKey(RegItem);}}

namespace  _FrameWork_ { // resume

void StartDragon(int flgs) { // (WinWindow theWin) // and stop when done
  int ix = 0, tmp;
  String theTxt;
  DWORD dwResult, dwEffect;
  int* thePkg = ClipDragPkg;
  FORMATETC formtetc = { CF_TEXT, 0, DVASPECT_CONTENT, -1, TYMED_HGLOBAL };
  STGMEDIUM storgmed = { TYMED_HGLOBAL, { 0 }, 0 };
  IDataObject* myDataObject;
  IDropSource* myDropSource;
  if (ClipDragPkg[3]<0) thePkg = (int*)int_2_ptr(ClipDragPkg[2]); // mine
  if (ClipDragStr == Null_Str) {
    tmp = _System_::PkgDataOffset(thePkg,ORD("+L&N"));
    ix = _System_::PkgDataOffset(thePkg,ORD("+Bbl")); // returns offset to type
    if (ix>0) ClipDragStr = FormBibleRef(thePkg[2+ix+1]);
    else if (tmp>0) {
      tmp = LoHalf(thePkg[2+tmp+1]);
      ix = tmp>>9;
      tmp = tmp&511;
      if (ix==94) {
        tmp = tmp+512;
        ix = 93;}
      ClipDragStr = String(ix)+String('.')+String(tmp);}
    else ClipDragStr = String("??");}
  if (Sys_Trace) if (DebugLogFile)
    SystemDebugLog(String("StartDrag ")+" #"+FormInt(PkgPending)
        + " "+FormInt(ClipDragPkg[1])+"/"+FormInt(ClipDragPkg[3])
        + FormTimeMS()+" '"+Substring(0,25,ClipDragStr)+"'");
  // if (theWin) SetCapture(theWin);
  Drag_Is_Busy = true;
  if (CRLF_LEN == 2) theTxt = ReplacAll(CR_LF,"\n",ClipDragStr);
    else theTxt = ClipDragStr;
  storgmed.hGlobal = StringToHandle(theTxt.text_());
  // ClipDragPkg[2+4] = LENGTH(ClipDragStr); // no longer used
  // ClipDragPkg[2+5] = ORD(Chr4Str(ClipDragStr));
  // theDrag.DrDataStr = Null_Str;
  CreateDropSource(&myDropSource);
  CreateDataObject(&formtetc,&storgmed,1,&myDataObject);
  try {
    // My_Test_Flag = true; // we get one initial call to QueryGetData
    dwResult = DoDragDrop(myDataObject,myDropSource,
        flgs|DROPEFFECT_COPY,&dwEffect); // DROPEFFECT_COPY=1, DROPEFFECT_MOVE=2
    if (dwResult == DRAGDROP_S_DROP) ix = dwEffect;}
  catch (...) {ix = -1;}
  myDataObject->Release();
  myDropSource->Release();
  // ReleaseCapture();
  ReleaseStgMedium(&storgmed); // cleanup
  Drag_Is_Busy = false;
  // if my drop is pending, can't dump data; else mark the pkg vacant..
  if (Sys_Trace|Drag_Trace) if (DebugLogFile)
    SystemDebugLog(String(" (StopDrag) ")+" #"+PkgPending+" "
       + FormInt(ClipDragPkg[1])+"/"+FormInt(ClipDragPkg[3])+FormTimeMS());
  if (PkgPending<3) _System_::Zap_TheDrag();} // ~StartDragon

void SysCacheInv(int rno) {_System_::CacheInvalid8(NULL,-1,rno,NULL);}

#else

// already (still) in _FrameWork_ namespace

void SysCacheInv(int rno) {}

#endif // MOS_System

////////////////////////////////////////////////////////////////////////

void Log_This_Evt(int msg) { // called also from Sys
  Logged_Events[(Ev_Log_Ix++)&(Event_Log-1)] = msg;
  Logged_Events[Ev_Log_Ix&(Event_Log-1)] = -1;} // ~Log_This_Evt

// MsgBox and AskText..

String Res_Str, Xfer_Str;

BOOL CALLBACK MyDlgProc(HWND theWin, UINT uMsg, WPARAM wp, LPARAM lp) {
  int ix;
  char DlgResBuf[256];
  Log_This_Evt(uMsg&0xFFF|0xD000);
  switch (uMsg) {
  case WM_INITDIALOG: // 0x0110
    Res_Str = LineOf(1,Xfer_Str);
    Xfer_Str = DeleteLineOf(1,Xfer_Str);
    SetDlgItemText(theWin,IDC_AskLabel,Xfer_Str.text_());
    SetDlgItemText(theWin,IDC_EdLn,Res_Str.text_());
    Signal32(_FrameWork_::GooLok);
    // SetWindowText(theWin,MsgBoxTitle); // dlog has no title
    SetFocus(GetDlgItem(theWin,IDC_EdLn));
    ix = LENGTH(Res_Str);
    if (ix>0) SendDlgItemMessage(theWin,IDC_EdLn,EM_SETSEL,0,ix);
    FrontedWin = theWin;
    BringWindowToTop(theWin);
    return false;
  case WM_COMMAND: // respond to button clicks.. // 0x0111
    switch (LOWORD(wp)) {
    case IDOK:
      ix = GetDlgItemText(theWin,IDC_EdLn,DlgResBuf,255);
      DlgResBuf[ix] = '\0';
      EndDialog(theWin,ix);
      Res_Str = String(DlgResBuf);
      FrontedWin = NULL;
      return true;
    case IDCANCEL:
      EndDialog(theWin,-1);
      FrontedWin = NULL;
      return true;
    default: return false;}}
  return false;} // ~MyDlgProc

 } // (temporarily) end namespace  _FrameWork_

String AskText(String msg, String txt, int opt) {
  boolean remote = (GetCurrentThreadId() != Gooey_Thread_ID);
  if (remote || opt >= 0) Wait32(_FrameWork_::GooLok);
  _FrameWork_::Xfer_Str = LineOf(1,txt)+"\n"+msg;
  if (remote) { // gotta do this in GUI thread..
    opt = Do_Gooey(WEV_AskTx,"AsTx",0,0,opt|HiBit); // releases, retakes GooLok
    txt = _FrameWork_::Xfer_Str;
    Signal32(_FrameWork_::GooLok);
    return txt;}
  if (DialogBox(_FrameWork_::Main_Context,MAKEINTRESOURCE(Ask_Text),
        _FrameWork_::theFrontWin,(DLGPROC)_FrameWork_::MyDlgProc) > 0)
    txt = _FrameWork_::Res_Str; // GooLok not retaken
  else txt = Null_Str;
  _FrameWork_::Fresh_Mods = true;
  if (opt<0) if (opt+opt >= 0) opt = opt&~HiBit;
  if (DebugLogFile) SystemDebugLog(String("? ")+opt+" '"+msg+"' = '"+txt+"'");
  return txt;} // ~AskText

int MsgBox(String msg, Char4 icn, int opt) {
  // Display msg with icon icn and buttons as specified in opt.
  // Return button # clicked by user
  int ix = 3, here = 8, tmp, bitz = 0, btnsAt = 0;
  if (GetCurrentThreadId() != Gooey_Thread_ID) { // gotta do this in GUI thread..
    Wait32(_FrameWork_::GooLok);
    _FrameWork_::Xfer_Str = msg;
    opt = Do_Gooey(WEV_MsgBx,"MsBx",ORD(icn),0,opt); // releases GooLok
    // Signal32(_FrameWork_::GooLok);
    return opt;}
  if (DebugLogFile) SystemDebugLog(String("!")+icn+" "+opt+" '"+msg+"'");
  if (opt==0) {
    bitz = 1;
    btnsAt = 1;}
  else while (ix>0) { // ix is the btn position: 1..3
    tmp = (((opt>>here)&7)-1); // msg: OK=0, Cancel=1, No=2, Yes=3
    if ((tmp&3)==tmp) if ((bitz&(1<<tmp))==0) {
      bitz = bitz+(1<<tmp); // got this msg
      btnsAt = btnsAt+(ix<<(tmp<<1));} // btn posn of this msg
    ix--;
    here = here-4;}
  if ((bitz&14)==14) {
    here = MB_YESNOCANCEL;
    if (tmp==1) here = here+MB_DEFBUTTON3;
    else if (tmp==2) here = here+MB_DEFBUTTON2;}
  else if ((bitz&3)==3) {
    here = MB_OKCANCEL;
    if (tmp==1) here = here+MB_DEFBUTTON2;}
  else if ((bitz&12)==12) {
    here = MB_YESNO;
    if (tmp==2) here = here+MB_DEFBUTTON2;}
  else here = MB_OK; // if (bitz&1) else return 0;
  if (icn==QuestIcon) here = here+MB_ICONQUESTION;
  else if (icn==StopIcon) here = here+MB_ICONHAND+MB_TASKMODAL;
  else if (icn==InfoIcon) here = here+MB_ICONEXCLAMATION;
  else if (icn>NoName) here = here+MB_ICONASTERISK;
  _FrameWork_::Fresh_Mods = true;
  _FrameWork_::New_MsgBx = true;
  _FrameWork_::Log_This_Evt(0xDBFF);
  // using theFrontWin keeps MsBx on top even after DoTree fronts the tree win..
  here = MessageBox(_FrameWork_::theFrontWin,msg.text_(),MsgBoxTitle,here);
  _FrameWork_::Log_This_Evt(here|0xDB00);               // ProgramName
  _FrameWork_::New_MsgBx = false;
  _FrameWork_::FrontedWin = NULL;
  switch (here) {
    case IDOK: return btnsAt&3;
    case IDCANCEL: return (btnsAt>>2)&3;
    case IDYES: return (btnsAt>>6)&3;
    case IDNO: return (btnsAt>>4)&3;}
  return 0;} // ~MsgBox

////////////////////////////////////////////////////////////////////////

// Win32 event support..

void SetCursor(int whom, Window win, RectAngle rc) { // my MOS version..
  if (win != NULL) if (whom >= 0) {                  // calls Win32 version..
    win->FoBgColo = Pack2ints(whom,win->FoBgColo);
    _FrameWork_::Cursor_Win = NULL;
    if (Drag_Trace) if (DebugLogFile) SystemDebugLog(String("  *SetCurs* ")
        + String(whom)+String(" ")+String(GetWindID(win)));
    return;}
  if (_FrameWork_::my_Cursor == whom) return;
  if ((whom&0xF0000000) == HiBit) whom = LoHalf(whom); // from Sys, don't log it
  else if (Drag_Trace) if (DebugLogFile)
    SystemDebugLog(String("  *SetCurs* ")+String(whom)
        + String("/")+String(_FrameWork_::my_Cursor));
  _FrameWork_::my_Cursor = whom;
  if (whom<0) _FrameWork_::OldCursor = _FrameWork_::HourGlass;
  else if (whom==0) _FrameWork_::OldCursor = _FrameWork_::PlainArrow;
  else if (whom==1) _FrameWork_::OldCursor = _FrameWork_::Ibeam;
  else if (whom==2) _FrameWork_::OldCursor = _FrameWork_::CrossHair;
  else if (whom==3) _FrameWork_::OldCursor = _FrameWork_::DiagonalAro;
  else if (whom==4) _FrameWork_::OldCursor = _FrameWork_::EastWest;
  else if (whom==5) _FrameWork_::OldCursor = _FrameWork_::NorthSouth;
    else _FrameWork_::OldCursor = _FrameWork_::PlainArrow;
  SetCursor(_FrameWork_::OldCursor);}

namespace  _FrameWork_ { // resume

void Zap_Cursor() {
  _FrameWork_::Cursor_Win = NULL;
  if (_FrameWork_::my_Cursor > 0) {
    _FrameWork_::my_Cursor = 0;
    _FrameWork_::OldCursor = _FrameWork_::PlainArrow;}
  SetCursor(_FrameWork_::PlainArrow);} // ~Zap_Cursor

void Refresh_Modifiers() {
  if ((GetKeyState(VK_CAPITAL)&1) !=0)
    Modifier_Keys = Modifier_Keys|CapsLockKey;
  else Modifier_Keys = Modifier_Keys&~CapsLockKey;
  if (GetKeyState(VK_SHIFT)<0) Modifier_Keys = Modifier_Keys|ShiftKey;
    else Modifier_Keys = Modifier_Keys&~ShiftKey;
  if (GetKeyState(VK_CONTROL)<0) Modifier_Keys = Modifier_Keys|ControlKey;
    else Modifier_Keys = Modifier_Keys&~ControlKey;
  if (GetKeyState(VK_APPS)<0) Modifier_Keys = Modifier_Keys|AppMenuKey;
    else Modifier_Keys = Modifier_Keys&~AppMenuKey;
  Fresh_Mods = false;} //~Refresh_Modifiers

unsigned char vKeyStates[256];
// UINT OEMscanCode; // not needed (VS DB broken)

char WinKey2ascii(char theKey) {
  WORD nuKeys[4];
  nuKeys[0] = 0;
  nuKeys[1] = 0;
  nuKeys[2] = 0;
  nuKeys[3] = 0;
  vKeyStates[theKey&255] = 255;
  if ((Modifier_Keys&ShiftKey) !=0) vKeyStates[VK_SHIFT] = 255;
  if (ToAscii((UINT)theKey,0,(PBYTE)vKeyStates,nuKeys,0) != 1)
    nuKeys[0] = 0;
  vKeyStates[VK_SHIFT] = 0;
  vKeyStates[theKey&255] = 0;
  return (char)(nuKeys[0]&255);}

boolean ShowHourGlass(WinWindow theWin) {
  Window myWin = FindWin_by32(theWin);
  if (myWin == NULL) return false;
  if (myWin->wMoreInfo !=0) SetCursor(HourGlass);
  else if (OldCursor) SetCursor(OldCursor);
    else return false;
  return true;} // ~ShowHourGlass

LRESULT CantPaint(HWND theWin, PAINTSTRUCT* Ppst, Char4 why) {
  HDC graf = BeginPaint(theWin,Ppst);
  EndPaint(theWin,Ppst);
    if (DebugLogFile) SystemDebugLog(String("? ---- CantPaint ")+why);
  return 0;} // ~CantPaint

LRESULT CALLBACK Wind_WinProc(HWND theWin, UINT msg, WPARAM wp, LPARAM lp) {
  int ix = 0, nx = 0, zx, hor, ver, wPrm = (int)wp;
  HDC graf;
  PAINTSTRUCT pst;
  DRAWITEMSTRUCT* dis;
  POINT wpt;
  Widgit wgt;
  Char4 whom;
  Window myWin;
  Process theProc;
  String theStr;
  unsigned char* TcanPtr;
  char ctitle[64];
  _RectAngle_ wnRect; // RECT wnRect;
  set_ary_size((void*)(&wnRect),-4);
  Log_This_Evt(msg); // Logged_Events[(Ev_Log_Ix++)&(Event_Log-1)] = msg;
  // Logged_Events[Ev_Log_Ix&(Event_Log-1)] = -1; // visual flag to mark edge
  if (msg>Windo_Event) if (msg<WEV_NoHigher) {
    // Valid8_Event((Event)lp);
    return (LRESULT)Do_WinEvent(msg,(Event)lp);}
  switch (msg) {
  case WM_COMMAND: // mostly menu stuff here.. // 0x0111
    if (lp==0) { // menu
      ix = wp-Menu_Comm_base;
      if (Pending_Event != NULL) if (ix<0) { // process waiting for popup..
        // Valid8_Event(Pending_Event);
        Pending_Event->EvInfo = -1-ix;
        Pending_Event = NULL;
        return 0;}
      // if (Popup_Result<0) { // process waiting for popup..
      //   if (ix<0) { // this is a popup selection..
      //     Popup_Result = -ix;
      //     return 0;}
      //   Popup_Result = 0;}
      if (ix<0) break;
      if (ix >= Num_Menus) break;
      myWin = FindWin_by32(theWin);
      if (myWin==NULL) break;
      whom = Menu_Comms[2+ix]; // .say;
      Do_Menu_Comm(whom,myWin,Proc4ch(myWin->wOwner));}
    else if (HIWORD(wp) == BN_CLICKED) {
      myWin = FindWin_by32(theWin);
      if (myWin==NULL) break;
      wgt = FindWgt_by32(myWin,(WinWindow)lp); // up-ref'd
      // wgt = FindThisWgt(wp&0xFFFF,"",myWin);
      if (wgt==NULL) break;
      if (wgt->wFlags&wManaged) if (myWin->wNxID == "PrgB") {
        down__ptr(wgt);
        Hiden_Die(myWin);
        return 0;}
      ClikPt = Mouse_Locn; // cuz we don't get a click
      ix = HiWord(Modifier_Keys)&31;
      whom = GetWindID(myWin);
      if (ix !=0) Asyn_Msg("ModK",(int)ClikPt,ix,ORD(whom),
          Proc4ch(myWin->wOwner));
      if (!Sys_Trace) if (DebugLogFile) SystemDebugLog(String("-- Clik ")
          + FormInt((int)ClikPt)+" "+wgt->wName+" "+whom+" "+myWin->wOwner);
      Asyn_Msg("Clik",(int)ClikPt,ORD(wgt->wName),ORD(whom),
          Proc4ch(myWin->wOwner));
      down__ptr(wgt);}
    else break;
    return 0;
  case WM_CHAR: // got data key (ignored for now) /// ---- (TCHAR) for unicode
    if (Pending_Event != NULL) PopRelease(0); // huh? lost menu
    break;
  case WM_KEYDOWN: // got keystroke
    // OEMscanCode = lp; // unneeded
    if (!Do_KeyDown(theWin,wp&255)) break;
    return 0;
  case WM_KEYUP: // modifiers: lock:0, shft:1, ctrl:2
    ix = wp&255;
    if (ix==VK_SHIFT) Modifier_Keys = Modifier_Keys&~ShiftKey;
    else if (ix==VK_CONTROL) Modifier_Keys = Modifier_Keys&~ControlKey;
    else if (ix==VK_LWIN) Modifier_Keys = Modifier_Keys&~WinKey;
    else if (ix==VK_RWIN) Modifier_Keys = Modifier_Keys&~WinKey;
    else if (ix==VK_APPS) Modifier_Keys = Modifier_Keys&~AppMenuKey;
    else if (ix==VK_CAPITAL) if ((GetKeyState(VK_CAPITAL)&1) ==0)
      Modifier_Keys = Modifier_Keys&~CapsLockKey;
    break;
  case WM_ACTIVATE: // got new front window..
    if (Pending_Event != NULL) PopRelease(0); // huh? lost menu
    myWin = FindWin_by32(theWin);
    if (myWin==NULL) break;
    Windo_Activ8(myWin,wPrm);
    // SetFocus(theWin); // makes everything fail
    break;
//case WM_MENURBUTTONUP: // dunno when these happen
//  break;
  case WM_MENUSELECT: // if selected, this happens before WM_EXITMENULOOP
    ix = Menu_Comm_base-1-LoHalf(wPrm);
    // if (Menu_Preselect>Menu_Hi_Chek) ... (deleted, didn't work very well
    break;
  case WM_ENTERMENULOOP:
    Menu_Is_Open = true;
    Zap_Cursor();
    break;
  case WM_EXITMENULOOP: // this happens when menu done, before WM_COMMAND
    Menu_Is_Open = false;
    if (Pending_Event != NULL) Pending_Event->EvDatum = 0; // PopReleas looks
    // if (Pending_Event != NULL) PopRelease(0);
    break;
#ifdef MOS_System
  case WM_NCLBUTTONDOWN: // click in "non-content" (frame/scrollbar)..
    Mouse_Buttons = Mouse_Buttons|HiBit; // client needs this
    wpt.x = LOWORD(lp); // capture possible scrollbar thumb position
    wpt.y = HIWORD(lp);
    // ClientToScreen(theWin,&wpt); // already is
    ScroRefLoc = (Point)Pack2ints(wpt.y,wpt.x); // absolute
    // wpt.x = 0; // save also window origin (unneed, Sys can get it)
    // wpt.y = 0;
    // ClientToScreen(theWin,&wpt);
    break; // ..then let default happen
  case WM_NCMOUSEMOVE:
    _System_::Frame_Cursor(theWin,lp);
    break;
  case WM_NCRBUTTONDOWN:
    Mouse_Buttons = Mouse_Buttons|0x20000000;
    break;
  case WM_NCMBUTTONDOWN:
    Mouse_Buttons = Mouse_Buttons|0x40000000;
    break;
//#else
//  Zap_Cursor();
#endif // MOS_System
  case WM_MOUSEMOVE:
  case WM_LBUTTONDOWN:  case WM_LBUTTONUP:   case WM_LBUTTONDBLCLK:
  case WM_RBUTTONDOWN:  case WM_RBUTTONUP:   case WM_RBUTTONDBLCLK:
  case WM_MBUTTONDOWN:  case WM_MBUTTONUP:   case WM_MBUTTONDBLCLK:
  case WM_NCLBUTTONUP:  case WM_NCRBUTTONUP: case WM_NCMBUTTONUP:
    switch (msg) {
    // case WM_MOUSEMOVE:
    //   nx = 0; // already =0
    //   break;
    case WM_LBUTTONDOWN:  case WM_RBUTTONDOWN:  case WM_MBUTTONDOWN:
      nx = 1;
      break;
    case WM_LBUTTONDBLCLK:  case WM_RBUTTONDBLCLK:  case WM_MBUTTONDBLCLK:
      nx = 2;
      break;
    case WM_LBUTTONUP:   case WM_RBUTTONUP:   case WM_MBUTTONUP:
    case WM_NCLBUTTONUP: case WM_NCRBUTTONUP: case WM_NCMBUTTONUP:
      nx = -1;
      break;}
    wpt.x = LOWORD(lp);
    wpt.y = HIWORD(lp);
    ClientToScreen(theWin,&wpt);
    Mouse_Locn = (Point)Pack2ints(wpt.y,wpt.x);
    if (wp&MK_CONTROL) ix = ControlKey;
    if (wp&MK_SHIFT) ix = ix+ShiftKey;
    if (wp&MK_LBUTTON) ix = ix+HiBit;
    if (wp&MK_RBUTTON) ix = ix+0x20000000;
    if (wp&MK_MBUTTON) ix = ix+0x40000000;
    // if (ix !=0) if (Pending_Event != NULL) PopRelease(0); // huh? lost menu
    Mouse_Buttons = (Mouse_Buttons&(0xFFFFFFF-ControlKey-ShiftKey))+ix;
    if ((Mouse_Buttons&0x08000000) !=0) break; // CxMn sent & pending
    // if ((Mouse_Buttons&0xE0000000)==0) {
    //   // if ((theDrag.wFlags&wiVisble) !=0) Zap_Time = 2;
    //   if (false) if (nx==0) break;} // no longer ignore unclicked rollovers
    if (nx>0) { // any kind of click..
      ClikPt = Mouse_Locn; // absolute
      if (Tix_100ms<0) Tix_100ms = -1;} // if (theWin==GooeyWin)
    // else if (nx<0) { // msg = WM_LBUTTONUP/WM_MBUTTONUP/WM_RBUTTONUP
    //   if (false) if ((theDrag->wFlags&wiVisble) !=0) Zap_Time = 3;}
    if (FlushClix>0) if ((Mouse_Buttons&0xE0000000) !=0) break;
    if (!Mouse_InDocX(theWin,nx)) break;
    return 0;
  case WM_VSCROLL:
    if (Pending_Event != NULL) PopRelease(0); // huh? lost menu
    if (!Scroll_Some(theWin,SB_VERT,(int)wp)) break;
    return 0;
  case WM_HSCROLL:
    if (Pending_Event != NULL) PopRelease(0); // huh? lost menu
    if (!Scroll_Some(theWin,SB_HORZ,(int)wp)) break;
    return 0;
  case WM_MOVE:
    if (Pending_Event != NULL) PopRelease(0); // huh? lost menu
    myWin = Rectify_MyWin(theWin,0);
    if (myWin==NULL) break;
    Asyn_Msg("MovW",0,0,ORD(GetWindID(myWin)),Proc4ch(myWin->wOwner));
    return 0;
  case WM_SIZE:
    if (Pending_Event != NULL) PopRelease(0); // huh? lost menu
    if (Windo_Resized(theWin)) return 0;
    break;
  case WM_CLOSE:
    if (Pending_Event != NULL) PopRelease(0); // huh? lost menu
    myWin = FindWin_by32(theWin);
    if (myWin==NULL) break;
    if (!Sys_Trace) if (DebugLogFile) SystemDebugLog(String("-- CloW ")
        + GetWindID(myWin)+" "+myWin->wOwner);
    Asyn_Msg("CloW",0,0,ORD(GetWindID(myWin)),Proc4ch(myWin->wOwner));
    return 0;
  case WM_PAINT: // this happens after a menu is released almost always
    if (Pending_Event != NULL) PopRelease(-1);
    myWin = FindWin_by32(theWin);
    if (myWin==NULL) break; /// return CantPaint(theWin,&pst,"Win0");
    if (myWin->wNxID != DocX4) if (myWin->wNxID != Imag4)
      if (myWin->wNxID != TxtD4) break;
    wgt = myWin->Get_Wgt(_wChild); // now up-ref'd
    if (wgt==NULL) break; /// return CantPaint(theWin,&pst,"fdw0");
    if (!Windo_Painting(theWin,&wnRect)) {
      down__ptr(wgt);
      break;}
    // if (myWin->wMoreInfo !=0) un_Pacify(myWin);
    ver = wnRect.rLeft; // = vert scroll in pixels; BR not used
    hor = wnRect.rTop; // W32 rev'd T/L
    zx = ScreenTall-1; // needed for hor>0; else SetDIBitsToDev fails to copy
    ix = ScreenWide-hor; // effective right edge of image
    MyBitMap.BitMapStuff.biWidth = ScreenWide;
    MyBitMap.BitMapStuff.biHeight = zx; // ScreenTall;
    graf = BeginPaint(theWin,&pst);
    GetClientRect(theWin,Win32Rect(&wnRect)); // retns TL=0,0; BR = actual size
    // brusk = (HBRUSH) GetStockObject(WHITE_BRUSH); // dunno if this leaks mem?

    if (wnRect.rBotm>ix) {
      wnRect.rTop = ix;
      FillRect(graf,Win32Rect(&wnRect),StockWhiteBrush);
      wnRect.rTop = 0;}
    wnRect.rLeft = zx-ver; // used to fill bottom if needed
    //  if (DebugLogFile) SystemDebugLog(String("  SetDIBits ")+wnRect.rRite+","
    //    + ScreenTall+"..."+DIB_RGB_COLORS);
    ix = SetDIBitsToDevice(graf,0,0,ix,zx,0,0,ver,zx,
        (unsigned char*)(theScreen+hor+8),theBitMap,DIB_RGB_COLORS);
        // +8 skips over array overhead; +hor does scroll, but wraps off end
    if (wnRect.rTop<wnRect.rBotm)
      FillRect(graf,Win32Rect(&wnRect),StockWhiteBrush);
    // if (ix==0) ix = GetLastError();
    nx = (wnRect.rBotm&-4)-18;
    zx = wnRect.rRite-17;
    // if (fdw->HotSpots != NULL) if (fdw->HotHit !=0)
    if (Get_HotRect(theWin,wgt,&wnRect)) InvertRect(graf,Win32Rect(&wnRect));
    // down__ptr(wgt); // Get_HotRec did
    ix = myWin->wFlags;
    if ((ix&(wiDragble|wiDropble)) == (wiDragble|wiDropble))
        if (myWin->wNxID == Imag4) { // add trashcan..
      if (ix&(1<<wiUsrBitShft)) TcanPtr = (unsigned char*)(FullCan+8);
        else TcanPtr = (unsigned char*)(eTrsCan+8);
      MyBitMap.BitMapStuff.biWidth = 16;
      MyBitMap.BitMapStuff.biHeight = 16;
      ix = SetDIBitsToDevice(graf,nx,zx,16,16,0,0,
          0,16,TcanPtr,theBitMap,DIB_RGB_COLORS);}
    EndPaint(theWin,&pst);
      ///if (DebugLogFile) SystemDebugLog(String("**"));
    unPacifyWin(myWin);
    return 0;
  case WM_DRAWITEM:
    myWin = FindWin_by32(theWin);
    if (myWin==NULL) break;
    if (myWin->wNxID != "PrgB") break;
    dis = (DRAWITEMSTRUCT*)lp;
    // graf = dis->hDC;
    // wnRect = dis->rcItem;
    ix = dis->rcItem.left;
    nx = dis->rcItem.right;
    wnRect.rBotm = nx; // Win32 reverses T/L, B/R..
    wnRect.rLeft = dis->rcItem.top;
    wnRect.rRite = dis->rcItem.bottom;
    nx = Get_Progress_Pix(myWin,nx-ix);
    wnRect.rTop = nx+ix;
    if (wnRect.rTop<wnRect.rBotm) // this is all about horiz posn..
      FillRect(dis->hDC,Win32Rect(&wnRect),PreProgBrush);
    wnRect.rBotm = wnRect.rTop;
    wnRect.rTop = ix;
    if (wnRect.rTop<wnRect.rBotm)
      FillRect(dis->hDC,Win32Rect(&wnRect),ProgressBrush);
    FrameRect(dis->hDC,&(dis->rcItem),StockBlackPen);
    return (LRESULT)TRUE; // (LRESULT)?
  case WM_SETCURSOR:
    if (ShowHourGlass(theWin)) return (LRESULT)TRUE;
    break;
  case WM_DROPFILES:
    //if (DragQueryFile((HDROP)wp,-1,NULL,0)>0)
    //  if (DragQueryFile((HDROP)wp,0,aStr,99)>0)
    //    MessageBox(NULL,aStr,AppName,0);
    return 0;
  case WM_TIMER:
    if (Pending_Event != NULL) PopRelease(-1); // in case nobody else notices
    Tix_100ms++;
    See_Panic(ORD("Time"));
    See_Panic(Tix_100ms);
    Tenth_Seconds++;
#ifdef MOS_System
    if ((Tenth_Seconds&3)==0) // if (theFrontWin != NULL)
      _System_::Carat_Blink_Chk(true);
    if (n_Timers) _System_::Delaying();
    // Tix_Now++;
    // if (Tix_Now >= 36000*24) _System_::RollDate();
    if ((Mouse_Buttons&0xE8000000) == 0x08000000) {
      Mouse_Buttons = Mouse_Buttons&~0x08000000; // so not send CxMn twice
      _System_::Clear_Hilites();}
#endif // MOS_System
    myWin = Pacify_wins;
    if (myWin) if (myWin->wMoreInfo !=0) {
      ix = GetMilliSecs();
      if (ix>myWin->wMoreInfo) {
        if (DebugLogFile)
          SystemDebugLog(String("  ... ")+theStr+FormTimeMS());
        theProc = Proc4ch(myWin->wOwner);       // if (!(theProc->Izzy_Busy()))
        if (false) if (Valid_Process(theProc)) if (myWin->wNxID == Imag4)
          Asyn_Msg("Imag",0,0,ORD(GetWindID(myWin)),theProc); // got lost
        myWin->wMoreInfo = MaskInts(ix+1536,myWin->wMoreInfo,255);
        ix = myWin->wMoreInfo&255;
        theStr = GetWinTitle(myWin);
        zx = LENGTH(theStr);
        if (zx==ix) {
          theStr = theStr + " ..";
          ShowHourGlass(myWin->Win32Win());}
        else theStr = theStr + ".";
        myWin->Set_Title(theStr);
        if (zx>ix+33) {
          if (!Sys_Trace) if (DebugLogFile) SystemDebugLog(String("-- CloW .. ")
              + GetWindID(myWin)+" "+myWin->wOwner);
          Asyn_Msg("CloW",0,0,ORD(GetWindID(myWin)),Proc4ch(myWin->wOwner));}}}
    if (Tix_100ms<0) if (Tix_100ms+Tix_100ms>0) Tix_100ms = 1;
    if (Tix_100ms==0) {
      if (Window_Count>0) {
        FrontedWin = NULL;
        myWin = FindWin_by32(GooeyWin);
        if (myWin != NULL) myWin->wFlags = myWin->wFlags&~wiVisble;
        ShowWindow(GooeyWin,SW_HIDE);
        Asyn_Msg("SplX",0,0,0,_FrameWork_::MainProgram);}
      else Tix_100ms = -1;}
    if (FlushClix>0) FlushClix--;
    return 0;
  case WM_SHOWWINDOW: // catch MsgBx on its way by (MB doesn't come here)..
    if (!New_MsgBx) break;
    if (GetWindowText(theWin,ctitle,60) != 12) break;
    if (strcmp(ctitle,MsgBoxTitle) !=0) break;
    Log_This_Evt(0xDB99);
    FrontedWin = theWin;
    New_MsgBx = false;
    break;
  case WM_CREATE:
  case WM_DESTROY:
    break;} // let default happen
  return DefWindowProc(theWin,msg,wp,lp);} // ~Wind_WinProc

////////////////////////////////////////////////////////////////////////

void StartGooey() {
  if (GooeyWin) return;
  strcpy(MsgBoxTitle,ProgramName);
  strcat(MsgBoxTitle," Message");
#ifdef Splash // #defined in Fonts.h, which is #included by System.h
  GooeyWin = _System_::Create_Splash((int*)Splasher);
  HasSplash = GooeyWin;
  FrontedWin = _FrameWork_::GooeyWin;
#endif
  if (GooeyWin == NULL) GooeyWin = CreateWindow(ProgramName,ProgramName,
      WS_OVERLAPPEDWINDOW,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,
      32,NULL,NULL,Main_Context,NULL);
  if (GooeyWin != NULL)
    SetTimer(GooeyWin,1,100,(TIMERPROC)NULL);} //~StartGooey

boolean InitializeAll(HINSTANCE hx, bool& GotOle) {
  int ix, rx = 0, gx = 0, bx = 0, colo;
  Gooey_Thread_ID = GetCurrentThreadId();
  NewSem32(MoveOK,1," MoveOK");
  NewSem32(StrnLok,1," StrnLok");
  NewSem32(ResFiLok,1," ResFiLok");
  NewSem32(DebugLok,1," DebugLok");
  NewSem32(PanicLok,1," PanicLok");
  NewSem32(EvenLok,1," EvenLok");
  NewSem32(GooLok,1," GooLok");
  Init_MOMAM();
  set_ary_size((void*)eTrsCan,-(sizeof(eTrsCan)/sizeof(int)-1));
  set_ary_size((void*)FullCan,-(sizeof(FullCan)/sizeof(int)-1));
  MyBitMap.BitMapStuff.biSize = sizeof(BITMAPINFOHEADER);
  MyBitMap.BitMapStuff.biPlanes = 1;
  MyBitMap.BitMapStuff.biBitCount = 8;
  MyBitMap.BitMapStuff.biCompression = BI_RGB;
  MyBitMap.BitMapStuff.biSizeImage = 0;
  MyBitMap.BitMapStuff.biXPelsPerMeter = 0;
  MyBitMap.BitMapStuff.biYPelsPerMeter = 0;
  MyBitMap.BitMapStuff.biClrUsed = 216;
  MyBitMap.BitMapStuff.biClrImportant = 0;
  for (ix=0; ix<256; ix++) vKeyStates[ix] = 0;
  for (ix=0; ix<216; ix++) { // initialize color table..
    colo = (((((rx*51)<<8)+gx*51)<<8)+bx*51);
    if (bx>0) if (bx<5) colo = colo+25; // brighten the gamma a little..
    if (gx>0) if (gx<5) colo = colo+(25<<8);
    if (rx>0) if (rx<5) colo = colo+(25<<16);
    MyBitMap.ColorTbl[ix] = colo;
    rx++;
    if (rx==6) {
      gx++;
      rx = 0;
      if (gx==6) {
        bx++;
        gx = 0;}}}
  HourGlass = LoadCursor(NULL,IDC_WAIT);
  CrossHair = LoadCursor(NULL,IDC_CROSS);
  Ibeam = LoadCursor(NULL,IDC_IBEAM);
  PlainArrow = LoadCursor(NULL,IDC_ARROW);
  EastWest = LoadCursor(NULL,IDC_SIZEWE);
  NorthSouth = LoadCursor(NULL,IDC_SIZENS);
  DiagonalAro = LoadCursor(NULL,IDC_SIZENWSE);
  Wind_type.style = 0;
  Wind_type.lpfnWndProc = Wind_WinProc;
  Wind_type.cbClsExtra = 0;
  Wind_type.cbWndExtra = 0;
  Wind_type.hInstance = hx;
  //Wind_type.hIcon = LoadIcon(NULL,IDI_APPLICATION);
  Wind_type.hIcon = LoadIcon(hx,MAKEINTRESOURCE(App_Icon));
  Wind_type.hCursor = PlainArrow;
  Wind_type.hbrBackground = (HBRUSH) GetStockObject(LTGRAY_BRUSH);
  Wind_type.lpszMenuName = NULL;
  Wind_type.lpszClassName = ProgramName;
  if (RegisterClass(&Wind_type) ==0) return false;
  StockWhiteBrush = (HBRUSH) GetStockObject(WHITE_BRUSH);
  StockBlackPen = (HBRUSH) GetStockObject(BLACK_PEN);
  PreProgBrush = (HBRUSH) GetStockObject(LTGRAY_BRUSH);
  ProgressBrush = (HBRUSH) GetStockObject(DKGRAY_BRUSH);
  Modifier_Keys = 0;
  Refresh_Modifiers();
  NoName = "    "; // dunno why init constructor fails, but this works..
  DocX4 = "DocX";
  Imag4 = "Imag";
  PrgB4 = "PrgB";
  TxtD4 = "TxtD";
  Typ4 = "Type";
  Next_Win_ID = "#000";
#ifdef Debug_Trace
#ifdef Debug_Trace_File
  DebugLogFile = Open_In_Dir(Debug_Trace_File,WriteOnly,NULL);
#else
  DebugLogFile = Open_In_Dir(Debug_Trace,WriteOnly,NULL);
#endif
  if (DebugLogFile)
    SystemDebugLog(String("---- ")+FormDateTime(GetTime(),0)+String(" ----"));
#endif
  GotOle = (OleInitialize(0) == S_OK); // Initialize COM and OLE
#ifdef MOS_System // clip stuff not defined otherwise
  ClipDragPkg[0] = ClpDrgPkSz;
  FILL0(ClipDragPkg,0,ClpDrgPkSz);
  _System_::System_Inits((int*)eTrsCan,(int*)FullCan); // in System.h
#else
  StartGooey();
#endif // MOS_System
  return true;} // ~InitializeAll

 } // (temporarily) end namespace  _FrameWork_

void _T2F_Process_::T2F_Quit() {
  if (_my_ID_ == Null4) return;
  while (_FrameWork_::Fetch_Event(NULL,this)) {} // release waiting evs
  _my_ID_ = Null4;
  if (_FrameWork_::MainProgram == this) {
    _FrameWork_::MainProgram = NULL;
    ExitProcess(0);}
  else ExitThread(0);} //~T2F_Quit

_T2F_Process_::~_T2F_Process_() {T2F_Quit();} // destructor no longer used

////////////////////////////////////////////////////////////////////////

namespace _FrameWork_ { // _FrameWork_::xxx

int* _tmp_0_; int* _tmp_00_;
int __aix__, __aiy__, __aiz__;









  int PIBsysErr = 0;

bool IsAlphaNum(char ch)
 { int* _tmp_00_; bool _tmp__;
  if ((ch < '0'))  { return 0;}
  if ((ch <= '9'))  { return 1;}
  if ((ch > 'z'))  { return 0;}
  if ((ch < 'A'))  { return 0;}
  if ((ch == '_'))  { return 1;}
  if ((ch < 'a'))  if ((ch > 'Z'))  { return 0;}
  { return 1;}} // ~IsAlphaNum

int Tokenize(int kin, int orn, String stx, int theLn)
 { int* _tmp_00_; int _tmp__;
  int ordn = MAX(orn,-1);
  int here = _len_(stx);
  int ptln = 0;
  int offs = 0;
  int xch;
  int sep;
  if ((here == 0))  { _tmp__ = 0; return _tmp__;}
  if ((ordn > orn))  orn = 0;
  if ((kin == (-2)))  kin = 10;
  sep = (kin&127);
  here = 0;
  xch = ORD((char)CharAt(0, stx));
  while (1) {
  if ((kin <= 0))  while ((xch <= 32)) {
  if ((xch == 0))  break;
  here++;
  xch = ORD((char)CharAt(here, stx));}
  if ((xch == 0)) {
  if ((ordn < 0))  { _tmp__ = MIN((orn-ordn),here); return _tmp__;}
  offs = here;
  break;}
  ordn--;
  if ((ordn == 0))  offs = here;
  if ((kin == 0))  while ((xch > 32)) {
  here++;
  xch = ORD((char)CharAt(here, stx));}
  else  if ((kin < 0)) {
  if (IsAlphaNum(CHR((xch&255))))  while (IsAlphaNum(CHR((xch&255)))) {
  here++;
  xch = ORD((char)CharAt(here, stx));}
  else {
  here++;
  xch = ORD((char)CharAt(here, stx));}}
  else  while ((xch != sep)) {
  here++;
  xch = ORD((char)CharAt(here, stx));
  if ((xch == 0))  break;}
  if ((ordn == 0)) {
  ptln = (here-offs);
  break;}
  if ((xch == 0)) {
  if ((ordn < 0))  if ((orn == 0))  orn = -1;
  continue;}
  if ((kin < 0))  continue;
  here++;
  xch = ORD((char)CharAt(here, stx));}
  if ((theLn != 0))  _POKE_((int)theLn,(int)ptln);
  { _tmp__ = offs; return _tmp__;}} // ~Tokenize

String Replace_Del(String NuSt, int offs, int nch, String theSt, bool del2)
 { int* _tmp_00_; String _tmp__;
  int ln1 = _len_(theSt);
  int ln2 = _len_(NuSt);
  if ((offs < 0))  offs = 0;
  if ((nch < 0))  nch = 0;
  if (del2) {
  if ((offs > 0))  offs--;
  nch++;}
  if ((ln1 == 0))  { _tmp__ = NuSt; return _tmp__;}
  if ((offs >= ln1))  { _tmp__ = _cat_(theSt, NuSt); return _tmp__;}
  if ((nch == 0))  if ((offs == 0))
    { _tmp__ = _cat_(NuSt, theSt); return _tmp__;}
  if ((offs == 0))  if ((nch >= ln1))  { _tmp__ = NuSt; return _tmp__;}
  nch = MIN(nch,(ln1-offs));
  if (((ln2+nch) == 0))  { _tmp__ = theSt; return _tmp__;}
  { _tmp__ = _cat_(_cat_(Substring(0, offs, theSt), NuSt), Substring((offs+nch), ln1, theSt)); return _tmp__;}} // ~Replace_Del

String Pos_Time(String prefx)
 { int* _tmp_00_; String _tmp__;
  int now = ((GetMilliSecs()/100)%36000);
  prefx = _cat_(_cat_(_cat_(prefx, String((int)(now/600))), String(":")), String((int)((now%600)+1000)));
  now = _len_(prefx);
  prefx = Replace_Del(Null_Str, (now-4), 1, prefx, 0);
  { _tmp__ = Replace_Del(String("."), (now-2), 0, prefx, 0); return _tmp__;}} // ~Pos_Time

String FormTimeMS()
 { int* _tmp_00_; String _tmp__;
  { return Pos_Time(String(" -- "));}}

void Log_Str_Cash(int whom, bool doWhat)
 { int* _tmp_00_;
  int ix;
  int best = -1;
  if ((_String_Log_ > 1))  if (DebugLogFile)
    if ((_FrameWork_::_Strings_Count_ >= 0)) {
  Wait32(StrnLok);
  for (  ix = (_FrameWork_::_Strings_Count_-1); !(ix < 0);  ix--)  {
  if ((_FrameWork_::the_String_Log[2+(ix)] == 0))  best = ix;
  else  if ((_FrameWork_::the_String_Log[2+(ix)] == whom)) {
  if ((!doWhat))  _FrameWork_::the_String_Log[2+(ix)] = 0;
  Signal32(StrnLok);
  { return;}}}
  if ((best >= 0))  if (doWhat)  _FrameWork_::the_String_Log[2+(best)] = whom;
  Signal32(StrnLok);}} // ~Log_Str_Cash

void Monitor_Strings(bool doWhat)
 { int* _tmp_00_;
  int ix = _FrameWork_::_Strings_Count_;
  int thar = _FrameWork_::Ev_Log_Ix;
  int ref;
  int nby;
  int* whom = 00;
  String aStr;
  if (DebugLogFile) {
  if ((Event_Log > 1)) {
  aStr = _cat_(_cat_(String("=== Events: "), String((int)thar)), String("   "));
  {  int _tmp_1_ = thar;
  for (  ref = MAX(((thar+1)-Event_Log),0); !(ref > _tmp_1_);  ref++)  {
  if (((((ref-1)-thar)&15) == 0))  aStr = _cat_(aStr, String("\n "));
  aStr = _cat_(_cat_(aStr, String(" ")), FormDateTime(_FrameWork_::Logged_Events[2+((ref&(Event_Log-1)))], 0x80000004));}}
  SystemDebugLog(_cat_(aStr, String("\n===")));}
  if ((_String_Log_ > 1))  while (1) {
  if (doWhat) {
  if ((_FrameWork_::_Strings_Count_ < 0))  _FrameWork_::_Strings_Count_ = 0;
  SystemDebugLog(_cat_(String("==== String Cache on.."), String((int)_FrameWork_::_Strings_Count_)));}
  else  if ((_FrameWork_::_Strings_Count_ < 0))  break;
  else  _FrameWork_::_Strings_Count_ = -1;
  while ((ix > 0)) {
  ix--;
  ref = _FrameWork_::the_String_Log[2+(ix)];
  if ((ref == 0))  continue;
  if ((!doWhat))  _FrameWork_::the_String_Log[2+(ix)] = 0;
  nby = _PEEK_((int)ref);
  {_tmp_00_ = (int*)whom;  whom = int_2_ptr((ref+4)); down__ptr(_tmp_00_);}
  aStr = TxtBlok2Str(8, MIN((nby&0xFFFFFF),44), (int*)up__ptr(whom));
  SystemDebugLog(_cat_(_cat_(_cat_(_cat_(
    _cat_(String(" === Cached: "), String((int)ix)), String("+")), String((int)_PEEK_((int)(ref+8)))),
   String(" ")), aStr));
  whom = 0;}
  if (doWhat)  SystemDebugLog(String(".."));
  else  SystemDebugLog(String("==== String Cache off"));
  break;}
  _FrameWork_::Exit_MOMAM();}
  down__ptr((int*)whom);} // ~Monitor_Strings

 
 } // (temporarily) end namespace  _FrameWork_

String Replace(String NuSt, int offs, int nch, String theSt)
 { int* _tmp_00_; String _tmp__;
  { _tmp__ = _FrameWork_::Replace_Del(NuSt, offs, nch, theSt, 0); return _tmp__;}}

String WordOf(int ordn, String theSt)
 { int* _tmp_00_; String _tmp__;
  int sz = 0;
  int here = _FrameWork_::Tokenize(0, ordn, theSt, ((int)(&sz)));
  if ((sz > 0))  { _tmp__ = Substring(here, sz, theSt); return _tmp__;}
  { _tmp__ = Null_Str; return _tmp__;}}

String LineOf(int ordn, String theSt)
 { int* _tmp_00_; String _tmp__;
  int sz = 0;
  int here = _FrameWork_::Tokenize(-2, ordn, theSt, ((int)(&sz)));
  if ((sz > 0))  { _tmp__ = Substring(here, sz, theSt); return _tmp__;}
  { _tmp__ = Null_Str; return _tmp__;}}

String TokenOf(int ordn, String theSt)
 { int* _tmp_00_; String _tmp__;
  int sz = 0;
  int here = _FrameWork_::Tokenize(-1, ordn, theSt, ((int)(&sz)));
  if ((sz > 0))  { _tmp__ = Substring(here, sz, theSt); return _tmp__;}
  { _tmp__ = Null_Str; return _tmp__;}}

String ItemOf(char sep, int ordn, String theSt)
 { int* _tmp_00_; String _tmp__;
  int sz = 0;
  int here = _FrameWork_::Tokenize(ORD((char)sep), ordn, theSt, ((int)(&sz)));
  if ((sz > 0))  { _tmp__ = Substring(here, sz, theSt); return _tmp__;}
  { _tmp__ = Null_Str; return _tmp__;}}

String PutWordOf(String NuSt, int ordn, String theSt)
 { int* _tmp_00_; String _tmp__;
  int sz = 0;
  int here = _FrameWork_::Tokenize(0, ordn, theSt, ((int)(&sz)));
  { _tmp__ = _FrameWork_::Replace_Del(NuSt, here, sz, theSt, 0); return _tmp__;}}

String PutLineOf(String NuSt, int ordn, String theSt)
 { int* _tmp_00_; String _tmp__;
  int sz = 0;
  int here = _FrameWork_::Tokenize(-2, ordn, theSt, ((int)(&sz)));
  { _tmp__ = _FrameWork_::Replace_Del(NuSt, here, sz, theSt, 0); return _tmp__;}}

String PutTokenOf(String NuSt, int ordn, String theSt)
 { int* _tmp_00_; String _tmp__;
  int sz = 0;
  int here = _FrameWork_::Tokenize(-1, ordn, theSt, ((int)(&sz)));
  { _tmp__ = _FrameWork_::Replace_Del(NuSt, here, sz, theSt, 0); return _tmp__;}}

String PutItemOf(String NuSt, char sep, int ordn, String theSt)
 { int* _tmp_00_; String _tmp__;
  int sz = 0;
  int here = _FrameWork_::Tokenize(ORD((char)sep), ordn, theSt, ((int)(&sz)));
  { _tmp__ = _FrameWork_::Replace_Del(NuSt, here, sz, theSt, 0); return _tmp__;}}

String InsertB4Word(String NuSt, int ordn, String theSt)
 { int* _tmp_00_; String _tmp__;
  int sz = 0;
  int here = _FrameWork_::Tokenize(0, ordn, theSt, ((int)(&sz)));
  { _tmp__ = _FrameWork_::Replace_Del(NuSt, here, 0, theSt, 0); return _tmp__;}}

String InsertB4Line(String NuSt, int ordn, String theSt)
 { int* _tmp_00_; String _tmp__;
  int sz = 0;
  int here = _FrameWork_::Tokenize(-2, ordn, theSt, ((int)(&sz)));
  { _tmp__ = _FrameWork_::Replace_Del(NuSt, here, 0, theSt, 0); return _tmp__;}}

String InsertB4Token(String NuSt, int ordn, String theSt)
 { int* _tmp_00_; String _tmp__;
  int sz = 0;
  int here = _FrameWork_::Tokenize(-1, ordn, theSt, ((int)(&sz)));
  { _tmp__ = _FrameWork_::Replace_Del(NuSt, here, 0, theSt, 0); return _tmp__;}}

String InsertB4Item(String NuSt, char sep, int ordn, String theSt)
 { int* _tmp_00_; String _tmp__;
  int sz = 0;
  int here = _FrameWork_::Tokenize(ORD((char)sep), ordn, theSt, ((int)(&sz)));
  { _tmp__ = _FrameWork_::Replace_Del(NuSt, here, 0, theSt, 0); return _tmp__;}}

String DeleteWordOf(int ordn, String theSt)
 { int* _tmp_00_; String _tmp__;
  int sz = 0;
  int here = _FrameWork_::Tokenize(0, ordn, theSt, ((int)(&sz)));
  { _tmp__ = _FrameWork_::Replace_Del(Null_Str, here, sz, theSt, 1); return _tmp__;}}

String DeleteLineOf(int ordn, String theSt)
 { int* _tmp_00_; String _tmp__;
  int sz = 0;
  int here = _FrameWork_::Tokenize(-2, ordn, theSt, ((int)(&sz)));
  { _tmp__ = _FrameWork_::Replace_Del(Null_Str, here, sz, theSt, 1); return _tmp__;}}

String DeleteTokenOf(int ordn, String theSt)
 { int* _tmp_00_; String _tmp__;
  int sz = 0;
  int here = _FrameWork_::Tokenize(-1, ordn, theSt, ((int)(&sz)));
  { _tmp__ = _FrameWork_::Replace_Del(Null_Str, here, sz, theSt, 1); return _tmp__;}}

String DeleteItemOf(char sep, int ordn, String theSt)
 { int* _tmp_00_; String _tmp__;
  int sz = 0;
  int here = _FrameWork_::Tokenize(ORD((char)sep), ordn, theSt, ((int)(&sz)));
  { _tmp__ = _FrameWork_::Replace_Del(Null_Str, here, sz, theSt, 1); return _tmp__;}}

int nWords(String theSt)
 { int* _tmp_00_; int _tmp__;
  { _tmp__ = _FrameWork_::Tokenize(0, -1, theSt, 0); return _tmp__;}}

int nTokens(String theSt)
 { int* _tmp_00_; int _tmp__;
  { _tmp__ = _FrameWork_::Tokenize(-1, -1, theSt, 0); return _tmp__;}}

int nLines(String theSt)
 { int* _tmp_00_; int _tmp__;
  { _tmp__ = _FrameWork_::Tokenize(-2, -1, theSt, 0); return _tmp__;}}

int n0Lines(String theSt)
 { int* _tmp_00_; int _tmp__;
  { _tmp__ = _FrameWork_::Tokenize(-2, -2, theSt, 0); return _tmp__;}}

int nItems(char sep, String theSt)
 { int* _tmp_00_; int _tmp__;
  { _tmp__ = _FrameWork_::Tokenize(ORD((char)sep), -1, theSt, 0); return _tmp__;}}

int n0Items(char sep, String theSt)
 { int* _tmp_00_; int _tmp__;
  { _tmp__ = _FrameWork_::Tokenize(ORD((char)sep), -2, theSt, 0); return _tmp__;}}

bool ParseBool(String stx)
 { int* _tmp_00_; bool _tmp__;
  int here = 0;
  int back = _len_(stx);
  char xch;
  while (1) {
  if ((here >= back))  { _tmp__ = 0; return _tmp__;}
  xch = CharAt(here, stx);
  here++;
  switch (xch) {
case 32:
case 44:
case 10:
case 13:
case 9:
  continue;
case 84:
case 116:
case 49:
  { _tmp__ = 1; return _tmp__;}
default:
  { _tmp__ = 0; return _tmp__;}
  } //~switch
  }} // ~ParseBool

int ParseInt(String stx)
 { int* _tmp_00_; int _tmp__;
  int parm = 0;
  int ndigs = 0;
  int here = 0;
  int sgn = 1;
  int nx = 5;
  int back = _len_(stx);
  char xch;
  while (1) {
  if ((here >= back))  break;
  xch = CharAt(here, stx);
  here++;
  switch (xch) {
case 32:
case 44:
case 10:
case 13:
case 9:
  if ((ndigs != 0))  break;
  continue;
case 45:
  if ((ndigs != 0))  break;
  sgn = -1;
  continue;
case 88:
case 120:
  if ((parm != 0))  break;
  if ((nx == 8))  break;
  nx = 8;
  continue;
case 48:
case 49:
case 50:
case 51:
case 52:
case 53:
case 54:
case 55:
case 56:
case 57:
  parm = (((parm << 1)*nx)+(ORD((char)xch)&15));
  ndigs++;
  continue;
case 65:
case 66:
case 67:
case 68:
case 69:
case 70:
case 97:
case 98:
case 99:
case 100:
case 101:
case 102:
  if ((nx == 5))  break;
  parm = (((parm << 4)+(ORD((char)xch)&15))+9);
  ndigs++;
  continue;
default:
  break;  } //~switch
    break;}
  if ((parm > 0))  if ((sgn < 0))  { _tmp__ = (-parm); return _tmp__;}
  { _tmp__ = parm; return _tmp__;}} // ~ParseInt

int Offset(String sot, String instr)
 { int* _tmp_00_; int _tmp__;
  int sotl = _len_(sot);
  int insl = _len_(instr);
  if ((sotl == 0))  { _tmp__ = 0; return _tmp__;}
  if ((insl < sotl))  { _tmp__ = -1; return _tmp__;}
  { _tmp__ = _FrameWork_::FindText(EXTRACT_Cstr(sot), sotl, EXTRACT_Cstr(instr), insl); return _tmp__;}} // ~Offset

String LastItem(String name, char delim)
 { int* _tmp_00_; String _tmp__;
  int nitms = _FrameWork_::Tokenize(ORD((char)delim), -1, name, 0);
  int sz = 0;
  int here = _FrameWork_::Tokenize(ORD((char)delim), (nitms-1), name, ((int)(&sz)));
  if ((nitms < 2))  { return name;}
  { _tmp__ = Substring(((here+sz)+1), _len_(name), name); return _tmp__;}} // ~LastItem

String ReplacAll(String nust, String seek, String theStr)
 { int* _tmp_00_; String _tmp__;
  int offs = 0;
  int olen = _len_(seek);
  int nlen = _len_(nust);
  int mxln;
  if ((olen == 0))  { _tmp__ = theStr; return _tmp__;}
  while (1) {
  mxln = _len_(theStr);
  if (((offs+olen) > mxln))  offs = -1;
  else  offs = FindStTxBlk(seek, offs, mxln, Text_Ary_Str(theStr));
  if ((offs < 0))  { _tmp__ = theStr; return _tmp__;}
  theStr = Replace(nust, offs, olen, theStr);
  offs = (offs+nlen);}} // ~ReplacAll

 
namespace  _FrameWork_ { // resuming

bool IsValidFile(File aFile)
 { int* _tmp_00_; bool _tmp__;
  File here = _FrameWork_::Next_File(0);
  if ((aFile == 0))  { return 0;}
  while ((here != 0)) {
  if ((here == aFile))  { return 1;}
  here = _FrameWork_::Next_File(here);}
  { return 0;}} // ~IsValidFile

File Find_Directry(String path, bool Sharon)
 { int* _tmp_00_; File _tmp__;
  int bitz;
  File aFile = 0;
  while (1) {
  aFile = Next_File(aFile);
  if ((aFile == 0))  { _tmp__ = 0; return _tmp__;}
  bitz = GetFileFlags(aFile);
  if (((bitz&64) != 0))  continue;
  if (((bitz&24) != 16)) {
  if ((!Sharon))  continue;
  if (((bitz&152) == 0))  continue;}
  if ((FullPath(aFile) == path))
    { _tmp__ = aFile; return _tmp__;}}} // ~Find_Directry

File Open_Fail(String why)
 { int* _tmp_00_; File _tmp__;
  if (DebugLogFile)  SystemDebugLog(_cat_(String("  -- fail: "), why));
  { _tmp__ = 0; return _tmp__;}} // ~Open_Fail

File Open_In_Dir(String name, int opt, File dir)
 { int* _tmp_00_; File _tmp__;
  int here;
  int sz;
  int nn;
  int nitms = 0;
  int abit = 0;
  File aFile = 0;
  WinFile wfi = 0;
  String thePath = Null_Str;
  if ((opt == 0))  opt = 0x400;
  if ((dir == 0))  dir = PIBwDir;
  if (DebugLogFile) {
  if ((dir != 0)) {
  thePath = FullPath(dir);
  nitms = GetFiRefNum(dir);}
  SystemDebugLog(_cat_(_cat_(_cat_(_cat_(
    _cat_(_cat_(_cat_(String("OpenFi "), String((int)opt)), String(" '")),
   name), String("' +")), String((int)nitms)), String(" ")), thePath));}
  if ((Offset(String("\\"), name) > 0)) {
  here = Tokenize(92, 1, name, ((int)(&sz)));
  thePath = Substring(0, sz, name);
  if ((thePath == Null_Str)) {
  if ((dir == 0))  thePath = String("C:");
  else {
  thePath = FullPath(dir);
  here = Tokenize(92, 1, thePath, ((int)(&sz)));
  thePath = Substring(0, sz, thePath);}
  name = _cat_(thePath, name);}
  else  if ((thePath == Two_Dots)) {
  if ((dir == 0))  { _tmp__ = Open_Fail(Two_Dots); return _tmp__;}
  thePath = FullPath(dir);
  nitms = Tokenize(92, 0, thePath, 0);
  here = Tokenize(92, (nitms-1), thePath, ((int)(&sz)));
  thePath = Substring(0, (here+sz), thePath);
  name = _cat_(thePath, name);}
  else  if ((thePath == One_Dot)) {
  if ((dir == 0))  { _tmp__ = Open_Fail(One_Dot); return _tmp__;}
  thePath = FullPath(dir);
  name = _cat_(thePath, name);}
  nitms = Tokenize(92, 0, name, 0);
  here = Tokenize(92, nitms, name, ((int)(&sz)));
  nitms = (here+sz);
  thePath = Substring(0, nitms, name);
  name = Substring((nitms+1), _len_(name), name);
  dir = _FrameWork_::Find_Directry(thePath, 0);
  if ((dir == 0)) {
  dir = Make_File(LastItem(thePath, '\\'), thePath, 0, 16);
  if ((dir == 0))  { _tmp__ = Open_Fail(String("new1")); return _tmp__;}}
  if ((_len_(name) == 0)) {
  if (DebugLogFile) {
  nitms = GetFiRefNum(dir);
  opt = GetFileFlags(aFile);
  thePath = FullPath(dir);
  SystemDebugLog(_cat_(_cat_(_cat_(_cat_(
    _cat_(String("  = "), String((int)nitms)), String(":")), String((int)opt)),
   String(" ")), thePath));}
  { _tmp__ = dir; return _tmp__;}}}
  if ((dir == 0)) {
  dir = Make_File(Null_Str, GetCurWinDirName(), 0, 16);
  if ((dir == 0))  { _tmp__ = Open_Fail(String("new2")); return _tmp__;}
  PIBwDir = dir;}
  else  if ((!IsValidFile(dir)))
    { _tmp__ = Open_Fail(String("IsVal")); return _tmp__;}
  nn = GetFileFlags(dir);
  if ((name == Null_Str)) {
  if (((opt&64) == 0))  if (((nn&64) == 0))
    { _tmp__ = Open_Fail(String("ref")); return _tmp__;}
  thePath = FullPath(dir);
  name = GetFileName(dir);}
  else  if (((nn&24) != 16))
    { _tmp__ = Open_Fail(_cat_(String("dir="), String((int)nn))); return _tmp__;}
  else  thePath = _cat_(_cat_(FullPath(dir), String("\\")), name);
  if (((opt&152) == 152)) {
  aFile = _FrameWork_::Find_Directry(thePath, 1);
  if ((aFile != 0)) {
  if (DebugLogFile) {
  nitms = GetFiRefNum(aFile);
  thePath = FullPath(aFile);
  opt = GetFileFlags(aFile);
  SystemDebugLog(_cat_(_cat_(_cat_(_cat_(
    _cat_(String("  sh "), String((int)nitms)), String(":")), String((int)opt)),
   String(" ")), thePath));}
  { _tmp__ = aFile; return _tmp__;}}}
  if (((opt&16) != 0)) {
  aFile = _FrameWork_::Find_Directry(thePath, 0);
  if ((aFile != 0)) {
  if (DebugLogFile) {
  nitms = GetFiRefNum(aFile);
  thePath = FullPath(aFile);
  opt = GetFileFlags(aFile);
  SystemDebugLog(_cat_(_cat_(_cat_(_cat_(
    _cat_(String("  dir "), String((int)nitms)), String(":")), String((int)opt)),
   String(" ")), thePath));}
  { _tmp__ = aFile; return _tmp__;}}}
  if ((Next_File(0) == 0))
    { _tmp__ = Open_Fail(String("PIBfi")); return _tmp__;}
  if (((opt&24) != 16))  if (((opt&64) == 0)) {
  if (((opt&128) != 0))  abit = FILE_ATTRIBUTE_HIDDEN;
  else  abit = FILE_ATTRIBUTE_NORMAL;
  nitms = (opt&7);
  here = 0;
  switch (nitms) {
case 0:
  nn = OPEN_ALWAYS;
  here = (GENERIC_READ|GENERIC_WRITE);
  abit = (abit|FILE_FLAG_RANDOM_ACCESS);
  break;
case 1:
  nn = OPEN_ALWAYS;
  here = GENERIC_READ;
  break;
case 2:
  nn = CREATE_ALWAYS;
  here = GENERIC_WRITE;
  break;
case 3:
  nn = OPEN_ALWAYS;
  here = GENERIC_WRITE;
  break;
case 4:
  nn = OPEN_EXISTING;
  here = (GENERIC_READ|GENERIC_WRITE);
  abit = (abit|FILE_FLAG_RANDOM_ACCESS);
  break;
case 5:
  nn = OPEN_EXISTING;
  here = GENERIC_READ;
  break;
case 6:
  nn = TRUNCATE_EXISTING;
  here = GENERIC_WRITE;
  break;
case 7:
  nn = OPEN_EXISTING;
  here = GENERIC_WRITE;
  break;  } //~switch
    wfi = CreateFile(EXTRACT_Cstr(thePath), here, 0, 0, nn, abit, 0);
  if ((wfi == INVALID_HANDLE_VALUE)) {
  nn = GetLastError();
  File_Destructor(aFile);
  { _tmp__ = Open_Fail(_cat_(String("Win32err="), String((int)nn))); return _tmp__;}}
  if ((nitms == 7))
    opt = Int_CAST((int*)SetFilePointer(wfi, 0, 0, FILE_END));}
  aFile = Make_File(LastItem(thePath, '\\'), thePath, wfi, opt);
  if ((aFile == 0))  { _tmp__ = Open_Fail(String("new3")); return _tmp__;}
  nitms = GetFiRefNum(aFile);
  thePath = FullPath(aFile);
  opt = GetFileFlags(aFile);
  if (DebugLogFile) {
  if ((abit == 0))  name = String("  rf ");
  else  name = String("  fi ");
  SystemDebugLog(_cat_(_cat_(_cat_(_cat_(
    _cat_(name, String((int)nitms)), String(":")), String((int)opt)), String(" ")),
   thePath));}
  { _tmp__ = aFile; return _tmp__;}} // ~Open_In_Dir

 
 } // (temporarily) end namespace  _FrameWork_

int GetFilePosn(File aFile)
 { int* _tmp_00_; int _tmp__;
  int bitz = GetFileFlags(aFile);
  if ((bitz == 0))  { return -1;}
  if (((bitz&24) != 16))  if (((bitz&64) == 0))
    { return Int_CAST((int*)SetFilePointer(_FrameWork_::Get_File_w32(aFile),
   0, 0, FILE_CURRENT));}
  { return -1;}} // ~GetFilePosn

int GetSizeFile(File aFile)
 { int* _tmp_00_; int _tmp__;
  int bitz = GetFileFlags(aFile);
  if ((bitz == 0))  { return -1;}
  if (((bitz&24) != 16))  if (((bitz&64) == 0))
    { return Int_CAST((int*)GetFileSize(_FrameWork_::Get_File_w32(aFile), 0));}
  { return -1;}} // ~GetSizeFile

int GetFileErr()
 { int* _tmp_00_; int _tmp__;
  { return _FrameWork_::PIBsysErr;}}

bool FileAtEOF(File aFile)
 { int* _tmp_00_; bool _tmp__;
  if ((!_FrameWork_::IsValidFile(aFile)))  { return 0;}
  { return (GetFilePosn(aFile) == GetSizeFile(aFile));}} // ~FileAtEOF

int FiSize_If_0(File aFile, int nby)
 { int* _tmp_00_; int _tmp__;
  if ((nby > 0))  { return nby;}
  { return MAX((GetSizeFile(aFile)-GetFilePosn(aFile)),0);}} // ~FiSize_If_0

File CloneFileRef(File aFile, Process frm)
 { int* _tmp_00_; File _tmp__;
  if ((aFile == 0))  { return 0;}
  { return _FrameWork_::Open_In_Dir(Null_Str, 64, aFile);}} // ~CloneFileRef

void SetEOF(File aFile)
 { int* _tmp_00_;
  if ((!_FrameWork_::IsValidFile(aFile)))  _FrameWork_::PIBsysErr = -254;
  else  if (((GetFileFlags(aFile)&24) == 16))  _FrameWork_::PIBsysErr = -252;
  else  if (((GetFileFlags(aFile)&64) != 0))  _FrameWork_::PIBsysErr = -252;
  else  if (SetEndOfFile(_FrameWork_::Get_File_w32(aFile)))
    _FrameWork_::PIBsysErr = 0;
  else  _FrameWork_::PIBsysErr = -252;} // ~SetEOF

void Seek(File aFile, int offs)
 { int* _tmp_00_;
  if ((!_FrameWork_::IsValidFile(aFile)))  _FrameWork_::PIBsysErr = -254;
  else  if (((GetFileFlags(aFile)&24) == 16))  _FrameWork_::PIBsysErr = -252;
  else  if (((GetFileFlags(aFile)&64) != 0))  _FrameWork_::PIBsysErr = -252;
  else  if ((Int_CAST((int*)SetFilePointer(_FrameWork_::Get_File_w32(aFile),
   offs, 0, FILE_BEGIN)) < 0))  _FrameWork_::PIBsysErr = -252;
  else  _FrameWork_::PIBsysErr = 0;} // ~Seek

File Close2Ref(File aFile)
 { int* _tmp_00_; File _tmp__;
  if ((!_FrameWork_::IsValidFile(aFile)))  { return 0;}
  if ((_FrameWork_::Using_ResFi == aFile))  _FrameWork_::Using_ResFi = 0;
  if (DebugLogFile)
    SystemDebugLog(_cat_(String("ClosFi +"), String((int)GetFiRefNum(aFile))));
  if (((GetFileFlags(aFile)&64) != 0))  { return aFile;}
  _FrameWork_::Win32FindClose(aFile);
  if (((GetFileFlags(aFile)&24) == 24))
    _FrameWork_::SysCacheInv((GetFiRefNum(aFile) << 16));
  if (((GetFileFlags(aFile)&24) != 16))
    if (((GetFileFlags(aFile)&64) == 0)) {
  CloseHandle(_FrameWork_::Get_File_w32(aFile));
  Zap_File_w32(aFile);}
  _FrameWork_::Set_File_Flagz(aFile, (GetFileFlags(aFile)|64));
  { return aFile;}} // ~Close2Ref

void FileClose(File aFile)
 { int* _tmp_00_;
  int flgz = GetFileFlags(aFile);
  if (((flgz&128) != 0))  { return;}
  aFile = Close2Ref(aFile);
  if (((flgz&24) == 16))  { return;}
  File_Destructor(aFile);} // ~FileClose

void FileDelete(File aFile)
 { int* _tmp_00_;
  String fiName = FullPath(aFile);
  if ((fiName == Null_Str))  { return;}
  if (((GetFileFlags(aFile)&192) != 64))  { return;}
  DeleteFile(EXTRACT_Cstr(fiName));
  File_Destructor(Close2Ref(aFile));} // ~FileDelete

void Beep()
 { int* _tmp_00_;
  if (DebugLogFile)  SystemDebugLog(String("!! --- Beep ---"));
  MessageBeep(-1);} // ~Beep

bool Valid_Process(Process whom)
 { int* _tmp_00_; bool _tmp__;
  T2F_Process here = (T2F_Process)up__ptr(_FrameWork_::Ready_List);
  T2F_Process whose = Cast_Proc(whom);
  while ((here != 0)) {
  if ((whose == here))  { _tmp__ = 1;
  down__ptr((int*)whose);
  down__ptr((int*)here); return _tmp__;}
  {_tmp_00_ = (int*)here;  here = (T2F_Process)up__ptr(here->_link_); down__ptr(_tmp_00_);}}
  { _tmp__ = 0;
  down__ptr((int*)whose);
  down__ptr((int*)here); return _tmp__;}}

Char4 ProcID(Process whom)
 { int* _tmp_00_; Char4 _tmp__;
  T2F_Process yo = 00;
  if ((whom == 0))  whom = MySelf();
  else  if ((!Valid_Process(whom)))  { _tmp__ = 0;
  down__ptr((int*)yo); return _tmp__;}
  if ((whom == 0))  { _tmp__ = 0;
  down__ptr((int*)yo); return _tmp__;}
  {_tmp_00_ = (int*)yo;  yo = Cast_Proc(whom); down__ptr(_tmp_00_);}
  { _tmp__ = yo->_my_ID_;
  down__ptr((int*)yo); return _tmp__;}}

Process Proc4ch(Char4 pnam)
 { int* _tmp_00_; Process _tmp__;
  T2F_Process whom = (T2F_Process)up__ptr(_FrameWork_::Ready_List);
  if ((pnam <= Char4("    ")))  { _tmp__ = MySelf();
  down__ptr((int*)whom); return _tmp__;}
  while ((whom != 0)) {
  if ((whom->_my_ID_ == pnam))  break;
  {_tmp_00_ = (int*)whom;  whom = (T2F_Process)up__ptr(whom->_link_); down__ptr(_tmp_00_);}}
  { _tmp__ = Proc_Cast((T2F_Process)up__ptr(whom));
  down__ptr((int*)whom); return _tmp__;}} // ~Proc4ch

void Stop4Debug(Char4 msg)
 { int* _tmp_00_;
  int ix = 0;
  int y = 1;
  if (DebugLogFile) {
  _FrameWork_::Log_Panic(String("!? ---- "));
  SystemDebugLog(_cat_(_cat_(String("!? ---- "), String((Char4)msg)), String(" ----")));}
  if (_Crash_and_Burn_)  ix = (y/ix);
  throw msg;} // ~Stop4Debug

void MonitorStrings(bool doWhat)
 { int* _tmp_00_;
  _FrameWork_::Monitor_Strings(doWhat);}
 
namespace  _FrameWork_ { // resuming
 
/*

void FrameWork_main()
 { int* _tmp_00_;
  Quit();}
*/

 } // (end namespace _FrameWork_) /// #include "FrameWork.h"

////////////////////////////////////////////////////////////////////////

int WINAPI WinMain(HINSTANCE hx, HINSTANCE px, PSTR cx, int sho) {
  // Win32 startup procedure, calls MOS_main() after initializing
  bool GotOle = false;
  MSG  theEvent;
  _FrameWork_::Main_Context = hx;
  if (!_FrameWork_::InitializeAll(hx,GotOle)) {
    MessageBox(NULL,"Oops",ProgramName,0);
    return 0;}
  // GotOle = (OleInitialize(0) == S_OK); // Initialize COM and OLE

  // MessageBox(NULL,"Hello",ProgramName,0); // this one works

#ifndef MOS_System // System.h does this after its inits..
  _FrameWork_::MainProgram = NewProgram(ProgramName,NULL,ProgramName);
#endif // MOS_System

  while (GetMessage(&theEvent,NULL,0,0)) {
    TranslateMessage(&theEvent);
    DispatchMessage(&theEvent);} //~while

  if (_FrameWork_::MainProgram != NULL) // shouldn't happen, but just in case..
    _FrameWork_::Quit_Proc(_FrameWork_::MainProgram);
  if (DebugLogFile) {
    _FrameWork_::Exit_MOMAM();
    FileClose(DebugLogFile);}
  DebugLogFile = NULL;
  if (GotOle) { // do the following when we actually quit...
    OleUninitialize();}

  return 0;} // ~WinMain

#ifdef __MSL__ // for CW linker
  void main() {WinMain(NULL,NULL,NULL,0);}
#endif

#undef _BASE_

#ifdef Debug_Trace
#undef Debug_Trace
#endif
#undef Post_Trace
#undef Sys_Trace
#undef SigWait_Trace
#undef Draw_Trace
#undef _Str_Logit_
#undef _Str_unLog_
#undef _String_Log_
#undef WinFile
#undef PathSize
#undef App_Icon
#undef Ask_Text
#undef IDC_EdLn
#undef IDC_AskLabel
#undef Int_CAST
#undef Fi_Find_32
#undef FlagMask
#undef File_Destructor
#undef Zap_File_w32
//#undef BaseM_LineHgt

#undef Verify_Valid_Ptr

#if Malloc_Log == 5
#define Verify_Valid_Ptr(ptr) Valid_Ptr_Verify(ptr)
#else
#define Verify_Valid_Ptr(ptr) 0
#endif

#undef PEEK_OFTEN
#undef Malloc_Log

#ifndef MOS_System // these are needed in System.h ..
//#undef _w32RECT_
#undef w32RECT
#undef wPOINT
//#undef _wPOINT_
#undef FILL_0
#undef int_2_ptr
#undef ptr_2_int
#undef C_string
#undef EXTRACT_Cstr
#undef MAKE_Cstr
#undef Literal_Cstr
#undef WinWindow
#undef Cast_Proc
#undef Proc_Cast

#undef _rTop
#undef _rLeft
#undef _rBotm
#undef _rRite
#undef _wKind
#undef _wName
#undef _wFlags
#undef _wLink
#undef _wParent
#undef _wChild
#undef _wWindo
#undef _FoBgColo
#undef _theText
#undef _MoInfo
#undef _wOwner
#undef _wNxID
#undef _wFocus
#undef _wTextVH
#undef _wUserData
#undef _wMoreInfo
#undef WinFlagsOffs

#undef Sem32
#undef NewSem32
#undef Signal32
#undef Wait32

#undef Windo_Event
#undef WEV_HidenShoWgt
#undef WEV_Delay
#undef WEV_MsgBx
#undef WEV_AskTx
#undef WEV_CheckMenu
#undef WEV_AddMenu
#undef WEV_UndoMenu
#undef WEV_NewMenBar
#undef WEV_PopMenu
#undef WEV_NewWindow
#undef WEV_CloseWind
#undef WEV_ShowWind
#undef WEV_WindowAcc
#undef WEV_HideWind
#undef WEV_InvalRect
#undef WEV_DocScroll
#undef WEV_WigitDraw
#undef WEV_IniWidgit
#undef WEV_PkgDone
#undef WEV_Splsh
#undef WEV_NoHigher

#undef Do_Gooey
#undef Gooey_Do
#undef Panic_Log
#undef Event_Log
#undef See_Panic

#undef Drag_Trace
#undef Menu_LineHeight
#undef Menu_Comm_base
#undef Max_Menus
#undef _Crash_and_Burn_
#undef DebugLogFile
#endif // MOS_System

#undef CR_LF
#undef CRLF_LEN

#undef FrameWork_Body
#endif // MOS_FrameWork

////////////////////////////////////////////////////////////////////////

// Standard type conversions and other built-ins..
// array management routines..
// Opaque Point type, implemented as an int..
// Predefined Rectangle class
// 4-character-as-int (Char4) class, intrinsic to Turk/2..
// Global declarations used in WinMain/InitializeAll..
// early forward declarations..
// ref-counted String class, intrinsic to Turk/2..
// Opaque File type, implemented as a pointer to a _File_ class..
// Forward function declarations, mostly defined in System.h ..
// array bounds check..
// Process management stuff..
// Predefined Event class, implemented as a pointer to class _Event_
// DragonDrop data and functions..
// Graphics-related operations..
// Opaque Window type, implemented as a pointer to a WindInfo class..
// Utility Functions not elsewhere defined..
// [Part 2] D&D code adapted from <http://www.catch22.net/tuts/dragdrop.asp>
// Win32 Registry Access..
// MsgBox and AskText..
// Win32 event support..
// #include "FrameWork.h"

////////////////////////////////////////////////////////////////////////

